/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Fri Jun  7 02:30:13 2002 by Jeff Dalton
 * Copyright: (c) 2001, 2002, AIAI, University of Edinburgh
 */

package ix.icore;

import java.net.UnknownHostException;

import javax.swing.JOptionPane;

import ix.icore.domain.Constraint;
import ix.ichat.ChatMessage;

import ix.iface.util.Reporting;
import ix.ispace.*;
import ix.util.*;

/**
 * Common class for I-X agents. <p>
 *
 * An agent is usually created by instantiating a subclass of IXAgent
 * and then calling its <code>mainStartup(String[] argv)</code> method.
 * That method will usually be inherited from IXAgent; it processes any
 * command line arguments and then calls <code>startup()</code>.
 * Thus, from the subclass's point of view, startup and initialization
 * has the following steps:
 * <ol>
 * <li>One of the subclass's constructurs, which will call superclass
 *     constructors.
 * <li><code>mainStartup(String[] argv)</code>, which calls
 *    <ol>
 *    <li><code>processCommandLineArguments()</code>
 *    <li><code>startup()</code>
 *    </ol>
 * </ol>
 * A subclass that redefines <code>processCommandLineArguments()</code>
 * or <code>startup()</code> should normally have them call the
 * corresponding <code>super</code> method to ensure that any
 * common processing is performed.  For example, the IXAgent
 * <code>processCommandLineArguments()</code> method will set up
 * interprocess communication as specified by the argument
 * <tt>-ipc=</tt><i>className</i>. <p>
 *
 * To support IPC, a subclass of IXAgent must assign a value
 * to the <code>ipcName</code>field, to be returned by the
 * <code>getAgentIPCName()</code> method, and will typically want
 * to define
 * <ul>
 * <li><code>handleNewIssue(Issue issue)</code>
 * <li><code>handleNewActivity(Activity activity)</code>
 * <li><code>handleNewConstraint(Constraint constraint)</code>
 * <li><code>handleNewReport(Report report)</code>
 * <li><code>handleNewChatMessage(ChatMessage message)</code>
 * </ul>
 * or, to intervene at an earlier stage, the method that calls
 * the ones above,
 * <ul>
 * <li><code>handleInput(IPC.InputMessage message)</code>
 * </ul>
 *
 * @see ix.test.SimpleIXAgent
 * @see <a href="../../../../ix/test/SimpleIXAgent.java">SimpleIXAgent.java 
 *      source code</a>
 */
public abstract class IXAgent {

    // The dreaded global variable ...
    protected static IXAgent thisAgent = null;

    /**
     * Name used for IPC purposes and returned by the
     * <code>getAgentIPCName()</code> method.
     */
    protected String ipcName = "anonymous";

    // /\/: The initial value of displayName is used as part of the
    // displayName default, which may be confusing.

    protected String displayName = "I-X Agent";
    protected String symbolName = "anonymous";

    protected ContactManager contactManager = new ContactManager();

    /**
     * Standard constructor. <p>
     */
    public IXAgent() {
	if (thisAgent != null)
	    throw new Error("Attempt to create multiple IX Agents in one VM");
	else
	    thisAgent = this;

	// /\/: Theme change doesn't work if done after objects created,
	// so is done here rather than in startup().
	// ix.iface.Iface.adjustLookAndFeel();
    }

    /**
     * Method called by main(String[] argv) to do initialization common
     * to all I-X agents.
     */
    public void mainStartup(String[] argv) {

	// Parse command-line arguments
	Parameters.processCommandLineArguments(argv);

	// Have agent to look at any arguments it's interested in.
	processCommandLineArguments();

	// Finish setting things up
	startup();

	// All command-line arguments should now have been processed.
	Parameters.checkParameterUse();

    }

    /**
     * Method called by the <code>mainStartup(String[] argv)</code> method
     * to perform any setup and initialization that should take place after
     * this agent's constructor has been called and command-line arguments
     * have been processed. <p>
     *
     * At present, this method does nothing, and all the work is done
     * in subclasses.
     */
    protected void startup() {
    }

    /**
     * Handles command-line arguments common to all I-X agents.
     * At present, this method also makes any standard I-X changes
     * to the look and feel.
     *
     * <p>The following are handled directly my this method:
     * <pre>
     *    -debug=<i>boolean</i>
     *    -ipc=<i>name</i>
     *    -ipc-name=<i>name</i>
     *    -symbol-name=<i>name</i>
     *    -display-name=<i>name</i>
     * </pre>
     * <tt>debug</tt> is used to set <code>Debug.on</code>.
     *
     * <p>The <i>name</i> in <tt>-ipc=</tt><i>name</i> argument will be
     * interpreted by the <code>IPC.getCommunicationStrategy(String)</code>
     * method.
     *
     * <p>The <i>name</i> in the <tt>-icp-name=</tt><i>name</i>
     * argument sets the name that this agent calls itself for IPC
     * and that is returned by the <code>getAgentIPCName()</code>
     * method.
     *
     * <p><i>Needs further explanation of ipc-name, and of symbol-name
     * and display-name. ...</i>
     *
     * <p>The <code>processCommandLineArguments</code> method of
     * this agent's contact manager is called to handle arguments
     * that list relationships with other agents.
     *
     * @see ix.iface.util.IfaceUtil#adjustLookAndFeel()
     * @see ix.util.IPC#getCommunicationStrategy(String methodName)
     * @see ix.ispace.ContactManager#processCommandLineArguments()
     * @see ix.util.Debug#on
     * @see ix.util.Parameters
     */
    protected void processCommandLineArguments() {

	ix.iface.util.IfaceUtil.adjustLookAndFeel();

	Debug.on = Parameters.getBoolean("debug", true);

	if (Parameters.haveParameter("ipc-name")) {
	    ipcName = Parameters.getParameter("ipc-name");
	    Debug.noteln("Using ipc name", ipcName);
	}
	else if (Parameters.haveParameter("symbol-name")) {
	    ipcName = Parameters.getParameter("symbol-name");
	    Debug.noteln("Using symbol name " + ipcName + " as ipc name");
	}
	else {
	    try { 
		ipcName = "IX-" + Util.getUserName() +"@"+ Util.getHostName();
	    }
	    catch (UnknownHostException e) {
		ipcName = "IX-" + Util.getUserName();
		Util.displayMessage(null, "Can't get host name; " +
				    "using ipc-name=" + ipcName);
	    }
	}

	symbolName  = Parameters.getParameter("symbol-name", ipcName);

	displayName = Parameters.haveParameter("display-name")
	    ? Parameters.getParameter("display-name")
	    : symbolName + " " + displayName;

	if (Parameters.haveParameter("ipc")) {
	    String ipc = Parameters.getParameter("ipc");
	    if (!(ipc.equals("false")			// from "-no -ipc"
		  || ipc.equals("none"))) 		// from "-ipc=none"
		// N.B. need to do this after setting ipcName
		startServer(getAgentIPCName(), ipc);
	}

	contactManager.processCommandLineArguments();

    }

    /**
     * Returns an object that represents the agent.
     * At present, there can be only one IXAgent per VM.
     */
    public static IXAgent getAgent() {
	return thisAgent;
    }

    /**
     * Returns the object used to represent the agent as an IPC "destination".
     * This object is usually, but not necessarily, a string containing the
     * agent's name.  The details will depend on the set of communication
     * strategies that might be used.
     *
     * @see ix.util.IPC
     * @see ix.util.IPC.CommunicationStrategy
     */
    public Object getAgentIPCName() {
	return ipcName;
    }

    /**
     * Returns this agent's symbol name.
     */
    public String getAgentSymbolName() {
	// /\/: Explain "symbol name"
	// /\/: Why not abstract?
	// /\/: Will symbol name and ipc name merge?
	return symbolName;
    }

    /**
     * Returns this agent's display name.
     */
    public String getAgentDisplayName() {
	return displayName;
    }

    /**
     * Returns this agent's contact manager.
     */
    public ContactManager getContactManager() {
	return contactManager;
    }

    /**
     * Set the default IPC communication strategy and set up
     * to receive messages.
     *
     * @see ix.util.IPC
     */
    protected void startServer(Object agentName, String strategyName) {
	// Create the communication strategy.
	IPC.setDefaultCommunicationStrategy(strategyName);
	// Set up to receive messages.
	IPC.setupServer(agentName, new IPC.MessageListener() {
	    public synchronized	void
		   messageReceived(IPC.InputMessage message) {
		pre_handleInput(message);
	    }
	});
    }

    /**
     * Gets the message before the <code>handleInput</code> method
     * and ensures that <code>handleInput</code> is called in the
     * AWT event dispatching thread.  This is perhaps a temporary
     * measure.  It also (once in the event dispatching thread)
     * catches any Errors or Exceptions thrown out of input handling
     * and reports them to the user.
     */
    protected void pre_handleInput(final IPC.InputMessage message) {
	// /\/: We have to do this somewhere, and it's awkward
	// to do it anywhere else for reports.  So for now ...
	javax.swing.SwingUtilities.invokeLater(new Runnable() {
	    public void run() {
		try {
		    notePossibleNewContact(message); // s.b. elsewhere?  /\/
		    handleInput(message);
		}
		catch (Throwable t) {
		    Debug.noteException(t);
		    tellUserOfException(message, t);
		}
	    }
	});

    }

    /**
     * Called by the pre_handleInput method to put up a dialog
     * describing an Error or Exception and the message that caused it
     * to be thrown.
     */
    protected void tellUserOfException(IPC.InputMessage message, Throwable t) {
	Object contents = message.getContents();
	String sender = contents instanceof Sendable
	    ? ((Sendable)contents).getSenderId().toString()
	    : "unknown sender";
	JOptionPane.showMessageDialog(null,
	    new Object[] {
	        "Problem handling message from " + sender,
		"Contents class " + contents.getClass().getName(),
		 t.getClass().getName(),
		 t.getMessage()},
	    "Error during message handling",
	    JOptionPane.ERROR_MESSAGE);
    }

    /**
     * Tells this agent's contact manager about the sender-id of the
     * message (if the id can be determined) in case it represents
     * a new contact.
     */
    public void notePossibleNewContact(IPC.InputMessage message) {
	Object contents = message.getContents();
	if (contents instanceof Sendable) {
	    Name senderId = ((Sendable)contents).getSenderId();
	    if (senderId != null)
		// /\/: Contact manager should work with Names, not Strings.
		contactManager.noteAgent(senderId.toString());
	}
    }

    /**
     * Handles external input in the form of an IPC.InputMessage
     * that contains an Issue or Report object.
     *
     * It calls one of <code>handleNewIssue</code>, 
     * <code>handleNewReport</code>, etc as appropriate.
     */
    public void handleInput(IPC.InputMessage message) {
	Object contents = message.getContents();
        if (contents instanceof Issue) {
            handleNewIssue((Issue)contents);
        }
	else if (contents instanceof Activity) {
	    handleNewActivity((Activity)contents);
	}
	else if (contents instanceof Constraint) {
	    handleNewConstraint((Constraint)contents);
	}
        else if (contents instanceof Report) {
            handleNewReport((Report)contents);
        }
	else if (contents instanceof ChatMessage) {
	    handleNewChatMessage((ChatMessage)contents);
	}
        else {
            displayMessage("Unexpected message contents " + contents);
        }
    }

    /**
     * Handles new issues from external sources.  Subclasses will
     * usually redefine this method.
     */
    public void handleNewIssue(Issue issue) {
        displayMessage(Reporting.issueDescription(issue));
    }

    /**
     * Handles new activities from external sources.  Subclasses will
     * usually redefine this method.
     */
    public void handleNewActivity(Activity activity) {
        displayMessage(Reporting.activityDescription(activity));
    }

    /**
     * Handles new constraints from external sources.  Subclasses will
     * usually redefine this method.
     */
    public void handleNewConstraint(Constraint constraint) {
        displayMessage(Reporting.constraintDescription(constraint));
    }

    /**
     * Handles new reports from external sources.  Subclasses will
     * usually redefine this method.
     */
    public void handleNewReport(Report report) {
        displayMessage(Reporting.reportDescription(report));
    }

    /**
     * Handles new chat messages from external sources.  Subclasses will
     * usually redefine this method.
     */
    public void handleNewChatMessage(ChatMessage message) {
        displayMessage(Reporting.chatMessageDescription(message));
    }

    /*
     * Default message display
     */

    /**
     * A text area in a separate frame used to display information about
     * incoming messages.
     */
    protected TextAreaFrame textFrame = null;

    /**
     * Adds the specified string to the default message display.
     */
    protected void displayMessage(String message) {
	if (textFrame == null)
	    textFrame = new TextAreaFrame("Messages to " + getAgentIPCName());
	textFrame.appendLine(message);
    }

}

// Issues:
// * Should keep a record of how IXAgent.getAgent() is used.
// * If a startup() method makes the GUI visible, subclass startup()
//   methods that call super.startup() will need to know this so they can
//   ensure that subsequent GUI stuff happens in the event-handling thread.
//   It might be better to have a separate makeVisible() that happens
//   after startup(); and maybe startup() should be renamed to something
//   like completeSetup().
