/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Wed Apr  3 23:19:57 2002 by Jeff Dalton
 * Copyright: (c) 2001 - 2002, AIAI, University of Edinburgh
 */

package ix.icore;

import java.util.*;
import java.io.Serializable;

import ix.util.*;
import ix.util.lisp.*;

/**
 * Something an agent might be asked to do or to handle.  Essentially
 * this class is just to hold what Issues and Activities have in common,
 * which at this level is almost everything.
 */
public abstract class TaskItem extends AbstractIXObject
    implements Serializable, Cloneable, Sendable {

    protected LList pattern;
    protected Status status = Status.BLANK;
    protected Priority priority = Priority.NORMAL;
    protected Name id;		// this item, w/in this agent
    protected Name senderId;	// source agent name, if source is external
    protected Name ref;		// ref to use in any reports back
    protected YesNo reportBack;	// set in forwarded copy

    protected String shortDescription; // set when pattern's set

    protected Set patternVars;	// set when pattern's set

    protected TaskItem() {
	this.id = generateId();
    }

    protected TaskItem(LList pattern) {
	this();
	setPattern(pattern);
    }

    public LList getPattern() { return pattern; }

    public void setPattern(LList pattern) {
	this.pattern = pattern;
	this.shortDescription = Lisp.elementsToString(pattern);
	this.patternVars = Variable.varsIn(pattern);
    }

    public Object getVerb() { return pattern.car(); }

    public LList getParameters() { return pattern.cdr(); }

    public String getShortDescription() { return shortDescription; }

    public Set getPatternVars() { return patternVars; }

    public Set getUnboundVars() {
	return Variable.unboundVarsIn(patternVars);
    }

    public Name getId() { return id; }
    // N.B. no setId method - always set by constructor

    public abstract Name generateId();

    public Status getStatus() { return status; }

    public void setStatus(Status status) { this.status = status; }

    public Priority getPriority() { return priority; }

    public void setPriority(Priority priority) { this.priority = priority; }

    public Name getSenderId() { return senderId; }

    public void setSenderId(Name id) { this.senderId = id; }

    public Name getRef() { return ref; }

    public void setRef(Name ref) { this.ref = ref; }

    public YesNo getReportBack() { return reportBack; }

    public void setReportBack(YesNo rep) { this.reportBack = rep; }

    /**
     * Sends a copy of this item to another agent, optionally asking
     * the destination agent to send reports back.
     */
    public void forwardTo(Object destination, boolean reportBack) {
	Object ipcName = IXAgent.getAgent().getAgentIPCName();
	try {
	    TaskItem copy = (TaskItem)clone();
	    // /\/: Remove Variables unless we get a better idea.
	    copy.setPattern((LList)Variable.removeVars(this.getPattern()));
	    copy.setSenderId(Name.valueOf(ipcName));
	    copy.setReportBack(YesNo.valueOf(reportBack));
	    copy.setRef(this.id);
	    IPC.sendObject(destination, copy);
	}
	catch (CloneNotSupportedException e) {
	    throw new Error("Can't clone " + this);
	}
    }

    public Object clone() throws CloneNotSupportedException {
	TaskItem copy = (TaskItem)super.clone();
	copy.id = generateId();
	return copy;
    }

}
