/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Mon May 27 01:59:54 2002 by Jeff Dalton
 * Copyright: (c) 2000 - 2002, AIAI, University of Edinburgh
 */

package ix.icore.domain;

import java.util.*;

import ix.icore.*;
import ix.icore.domain.event.*;
import ix.util.*;
import ix.util.lisp.*;

/**
 * A Domain contains descriptions of process products and activities.
 */

public class Domain extends AbstractIXObject implements Named, Cloneable {

    protected String name;

    protected List refinements = new LListCollector();

    protected List listeners = new LListCollector();

    protected Map nameMap = new HashMap();

    public Domain() {
	super();
    }

    public String getName() {
	return name;
    }

    public void setName(String name) {
	this.name = name;
    }

    public List getRefinements() {
	return refinements;
    }

    public void setRefinements(List refinements) {
	this.refinements = refinements;
    }

    /** Merges two domains */
    public void takeFrom(Domain other) {
	Debug.noteln(this + " takes from " + other);
	// Name
	if (other.name != null) {
	    if (name != null) {
		List parts = Strings.breakAt("+", name);
		if (!parts.contains(other.name)) {
		    name = name + "+" + other.name;
		    Debug.noteln("Extended name of", this);
		}
	    }
	    else {
		name = other.name;
		Debug.noteln("New name for", this);
	    }
	}
	// Refinements
	for (Iterator i = other.getRefinements().iterator(); i.hasNext();) {
	    addRefinement((Refinement)i.next());
	}
    }

    public boolean isEmpty() {
	return refinements.isEmpty();
    }

    public void clear() {
	Debug.noteln(this + " cleared");
	refinements.clear();
	nameMap.clear();
    }

    public Refinement getNamedRefinement(String name) {
	return (Refinement)nameMap.get(name);
    }

    public void addRefinement(Refinement r) {
	Debug.noteln(this + " adds " + r);
	nameMap.put(r.getName(), r);
	refinements.add(r);
	fireRefinementAdded(r);
    }

    public void deleteNamedRefinement(String name) {
	deleteRefinement(getNamedRefinement(name));
    }

    public void deleteRefinement(Refinement r) {
	Debug.noteln(this + " deletes " + r);
	// /\/: Use with great caution.  Also, this metod should
	// tell the listeners but doesn't.
	refinements.remove(r);
	nameMap.remove(r.getName());
    }

    public void replaceNamedRefinement(String name, Refinement replacement) {
	replaceRefinement(getNamedRefinement(name), replacement);
    }

    public void replaceRefinement(Refinement old, Refinement neu) {
	Debug.noteln(this + " replaces " + old + " with " + neu);
	// /\/: Use with great caution.  Also, this metod should
	// tell the listeners but doesn't.
	// N.B. The Refinement's name might change.

	// Give the new refinement the old one's place in the list
	int i = refinements.indexOf(old);
	refinements.remove(i);
	refinements.add(i, neu);
	// Make the name map to the new refinement instead of the old.
	nameMap.put(old.getName(), neu);
    }

    public void addDomainListener(DomainListener listener) {
	listeners.add(listener);
    }

    public void fireRefinementAdded(Refinement r) {
	RefinementEvent event = new RefinementEvent(this, r);
	for (Iterator i = listeners.iterator(); i.hasNext();) {
	    DomainListener listener = (DomainListener)i.next();
	    listener.refinementAdded(event);
	}
    }

    public void analyseDomain() {

	// checkRefinementReferences();

    }

    public Object clone() throws CloneNotSupportedException {
	return super.clone();
    }

    public String toString() {
	return "Domain[" + name + ", "
	    + refinements.size() + " refinements]";
    }

}

// Issues:
// * Need a way to deal with a set of changes to a domain as a unit,
//   because it may be impossible to keep things consistent when changes
//   (such as refinement replacements) are done one at a time.
