/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Thu May  2 01:35:56 2002 by Jeff Dalton
 * Copyright: (c) 2002, AIAI, University of Edinburgh
 */

package ix.icore.domain;

import java.util.*;

import ix.icore.*;
import ix.util.*;
import ix.util.lisp.*;
import ix.util.match.*;

/**
 * A Refinement describes one way to carry out a possible activity
 * in a process being modelled.
 */

public class Refinement extends AbstractIXObject implements Named, Cloneable {
    protected String name;
    protected LList pattern;
    protected List nodes;	// List of NodeSpecs
    protected List orderings;	// List of Orderings, for now  /\/
    protected List constraints;
    protected List issues;
    // annotations
    protected String comments = "";

    // Need a better place for these, here and in the Ip2 Model Manager /\/
    public static final Symbol
	S_WORLD_STATE = Symbol.intern("world-state"),
	S_CONDITION   = Symbol.intern("condition"),
	S_EFFECT      = Symbol.intern("effect");
    
    public Refinement() {
    }

    public Refinement (String name, LList pattern) {
	this.name = name;
	this.pattern = pattern;
	Debug.noteln("New blank refinement " + Util.quote(name) + " for",
		     pattern);
    }

    public String getName() { return name; }
    public void setName(String name) { this.name = name; }

    public LList getPattern() {	return pattern; }
    public void setPattern(LList pattern) { this.pattern = pattern; }

    public List getNodes() { return nodes; }
    public void setNodes(List nodes) { this.nodes = nodes; }

    public List getOrderings() { return orderings; }
    public void setOrderings(List orderings) { this.orderings = orderings; }

    public List getConstraints() { return constraints; }
    public void setConstraints(List cs) { this.constraints = cs; }

    public List getIssues() { return issues; }
    public void setIssues(List issues) { this.issues = issues; }

    public String getComments() { return comments; }
    public void setComments(String comments) { this.comments = comments; }

    public List getFilterConditions() {
	// N.B. for now returns a list of PatternAssignments /\/
	List result = new LinkedList();
	for (Iterator i = Collect.iterator(constraints); i.hasNext();) {
	    Constraint c = (Constraint)i.next();
	    if (c.getType() == S_WORLD_STATE
		    && c.getRelation() == S_CONDITION) {
		PatternAssignment pv = (PatternAssignment)c.getParameter(0);
		result.add(pv);
	    }
	}
	return result;
    }

    public void checkConsistency() {
    }

    public Refinement instantiate(MatchEnv env) {
	return instantiate(env, new Function1() {
	    public Object funcall(Object name) {
		throw new RuntimeException("Unbound variable " + name);
	    }
	});
    }

    public Refinement instantiate(MatchEnv env, Function1 ifUnbound) {
	try {
	    Refinement r = (Refinement)clone();
	    // Pattern
	    r.pattern = (LList)env.instantiateTree(r.pattern, ifUnbound);
	    // Nodes
	    if (r.nodes != null) {
		r.nodes = LList.newLList(r.nodes);	// copy list
		for (ListIterator i = r.nodes.listIterator(); i.hasNext();) {
		    NodeSpec spec = (NodeSpec)i.next();
		    i.set(spec.instantiate(env, ifUnbound));
		}
	    }
	    // Issues
	    if (r.issues != null) {
		r.issues = LList.newLList(r.issues);
		for (ListIterator i = r.issues.listIterator(); i.hasNext();) {
		    Issue issue = (Issue)i.next();
		    i.set(issue.instantiate(env, ifUnbound));
		}
	    }
	    // Constraints
	    if (r.constraints != null) {
		r.constraints = LList.newLList(r.constraints);
		for (ListIterator i = r.constraints.listIterator();
		     i.hasNext();) {
		    Constraint c = (Constraint)i.next();
		    i.set(c.instantiate(env, ifUnbound));
		}
	    }
	    return r;
	}
	catch (CloneNotSupportedException e) {
	    Debug.noteException(e);
	    return null;
	}
    }

    public Object clone() throws CloneNotSupportedException {
	return super.clone();
    }

    public String toString() {
	return "Refinement[" + name + " expands " + pattern + "]";
    }

}

// Issues:
// * Should we define an Instantiable interface for objects that
//   provide an instantiate(MatchEnv, Function1) method?
// * When a field F has declared type List, you can't call clone() without
//   casting it to a concrete class (and we don't want to depend on which
//   concrete class it is), and to make an LList of a List that might be
//   empty, you can't use a 1-argument constructor; so we use the idiom:
//      field = LList.newLList(field)
//   N.B. "newLList", no space.
