/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Wed May  8 02:23:43 2002 by Jeff Dalton
 * Copyright: (c) 2001, 2002, AIAI, University of Edinburgh
 */

package ix.icore.process;

import java.util.*;

import ix.icore.domain.*;
import ix.icore.Variable;
import ix.icore.Status;
import ix.util.*;
import ix.util.lisp.*;
import ix.util.match.*;

/**
 * A node that represents an action in a process.  This class
 * supports expansion into subactions and the propagation of status
 * changes, thus acting as a simple process model / simulator and
 * expansion planner.
 */
public abstract class PNode {

//  public static Map varEnv = new HashMap(); // s.b. per model manager /\/

    public ProcessModelManager modelManager;

    public int level = 0;		// expansion level, 1 + the parent's

    public PNode parent = null;
    public LList children = Lisp.NIL;

    public Map nameToChildMap = new HashMap();

    public Map varTable = new HashMap(); // when instantiating a refinement
    // maps ?name to instance of Variable.

    public List preNodes  = new LinkedList(); // nodes directly before this
    public List postNodes = new LinkedList(); // nodes directly after this

    public Refinement refinement = null;

    public PNode(PNode parent) {

	this.parent = parent;
	if (parent != null)
	    level = parent.level + 1;

	// computeStatus();	// may be too soon from subclass POV /\/

    }

    public abstract Status getStatus();

    public boolean isAncestorOf(PNode n) {
	for (PNode p = n.parent; p != null; p = p.parent) {
	    if (this == p)
		return true;
	}
	return false;
    }

    public boolean isDescendentOf(PNode n) {
	return n.isAncestorOf(this);
    }

    /**
     * Creates child nodes as specified by a refinement and installs
     * ordering links.  The subnodes are constructed by a NodeMaker
     * so that they can be instances of a subclass of Node and to
     * allow other, related objects to be constructed at the same time.
     */
    public void expandOneLevel(Refinement sourceRefinement,
			       MatchEnv env,
			       final PNodeMaker nodeMaker) {

	// Instantiate the refinement
	Debug.assert(refinement == null);
	Function1 ifUnbound = new MakeVarIfUnbound(varTable);
	refinement = sourceRefinement.instantiate(env, ifUnbound);
	Debug.noteln("varTable after expansion", varTable);

	// Create the child nodes specified by the refinement.
	LListCollector c = new LListCollector();
	for (Iterator ni = Collect.iterator(refinement.getNodes());
	     ni.hasNext();) {
	    NodeSpec spec = (NodeSpec)ni.next();
	    LList childPattern = (LList)spec.getPattern();
	    PNode child = nodeMaker.makePNode(PNode.this, childPattern);
	    c.add(child);
	    nameToChildMap.put(spec.getId(), child);
	}
	children = c.contents();

	// Add the orderings
	// N.B. uses fields children and nameToChildMap.
	if (refinement.getOrderings() != null)
	    processOrderings(refinement.getOrderings());

	// Have all the children determine their status
	allComputeStatus(children);

    }

    public class MakeVarIfUnbound implements Function1 {

	protected Map nameToVarMap;

	public MakeVarIfUnbound(Map nameToVarMap) {
	    this.nameToVarMap = nameToVarMap;
	};

	public Object funcall(Object name) {
	    Variable v = (Variable)nameToVarMap.get(name);
	    if (v == null) {
		v = new Variable(name);
		Debug.noteln("Creating variable " + v + " for " + name);
		// v.sourceNode = PNode.this;
		// The refinement will be null for Variables 
		// in top-level patterns
		v.sourceRefinement = PNode.this.refinement;
		nameToVarMap.put(name, v);
	    }
	    return v;
	}

    }
	
    protected void processOrderings(List orderings) {
	for (Iterator ords = orderings.iterator(); ords.hasNext();) {
	    Ordering ord = (Ordering)ords.next();
	    // /\/: Handles only end->begin links for now.
	    Debug.assert(ord.getFrom().getEnd() == End.END);
	    Debug.assert(ord.getTo().getEnd() == End.BEGIN);
	    Name n1 = ord.getFrom().getNode();
	    Name n2 = ord.getTo().getNode();
	    PNode before = getChild(n1);
	    PNode after = getChild(n2);
	    before.linkBefore(after);
	}
    }

    protected PNode getChild(Name name) {
	PNode child = (PNode)nameToChildMap.get(name);
	if (child == null)
	    throw new RuntimeException("No child " + name);
	else
	    return child;
    }

    protected void linkBefore(PNode after) {
	Debug.noteln("- " + this + " --> " + after);
	postNodes.add(after);
	after.preNodes.add(this);
	// Since after has a new preNode, its status may change.
	// after.computeStatus();
    }

    /**
     * computeStatus() is used to change the status of related nodes
     * when a node changes its status.
     */
    public void computeStatus() {
	// We assume that the current status may not be correct,
	// but only certain transitions are handled.

	// If the node is BLANK (waiting to become POSSIBLE), then:
	// if the parent is EXECUTING (or null) and all predecessors 
	// are COMPLETE, then this node should become POSSIBLE.
	if (getStatus() == Status.BLANK) {
	    if ((parent == null || parent.getStatus() == Status.EXECUTING)
		&& allHaveStatus(preNodes, Status.COMPLETE))
		setStatus(Status.POSSIBLE);
	}

	// If the node is EXECUTING, then:
	// if all children are COMPLETE, this this node should
	// become COMPLETE.
	else if (getStatus() == Status.EXECUTING) {
	    Debug.assert(!children.isNull(), "Executing w/o children");
	    if (allHaveStatus(children, Status.COMPLETE))
		setStatus(Status.COMPLETE);
	}
    }

    /**
     * Changes the node's status and then that of related nodes.
     */
    public void setStatus(Status status) {
	// A change in a node's status may affect the parent,
	// successor siblings, or children.
	Debug.noteln("Setting status of " + this + " to " + status);
	// /\/: The status value might not even be stored in this object.
	// this.status = status;

	// Let the model manager see the status change
	modelManager.statusChanged(this);

	// Propagate status change to related nodes.
	if (status == Status.COMPLETE) {
	    if (parent != null)
		parent.computeStatus();			// may -> COMPLETE
	    allComputeStatus(postNodes);	 	// may -> POSSIBLE
	}
	else if (status == Status.EXECUTING) {
	    allComputeStatus(children); 		// may -> POSSIBLE
	}
	else if (status == Status.POSSIBLE) {
	    // When a node that already had children becomes POSSIBLE,
	    // we automatically make it start EXECUTING
	    if (!children.isEmpty())
		setStatus(Status.EXECUTING);
	}
    }

    public static boolean allHaveStatus(List l, Status status) {
	Iterator i = l.iterator();
	while(i.hasNext()) {
	    PNode n = (PNode)i.next();
	    if (n.getStatus() != status)
		return false;
	}
	return true;
    }

    public static void allComputeStatus(List l) {
	Iterator i = l.iterator();
	while (i.hasNext()) {
	    PNode n = (PNode)i.next();
	    n.computeStatus();
	}
    }

}

