/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Thu May  2 00:57:15 2002 by Jeff Dalton
 * Copyright: (c) 2001 - 2002, AIAI, University of Edinburgh
 */

package ix.icore.process;

import java.util.*;

import ix.icore.Variable;
import ix.icore.process.event.*;

import ix.icore.domain.Constraint;

import ix.util.*;
import ix.util.lisp.*;
import ix.util.match.*;

public abstract class ProcessModelManager {

    protected List nodes = new LinkedList();

    protected Map varEnv = new HashMap();

    protected List processStatusListeners = new LinkedList();

    protected List sentinels = new LinkedList();

    public ProcessModelManager() {
    }

    public void reset() {
	varEnv.clear();
    }

    public void addNode(PNode node) {
	Debug.noteln("PMM adding node", node);
	node.modelManager = this;
	nodes.add(node);
    }

    public void addConstraints(PNode node, List constraints) {
	Debug.noteln("PMM adding constraints for node", node);
	for (Iterator ci = constraints.iterator(); ci.hasNext();) {
	    Constraint c = (Constraint)ci.next();
	    Debug.noteln("--- adding ", c);
	    addConstraint(node, c);
	}
    }

    protected abstract void addConstraint(PNode node, Constraint c);

    // /\/: Not clear that this single-constrint version should be public
    public abstract void addConstraint(Constraint constraint);

    public abstract MatchEnv evalFilters(List conds, MatchEnv env);

    /**
     * Tells the model manager to bind the indicated variables.
     * This is the only way in which values should ever be assigned.
     *
     * @see ix.icore.Variable
     */
    public abstract void newBindings(Map bindings);

    /**
     * Called by a node when its status changes.
     */
    public abstract void statusChanged(PNode node);

    /**
     * Returns a copy of the pattern in which ?-vars have been replaced
     * by Variables.
     */
    public LList putVariablesInPattern(LList pattern) {
	// When a subnode is introduced by expansion with a Refinement,
	// all ?-vars in its pattern are repladed by values or by
	// Variables when the Refinement is instantiated.  But when
	// a level-0 node is introduced, there is no refinement,
	// and this method is used to perform the replacement of
	// ?-vars by Variables using a global (one per model manager)
	// mapping.
	return
	    (LList)SimpleMatcher.emptyEnv
	        .instantiateTree(pattern, new MakeVarIfUnbound());
	// /\/: Don't need to make a new MakeVarIfUnbound each time.
    }

    class MakeVarIfUnbound implements Function1 {
	MakeVarIfUnbound() { }
	public Object funcall(Object name) {
	    Variable v = (Variable)varEnv.get(name);
	    if (v == null) {
		v = new Variable(name);
		// v.sourceNode = null;
		v.sourceRefinement = null;
		varEnv.put(name, v);
	    }
	    return v;
	}
    }


    /*
     * Sentinels
     */

    protected interface Sentinel extends Runnable {
	boolean isReady();
    }

    public void addSentinel(Sentinel r) {
	sentinels.add(r);
    }

    public void removeSentinel(Sentinel r) {
	sentinels.remove(r);
    }

    public void runSentinels() {
	for (Iterator i = sentinels.iterator(); i.hasNext();) {
	    Sentinel r = (Sentinel)i.next();
	    if (r.isReady()) {
		i.remove();
		r.run();
	    }
	}
    }


    /*
     * Listeners
     */

    public void addProcessStatusListener(ProcessStatusListener listener) {
	processStatusListeners.add(listener);
    }

    public void fireStatusUpdate() {
	ProcessStatusEvent event = new ProcessStatusEvent(this);
	for (Iterator i = processStatusListeners.iterator(); i.hasNext();) {
	    ProcessStatusListener listener = (ProcessStatusListener)i.next();
	    listener.statusUpdate(event);
	}
    }

    public void fireNewBindings(Map bindings) {
	ProcessStatusEvent event = new ProcessStatusEvent(this);
	for (Iterator i = processStatusListeners.iterator(); i.hasNext();) {
	    ProcessStatusListener listener = (ProcessStatusListener)i.next();
	    listener.newBindings(event, bindings);
	}
    }

    public void fireStateChange(Map delta) {
	ProcessStatusEvent event = new ProcessStatusEvent(this);
	for (Iterator i = processStatusListeners.iterator(); i.hasNext();) {
	    ProcessStatusListener listener = (ProcessStatusListener)i.next();
	    listener.stateChange(event, delta);
	}
    }

}
