/****************************************************************************
 * An interface model to be used to fill an IFormPanel.
 *
 * @author Jussi Stader
 * @version 2.0
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 *
 *****************************************************************************
 */
package ix.iface.ui;

import ix.*;
import ix.util.Debug;
import ix.util.lisp.*;
import ix.iface.ui.util.*;

import java.util.*;
import java.lang.reflect.*;
//import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/**
 * An abstract model class to be used to fill an IFormPanel.
 *
 */
public abstract class AbstractIFormModel implements IFormModel {
  List nameFields = Lisp.NIL;
  List displayFields = Lisp.NIL;
  HashMap fieldNames = new HashMap();
  HashMap fieldClasses = new HashMap();
  protected Object baseObject;
  Class baseObjectClass;
  HashSet listeners = new HashSet();
  boolean isUpdating = false;

  public AbstractIFormModel(Class objectClass) {
    super();
    baseObjectClass = objectClass;
    setFields();
  }
  public AbstractIFormModel(Object object) {
    this(object.getClass());
    baseObject = object;
  }

  public void clearData() {
    setObject(null);
  }


  public String getFieldName(String field) {
    /*
    if (fieldNames.isEmpty()) {
      List fields = getNameFields();
      fields.append(getDisplayFields());
      deriveNames(fields);
      fireModelChanged("fieldNames");
    }
    */
    return (String) fieldNames.get(field);
  }
  public Class getFieldClass(String field) {
    /*
    if (fieldClasses.isEmpty()) {
      List fields = getNameFields();
      fields.append(getDisplayFields());
      deriveClasses(fields);
      fireModelChanged("fieldClasses");
    }
    */
    return (Class) fieldClasses.get(field);
  }
  public double getFieldWeight(String field) {
    return 1.0;
  }


  public void setNameFields(List fields) {
    nameFields = fields;
    fieldNames.clear();
    fieldClasses.clear();
    deriveNames();
    deriveClasses();
    fireModelChanged("nameFields");
  }
  public void setNameFields(String[] fields) {
    List list = new LListCollector(Arrays.asList(fields));
    setNameFields(list);
  }
  public void setDisplayFields(List fields) {
    displayFields = fields;
    fieldNames.clear();
    fieldClasses.clear();
    deriveNames();
    deriveClasses();
    fireModelChanged("displayFields");
  }
  public void setDisplayFields(String[] fields) {
    List list = new LListCollector(Arrays.asList(fields));
    setDisplayFields(list);
  }

  /**
   * Get the fields that correspond to the object's name (e.g. "name", "id")
   */
  public List getNameFields() {
    /*
    if ((nameFields == null) || (nameFields.isEmpty())) {
      nameFields = this.deriveNameFields();
      fireModelChanged("nameFields");
    }
    */
    return nameFields;
  }

  /**
   * Get the titles to use for the object's name fields (e.g. "Name", "Id").
   * Derive them if they are not there.
   */
  public List getNameNames() {
    return getFieldNames(nameFields);
  }

  /**
   * Get the classes of the object's name fields 
   * (e.g. String.class, Object.class)
   */
  public List getNameClasses() {
    return getFieldClasses(nameFields);
  }

  /**
   * Get the fields of the object that are to be displayed (exc. names)
   */
  public List getDisplayFields() {
    /*
    if ((displayFields == null) || (displayFields.isEmpty())) {
      displayFields = this.deriveDisplayFields();
      fireModelChanged("displayFields");
    }
    */
    return displayFields;
  }

  /**
   * Get the titles to use for the object's display fields
   */
  public List getDisplayNames() {
    return getFieldNames(displayFields);
  }

  /**
   * Get the classes of the object's display fields 
   * (e.g. String.class, Object.class, Collection.class)
   */
  public List getDisplayClasses() {
    return getFieldClasses(displayFields);
  }


  public void setObject(Object o) {
    boolean changed = false;
    //Debug.note("form model: setting object");
    //Debug.noteln("Object class is " + baseObjectClass.getName());
    if ((o != null) && (!(o.getClass().equals(baseObjectClass)))) {
      setObjectClass(o.getClass());
      changed = true;
    }
    if (!((o == null) && (baseObject == null)) &&
	!((o != null) && o.equals(baseObject)))
      changed = true;
    baseObject = o;
    if (changed) fireModelChanged("object"); //real change (different object)
    else if (baseObject != null)
      fireModelChanged("allFields"); //just in case there were any changes
  }

  public Object getObject() {
    return baseObject;
  }

  /**
   * Set the class of the model's base object
   */
  public void setObjectClass(Class c) {
    if (baseObjectClass.equals(c)) return;
    if (baseObjectClass != null) clearFields();
    baseObjectClass = c;
    setFields(); //fires "allFields"
    //fireModelChanged("objectClass");
  }
  /**
   * Get the class of the model's base object
   */
  public Class getObjectClass() {
    if (baseObjectClass == null) 
      if (baseObject != null) 
	baseObjectClass = baseObject.getClass();
    return baseObjectClass;
  }


  /**
   * Get the current object's value for the given field.

   */
  public Object getValue(String field) {
    return UIUtil.getObjectFieldValue(baseObject, field);
  }
  

  public void setValue(String field, Object value) {
    setValue(baseObject, field, value);
    fireModelChanged("Field " + field);
  }
  public void setValue(Object object, String field, Object value) {
    Class fClass = getFieldClass(field);
    UIUtil.setObjectFieldValue(object, field, fClass, value);
  }


  public Field stringToField(String field) throws 
                              NoSuchFieldException, IllegalAccessException {
    Class oClass = getObjectClass();
    if (oClass == null) return null;
    else return oClass.getField(field);
  }
  public Field stringToField(Class oClass, String field) throws 
                              NoSuchFieldException, IllegalAccessException {
    if (oClass == null) return null;
    else return oClass.getField(field);
  }



  /**
   * Derives the name fields, display fields and the field names and classes.
   */
  public void setFields() {
    if (baseObjectClass == null) {
      Debug.noteln("Cannot set fields for a null base object class.");
      return;
    }
    else {
      isUpdating = true;
      if (nameFields.isEmpty()) nameFields = this.deriveNameFields();
      if (displayFields.isEmpty()) displayFields = this.deriveDisplayFields();
      List allFields = new LListCollector(nameFields);
      allFields.addAll(displayFields);
      if (fieldNames.isEmpty()) deriveNames(allFields);
      if (fieldClasses.isEmpty()) deriveClasses(allFields);
      isUpdating = false;
      fireModelChanged("allFields");
    }
  }

  private void clearFields() {
    nameFields = Lisp.NIL;
    displayFields = Lisp.NIL;
    fieldNames.clear();
    fieldClasses.clear();
    baseObject = null;
  }


  protected List lookupFieldNames(List fields) {
    return lookupFieldMaps(fields, fieldNames);
  }
  protected List lookupFieldClasses(List fields) {
    return lookupFieldMaps(fields, fieldClasses);
  }

  /**
   * Get the mappings of the given fields from the given list. 
   * Do not try to fill the map if it is not already there.
   */
  private static List lookupFieldMaps(List fields, HashMap map) {
    if (fields.isEmpty()) return Lisp.NIL;
    else {
      LListCollector values = new LListCollector();
      Iterator i = fields.iterator();
      while (i.hasNext()) {
	String field = (String)i.next();
	values.addElement(map.get(field));
      }
      return values;    
    }

  }


  /**
   * Get or derive the names of the given fields 
   */
  private List getFieldNames(List fields) {
    /*
    //if there are no field names, try to derive them
    if ((fieldNames == null) || (fieldNames.isEmpty())) {
      this.deriveNames(fields);
      fireModelChanged("fieldNames");
    }
    */
    return lookupFieldNames(fields);
  }
  /**
   * Get or derive the classes of the given fields 
   */
  private List getFieldClasses(List fields) {
    /*
    //if the map is empty, try to fill it
    if ((fieldClasses == null) || (fieldClasses.isEmpty())) {
      this.deriveClasses(fields);
      fireModelChanged("fieldClasses");
    }
    */
    return lookupFieldClasses(fields);
  }

  /** Collects name and display fields and calls deriveNames(fields). */
  protected void deriveNames() {
    List allFields = new LListCollector(nameFields);
    allFields.addAll(displayFields);
    deriveNames(allFields);
  }
  /** Collects name and display fields and calls deriveClasses(fields). */
  protected void deriveClasses() {
    List allFields = new LListCollector(nameFields);
    allFields.addAll(displayFields);
    deriveClasses(allFields);
  }

  /**
   * Derives the names of the given fields and puts them into
   * HashMap fieldNames.
   */
  abstract protected void deriveNames(List fields);

  /**
   * Derives the classes of the given fields and puts them into
   * HashMap fieldClasses.
   */
  abstract protected void deriveClasses(List fields);

  /**
   * Derives the fields to be used for displaying the object name(s)
   */
  abstract protected List deriveNameFields();

  /**
   * Derives the fields to be used for displaying the object information
   */
  abstract protected List deriveDisplayFields();

  public static Collection addIfField(Collection fields, 
				      Class c, String sField) {
    try {
      Field field = c.getField(sField);
      fields.add(sField);
      return fields;
    }
    catch (NoSuchFieldException fe) { return fields; }
  }

  public boolean isEditable(String field) {
    return true;
  }


  public void startUpdate() {
    isUpdating = true;
  }
  public void endUpdate() {
    isUpdating = false;
    fireModelChanged("allFields");
  }
  public void ignoreUpdate() {
    isUpdating = false;
  }


  public void fireModelChanged(String change) {
    if (isUpdating) return;
    if (listeners.isEmpty()) return;
    else {
      Iterator e = listeners.iterator();
      while (e.hasNext()) {
	FormModelListener l = (FormModelListener)e.next();
	l.formModelChanged(change);
      }
    }    
  }


  public void addModelListener(FormModelListener l) {    
    listeners.add(l);
  }


}
