/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Tue May 28 21:52:24 2002 by Jeff Dalton
 * Copyright: (c) 2001 - 2002, AIAI, University of Edinburgh
 */

package ix.ip2;

import java.util.*;

import ix.iface.util.Reporting;

import ix.ip2.event.*;
import ix.icore.*;
import ix.icore.process.PNode;
import ix.icore.domain.Refinement;
import ix.util.*;
import ix.util.lisp.*;
import ix.util.match.MatchEnv;

/**
 * An entry in an Agenda.
 */
public abstract class AgendaItem extends PNode {

    protected TaskItem about;

    protected String comments = "";

    protected List actions = new LinkedList();
    protected List reports = new LinkedList();
    protected List listeners = new LinkedList();

    public AgendaItem(TaskItem about) {
	this(null, about);
    }

    public AgendaItem(AgendaItem parent, TaskItem about) {
	super(parent);
	this.about = about;
	if (parent != null)
	    about.setPriority(parent.getPriority());
	// /\/: Don't make any status decisions until the item
	// is added to the Agenda.  This is so that ActivityItems
	// can consult the model manager.
	// computeStatus();
    }

    public TaskItem getAbout() {
	return about;
    }

    public LList getPattern() {
	return about.getPattern();
    }

    public Set getPatternVars() {
	return about.getPatternVars();
    }

    public Set getUnboundVars() {
	// Start with the ones from the issue or activity -
	// the vars in the pattern.
	Set vars = about.getUnboundVars();

	// Add unbound vars from this item's varTable -
	// from the top level of expansion.
	vars.addAll(Variable.unboundVarsIn(varTable.values()));

	// /\/: How about from all descendent nodes?

	return vars;
    }

    public void setPattern(LList pattern) {
	about.setPattern(pattern);
    }

    public String getShortDescription() {
	// N.B. changes when variables become bound, but since we
	// recompute it every time, we should always have the right
	// value.
	return Lisp.elementsToString(getPattern());
    }

    public int getLevel() {
	return level;
    }

    public AgendaItem getParent() {
	return (AgendaItem)parent;
    }

    public List getChildren() {
	return children;
    }

    public Status getStatus() {
	return about.getStatus();
    }
    public void setStatus(Status status) {
	if (this.getStatus() != status) {
	    if (about.getReportBack() == YesNo.YES)
		handleReportBack(status);
	    // Status isn't changed until after any report-back
	    // in case we can't send the report.  But is that right? /\/
	    about.setStatus(status);
	    super.setStatus(status);
	    fireStatusChanged();
	}
    }

    public Priority getPriority() {
	return about.getPriority();
    }
    public void setPriority(Priority priority) {
	Debug.noteln("Changing priority of " + this + " to " + priority);
	about.setPriority(priority);
	firePriorityChanged();
    }

    public String getComments() {
	return comments;
    }
    public void setComments(String comments) {
	this.comments = comments;
	fireAgendaItemEdited();
    }

    public List getActions() {
	return actions;
    }

    /**
     * Gives this item its say in whether an item-handler should
     * be able to give it handler actions.
     *
     * @see ItemHandler#appliesTo(AgendaItem item)
     */
    public boolean wantsActionsFrom(ItemHandler handler) {
	return true;
    }

    /**
     * Add an action to this issue.  Note that an action with the same
     * description may already exist.  We regard that as a problem for
     * the AgendaItemListeners.  For example, a user interface might use
     * the description as a way to identify the action and may not want
     * any ambiguity.
     */
    public void addAction(HandlerAction act) {
	Debug.assert(!hasAction(act), "Action added twice", act);
	HandlerAction exists = findAction(act.getActionDescription());
	if (exists != null) {
	    Debug.noteln("Action description not unique",
			 Util.quote(act.getActionDescription()));
	}
	actions.add(act);
	act.computeStatus();
	fireNewHandlerAction(act);
    }

    boolean hasAction(HandlerAction act) {
	return actions.contains(act);
    }

    HandlerAction findAction(String actionDescription) {
	// N.B. returns the 1st that matches; there may be > 1.
	for (Iterator i = actions.iterator(); i.hasNext();) {
	    HandlerAction act = (HandlerAction)i.next();
	    if (act.getActionDescription().equals(actionDescription))
		return act;
	}
	return null;
    }

    boolean actionCanBeTakenNow(HandlerAction act) {
	// Allow any action when status is POSSIBLE and no action when
	// the status is something else.
	// /\/: Not clear what we should do when the usuer manually
	// specifies an expansion.  Right now, this method isn't called.
	return getStatus() == Status.POSSIBLE
	    && act.isReady();
    }

    boolean wantsReport(Report report) {
	Name ref = report.getRef();
	return ref != null && ref.equals(about.getId());
    }

    void addReport(Report report) {
	Debug.assert(wantsReport(report), "unwanted report", report);
	Debug.noteln(this + " accepts " + report);

	String lineSeparator = System.getProperty("line.separator");
	comments = comments
	    + (comments.equals("") || comments.endsWith(lineSeparator)
	       ? "" : lineSeparator)
	    + Reporting.reportDescription(report) + lineSeparator;

	reports.add(report);
	fireNewReport(report);

	setStatusBasedOn(report);

    }

    void setStatusBasedOn(Report report) {
	// /\/: Change status even though officially we may not
	// be able to make this transition (e.g. from COMPLETE
	// to EXECUTING).
	if (!report.isCompletion())
	    setStatus(Status.EXECUTING);
	else if (report.isSuccess())
	    setStatus(Status.COMPLETE);
	else
	    setStatus(Status.IMPOSSIBLE);
    }

    List getReports() {
	return reports;
    }

    void handleReportBack(Status newStatus) {
	Debug.assert(about.getReportBack() == YesNo.YES);

	// See what we need to say
	ReportType type;
	if (newStatus == Status.COMPLETE)
	    type = ReportType.SUCCESS;
	else if (newStatus == Status.IMPOSSIBLE)
	    type = ReportType.FAILURE;
	else return;

	// Construct the Report.
	Report report = new Report("done - " + type);
	report.setReportType(type);
	report.setSenderId(Name.valueOf(IXAgent.getAgent().getAgentIPCName()));
	report.setRef(about.getRef());

	// Send.
	IPC.sendObject(about.getSenderId() // whoever sent this item to us
		            .toString(),   // /\/ 'cause it's a Name
		       report);
    }

    void expandOneLevel(Refinement refinement, MatchEnv env) {
	throw new Error("Attempt to expandOneLevel " + refinement);
    }

    void addItemListener(AgendaItemListener listener) {
	// /\/: What is the Java convention on adding if already there?
	if (!listeners.contains(listener))
	    listeners.add(listener);
    }

    void fireStatusChanged() {
	AgendaItemEvent event = new AgendaItemEvent(this);
	// This method may be called before listeners has been
	// initialized because PNode init calls computeStatus.
	if (listeners != null) {
	    for (Iterator i = listeners.iterator(); i.hasNext();) {
		AgendaItemListener listener = (AgendaItemListener)i.next();
		listener.statusChanged(event);
	    }
	}
    }

    void firePriorityChanged() {
	AgendaItemEvent event = new AgendaItemEvent(this);
	for (Iterator i = listeners.iterator(); i.hasNext();) {
	    AgendaItemListener listener = (AgendaItemListener)i.next();
	    listener.priorityChanged(event);
	}
    }

    void fireHandlerActionsChanged() {
	AgendaItemEvent event = new AgendaItemEvent(this);
	for (Iterator i = listeners.iterator(); i.hasNext();) {
	    AgendaItemListener listener = (AgendaItemListener)i.next();
	    listener.handlerActionsChanged(event);
	}
    }

    void fireNewHandlerAction(HandlerAction act) {
	AgendaItemEvent event = new AgendaItemEvent(this);
	for (Iterator i = listeners.iterator(); i.hasNext();) {
	    AgendaItemListener listener = (AgendaItemListener)i.next();
	    listener.newHandlerAction(event, act);
	}
    }

    void fireNewReport(Report report) {
	AgendaItemEvent event = new AgendaItemEvent(this);
	for (Iterator i = listeners.iterator(); i.hasNext();) {
	    AgendaItemListener listener = (AgendaItemListener)i.next();
	    listener.newReport(event, report);
	}
    }

    void fireAgendaItemEdited() {
	AgendaItemEvent event = new AgendaItemEvent(this);
	for (Iterator i = listeners.iterator(); i.hasNext();) {
	    AgendaItemListener listener = (AgendaItemListener)i.next();
	    listener.agendaItemEdited(event);
	}
    }

    public String toString() {
	return "AgendaItem[about " + about + "]";
    }

}

// Issues:
// * Should reports be recorded in the "about" object instead?
