/****************************************************************************
 * A table viewer for an agenda (set of agenda items).
 * 
 * @author Jussi Stader
 * @version 2.0
 * Copyright: (c) 2002, AIAI, University of Edinburgh
 */

package ix.ip2;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import ix.ip2.event.*;

import ix.icore.*;
import ix.icore.domain.Refinement;

import ix.iface.util.*;
import ix.iface.ui.*;
import ix.iface.ui.table.*;

import ix.iview.util.*;
import ix.iview.table.*;

import ix.util.*;
import ix.util.lisp.*;

/**
 * A table viewer for an agenda (set of agenda items).
 * The items are managed by an AgendaManager (usually an Agenda).  
 */
abstract public class AgendaTableViewer extends IXTreeTable
  implements AgendaViewer, AgendaListener
{

  protected Ip2 ip2;

  AgendaManager agendaManager;
  ItemEditor itemEditor;

  AgendaItemTableModel model;

  private static final ImageIcon editableIcon = 
     Util.resourceImageIcon("ix-symbol-editable.gif");

  /** 
   * Says which action to use when an item has more than one that
   * have the same description.  The two keys are the item and
   * the description.
   */
  private TwoKeyHashMap actionShadowingTable = new TwoKeyHashMap();


  /**
   * Constructs a viewer for the indicated agent.
   */
  public AgendaTableViewer(Ip2 ip2) {
    super();
    setupAgendaTable();      
    this.ip2 = ip2;
    addMouseListener(new AgendaTableMouser());
  }

  private void setupAgendaTable() {
    model = new AgendaItemTableModel(this);
    setModel(model);
    setAgendaRenderersNEditors();
    setPreferredScrollableViewportSize(new Dimension(500, 90));
    //setCellSelectionEnabled(true);
  }
      

  private void setAgendaRenderersNEditors() {
    //Priorities
    //EditablePriorityRenderer epr = new EditablePriorityRenderer(false);
    //epr.setToolTipText("Click for choices");
    PriorityRenderer pr = new PriorityRenderer(false);
    pr.setToolTipText("Click for choices");
    setDefaultRenderer(Priority.class, pr);
    setDefaultRenderer(EditablePriority.class, pr);
    IXComboBox comboBox = new ix.iview.util.PriorityComboBox();
    setDefaultEditor(Priority.class, new IXDefaultCellEditor(comboBox));
    setDefaultEditor(EditablePriority.class, 
		     new IXDefaultCellEditor(comboBox));
    //Actions
    pr.setToolTipText("Click for choices");
    setDefaultRenderer(Priority.class, pr);

    //HandlerActionEditor sets its own renderer
    DefaultCellEditor ae = new HandlerActionEditor(this, new IXComboBox());
    setDefaultEditor(FourField.class, ae);
  }


  public void setItemEditor(ItemEditor itemEditor) {
    this.itemEditor = itemEditor;
  }
  public void ensureItemEditor() {
    if (itemEditor == null) itemEditor = makeItemEditor();
    itemEditor.setVisible(true);
  }

  abstract protected ItemEditor makeItemEditor();

  /**
   * Sets the viewer back to something approximating its initial state.
   */
  public synchronized void reset() {
    clearTable();
    invalidate();
  }

  /**
   * Clears what's displayed without resetting anything else.
   * NOTE: the model clears its rows and nodes and notifies listeners.
   */
  public synchronized void clearTable() {
    model.clearData();
  }

  /**
   * Clears the table then puts everything back, taking account
   * of any changes in open/closed status.
   */
  //***cannot restore after a clear like that!
  /*
  public synchronized void redisplay() {
    clearTable();
    restoreTable();
    invalidate();
    SwingUtilities.getRoot(this).validate();
  }
  */

  /**
   * Undoes a <code>clearTable()</code>, taking into acount any
   * changes in open/closed status.
   */
  /*
  public void restoreTable() {
    Debug.noteln("Restoring item viewing table***");
  }
  */

  public HandlerAction findHandlerAction(AgendaItem item,
					 String shortDescription) {
    HandlerAction shadowingAction
      = (HandlerAction)actionShadowingTable.get(item, shortDescription);
    return (shadowingAction != null)
      ? shadowingAction
      : item.findAction(shortDescription);
  }

  protected void setShadowingAction(AgendaItem item,
				    String shortDescription,
				    HandlerAction action) {
    actionShadowingTable.put(item, shortDescription, action);
  }


  public void actionSelected(AgendaItem item, String actionName) {
    HandlerAction action = findHandlerAction(item, actionName);
    Debug.assert(action != null, "Can't find action", actionName);
    HandlerAction savedAction = model.getHandlerAction(item);
    if (action.equals(savedAction) || 
	(((savedAction == null) || 
	  HandlerAction.NoAction.class.isInstance(savedAction))
	 && HandlerAction.NoAction.class.isInstance(action)) ||
	((savedAction != null) &&
	 actionName.equals(savedAction.getActionDescription())))
      return;

    // The user can mess around with the actions while the status is BLANK
    // but can't get anything to happen unless the status is POSSIBLE.
    // /\/: This now depends on the item and action classes.
    if (item.actionCanBeTakenNow(action)) {
      // Test was: (item.status == Status.POSSIBLE)
      handleItem(item, action);
    }
    // Put selection back to what it was before
    else {
      Debug.noteln("Ignoring action selection");
      JOptionPane
	.showMessageDialog(this, 
			   "The selected action cannot be taken now",
			   "Action selection",
			   JOptionPane.INFORMATION_MESSAGE);
      model.setHandlerAction(item, savedAction);
    }
  }

  public void handleItem(AgendaItem item, HandlerAction action) {
    HandlerAction savedAction = model.getHandlerAction(item);
    model.setHandlerAction(item, action);
    try { agendaManager.handleItem(item, action); }
    catch (Exception e) {
      Debug.noteln("Exception while handling item", item);
      Debug.noteException(e);
      JOptionPane
	.showMessageDialog(this, new Object[] 
			   {"Problem while handling item", e.getMessage()},
			   "Problem while handling item",
			   JOptionPane.ERROR_MESSAGE);
      model.setHandlerAction(item, savedAction);
    }
  }


  public void newHandlerAction(AgendaItem item, HandlerAction act) {
    //Ignore the new action if the item staus indicates that it cant be used
    Status status = item.getStatus();
    if (status == Status.COMPLETE 
	|| status == Status.EXECUTING
	|| status == Status.IMPOSSIBLE) 
      return;
    
    // Make sure the user can see which item we're talking about.
    int savedSelected = getSelectedRow();
    
    int row = model.getObjectRow(item);
    try { setRowSelectionInterval(row, row);}
    catch (Exception e) {}
    
    // do we already know about an action with the same description?
    String target = act.getActionDescription();
    String exists = null;
    java.util.List handlerActions = item.getActions();
    for (Iterator i = handlerActions.iterator(); i.hasNext(); ) {
      HandlerAction ha = (HandlerAction)i.next();
      if (!ha.equals(act) && (ha.getActionDescription().equals(target))) {
	exists = ha.getActionDescription();
	break;
      }
    }
    
    if (exists != null) {
      //There is already an action with the same description.
      //See if the user wants the new action or the old one. We assume that
      // if we do nothing we will continue to find the same action as before.
      if (shouldReplaceAction(item, exists)) {
	// The user wants the new action, so shadow the old.
	setShadowingAction(item, act.getActionDescription(), act);
	Debug.assert(findHandlerAction(item, exists) == act);
      }
      else {
	// Check that lookup doesn't get the new action.
	Debug.assert(findHandlerAction(item, exists) != act);
      }
    }
    else {
      //We have a new action that does not match an existing description,
      // so add it to the list of possible ones
      //Do not mention it if it is a "no action" or a "not appliccable"
      //***This should really not mention things that are just created, only 
      //***ones that are added when things are on the agenda already
      String descript = act.getActionDescription();
      Debug.noteln("Item", Util.quote(item.getShortDescription()));
      Debug.noteln(" has new action", descript);
      String noapp = 
	(new HandlerAction.NotApplicable()).getActionDescription();
      String noact = (new HandlerAction.NoAction()).getActionDescription();
      String done = (new HandlerAction.Manual()).getActionDescription();
      if (!descript.equals(noapp) &&
	  !descript.equals(noact) &&
	  !descript.equals(done) &&
	  !descript.equals("Expand as below"))
	JOptionPane.showMessageDialog(this, new Object[] {
	  "New action " + Util.quote(act.getActionDescription()),
	    "For item " + Util.quote(item.getShortDescription())
	    },
	  "New action",
	  JOptionPane.INFORMATION_MESSAGE);
      //handlerActions.add(act);
    }
    
    // Restore selected item to before highlighting this one
    if (savedSelected >= 0)
      setRowSelectionInterval(savedSelected, savedSelected);
    else clearSelection();
    
    //no need to fire change - the possible actions are not shown in table
  }

  boolean shouldReplaceAction(AgendaItem item, String actDescription) {
    Object[] answers = {"Replace existing action","Ignore new action"};
    Object selected = JOptionPane.showInputDialog(this,	new Object[] {
      "Item " + Util.quote(item.getShortDescription()),
	"Already has an action " + 
	Util.quote(actDescription)},
      "New Action",			// title
      JOptionPane.INFORMATION_MESSAGE,
      null,				// icon
      answers,
      answers[0]);
    if (selected == null) return false;
    else return selected.equals("Replace existing action");
  }



  /*
   * Instructions from the frame menu bar
   */
  
  public void getNewItemFromUser() {
    ensureItemEditor();
    itemEditor.showNewItem();
  }




  //------------------- AgendaViewer implementation -------------------------

  public AgendaItem makeItem(LList pattern) {
    if (agendaManager == null) {
      Debug.error("AgendaManager not set!");
      return null;
    }
    else return agendaManager.makeItem(pattern);
  }

  public void addItem(AgendaItem i) {
    agendaManager.addItem(i);
  }

  public void saveExpansion(Refinement data) {
    ip2.frame.getDomainEditor().saveExpansion(data);
  }

  public void expandItem(AgendaItem i, Refinement instructions) {
    agendaManager.expandItem(i, instructions);
  }


  /**
   * Useful methods for adding test items.  They assume that the
   * text can be parsed w/o error.  Note that the ItemEditor does
   * not call these methods -- they are called only "internally",
   * specifically for test items.
   */
  public void addItem(String text) {
    addItem(agendaManager.makeItem(text));
  }

  public void addItem(Priority priority, String text) {
	// Assumes that the text can be parsed w/o error.
	// Note that the ItemEditor does not call this method --
	// it's called only "internally", for test items.
	AgendaItem i = agendaManager.makeItem(text);
	i.setPriority(priority);
	addItem(i);
  }


  //------------------- AgendaListener implementation ---------------------


  public void setAgendaManager(AgendaManager agendaManager) {
    this.agendaManager = agendaManager;
  }

  public void itemAdded(AgendaEvent event, AgendaItem i) {
    itemAdded(i);
  }
  /**
   * Called when an itemhas been added, adds the item to the table.
   * An AgendaViewer thing - also used by AgendaListener method.
   */
  public synchronized void itemAdded(AgendaItem i) {
    //Debug.noteln("****AgendaViewer adding item", i);
    model.itemAdded(i);
  }

  public void itemRemoved(AgendaEvent event, AgendaItem i) {
    // /\/: What if the item has children?
    //Debug.noteln("Viewer removing item", i);
    model.itemRemoved(i);
  }

  public void itemHandled(AgendaEvent ae, AgendaItem item, HandlerAction act) {
    // Called when an item has been handled in some way other than
    // the user selecting an action from the menu - in particular when
    // the user has specified a manual expansion.  The JComboBox's
    // listener will be told when we set the selected item and may
    // need to know this is not the usual case.  At present, we let
    // it know by disabling the combobox.  /\/
    model.itemHandled(ae, item, act);
  }

  public void newBindings(AgendaEvent event, Map bindings) {
    //This works but is a bit too sweeping - perhaps better to go through
    // the rows and only change ones that contain variables in question
    //Debug.noteln("Viewer binding variables");
    model.fireTableChanged();
  }


  //------------------- Actions -----------------------------------------------

  public void editingCanceled(ChangeEvent e) {
    super.editingCanceled(e);
    model.fireTableChanged();
  }
  public void editingStopped(ChangeEvent e) {
    super.editingStopped(e);
    model.fireTableChanged();
  }



  private class AgendaTableMouser extends MouseAdapter {
    /**
     * If the left button was clicked on the tree-column, expand/collapse node
     * as per model.
     * If the left button was clicked in a non-editable column, bring up 
     * item editor.
     * If the right button was clicked, bring up popup menu.
     */
    public void mouseClicked(MouseEvent me) {
      //these clicks do not need the row to be selected - work on where clicked
      int row = rowAtPoint(me.getPoint());
      int column = columnAtPoint(me.getPoint());
      AgendaItem item = (AgendaItem)model.getRowObject(row);
      //Debug.noteln("AgendaTableViewer has mouse event", me.paramString());
      //Debug.noteln(" row " + row + " column ", column);
      if (item == null) {
	Debug.noteln("Clicked on empty item");
	return;
      }
      else if (SwingUtilities.isRightMouseButton(me)) {
	ItemPopupMenu popup = new ItemPopupMenu(item);
	popup.noticeItemState();
	popup.show(me.getComponent(), me.getX(), me.getY());
      }
      else if (SwingUtilities.isLeftMouseButton(me) && 
	       ((column == 0) || 
		(column == model.PRIORITY_COL) || 
		(column == model.ACTION_COL))) {
	model.cellClicked(me, row, column);
      }
      /*
      else if (SwingUtilities.isLeftMouseButton(me)) {
	ensureItemEditor();
	itemEditor.showItem(item);
      }
      */
    }
  }

  private class ItemPopupMenu extends JPopupMenu implements ActionListener {
    AgendaItem item;
    
    JMenuItem foldItem = makeMenuItem("Fold");
    JMenuItem bindItem = makeMenuItem("Bind Variables");

    ItemPopupMenu(AgendaItem item) {
      super();
      this.item = item;
      
      add(makeMenuItem("Show Details"));
      add(foldItem);
      foldItem.setEnabled(false);
      
      add(bindItem);
      bindItem.setEnabled(false);
      // Other items might be added depending on the item's state.
      
      if (item.getAbout().getReportBack() == YesNo.YES)
	add(makeMenuItem("Send report"));
    }

    JMenuItem makeMenuItem(String text) {
      JMenuItem item = new JMenuItem(text);
      item.addActionListener(this);
      return item;
    }

    void noticeItemState() {
      IXTreeTableNode node = (IXTreeTableNode)model.getTreeNode(item);
      if (!node.isLeaf())
	foldItem.setEnabled(true);
      if (node.expanded) foldItem.setText("Fold");
      else foldItem.setText("Unfold");
      if (item.getPatternVars().isEmpty()) bindItem.setEnabled(false);
      else bindItem.setEnabled(true);
    }

    public void actionPerformed(ActionEvent e) {
      String command = e.getActionCommand();
      Debug.noteln("Item popup command", command);
      if (command.equals("Show Details")) {
	ensureItemEditor();
	itemEditor.showItem(item);
      }
      else if (command.equals("Fold")) {
	IXTreeTableNode node = (IXTreeTableNode)model.getTreeNode(item);
	int row = model.getObjectRow(item);
	model.unexpandNode(node, row);
	foldItem.setText("Unfold");
      }
      else if (command.equals("Unfold")) {
	IXTreeTableNode node = (IXTreeTableNode)model.getTreeNode(item);
	int row = model.getObjectRow(item);
	model.expandNode(node, row);
	foldItem.setText("Fold");
      }
      else if (command.equals("Bind Variables")) {
	Set unboundVars = item.getUnboundVars();
	Debug.assert(!unboundVars.isEmpty(), "No variables to bind.");

	// /\/: Trick to keep the compiler from saying
	// "variable binder might not have been initialized".
	// Should work instead by BindingViewer having a
	// method we can override in a local or anon class.
	final BindingViewer[] binder = new BindingViewer[1];
	binder[0] = new BindingViewer
	  (ip2, unboundVars,
	   new Runnable() { 
	    public void run() {
	      if (binder[0].newBindings != null) { // next bit is messy /\/
		ip2.controller.modelManager.newBindings(binder[0].newBindings);
	      }
	    }
	  });
      }
      else if (command.equals("Send report")) {
	ip2.frame.ensureChat();
	ip2.frame.chatFrame.initReport(item.getAbout());
      }
      else {
	Debug.noteln("Nothing to do for", command);
      }
    }

  }
}

// Issues:
// * What should be done when removing an item that has children?
//   Don't allow it?  Remove descendents as well?
