/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Sun May  5 02:09:36 2002 by Jeff Dalton
 * Copyright: (c) 2002, AIAI, University of Edinburgh
 */

package ix.ip2;

import java.util.*;

import ix.util.*;
import ix.icore.*;

/**
 * One of the ways in which an agenda item might be handled.
 */
public class HandlerAction {

    /**
     * Briefly describes what this action does.
     */
    protected String shortDescription;

    /**
     * Constructor used by subclasses.
     */
    HandlerAction() { }

    /**
     * Handle the item in the manner appropriate to this handler-action.
     * This method merely changes the item's status to COMPLETE and
     * is overridden in subclasses that handle the item in more
     * interesting ways.
     */
    public void handle(AgendaItem item) {
	Debug.noteln("HandlerAction " + shortDescription +
		     " handling item " + item);
	item.setStatus(Status.COMPLETE);
    }

    /**
     * Called when something has happened that might mean this action
     * no longer makes sense and should be deleted.  This method just
     * returns true and is overridden in subclasses that have instances
     * that may cease to be valid.
     */
    public boolean isStillValid() {
	return true;
    }

    /**
     * Indicates whether this action is ready to be used or else
     * is waiting for something that would enable it to be used.
     * This method just returns true and is overridden in subclasses
     * that have instances that are not always ready.
     */
    public boolean isReady() {
	return true;
    }

    /**
     * Called when something has happened that might have changed
     * this action's ready status.  This method does nothing and
     * is overridden in subclasses that have instances that are not
     * always ready.
     */
    public void computeStatus() {
    }

    /**
     * Returns a brief description of how this action would handle
     * an item.
     */
    public String getActionDescription() {
	return shortDescription;
    }

    public String toString() {
	return "HandlerAction[" + shortDescription + "]";
    }


    /*
     * Some simple subclasses
     */

    /**
     * Handles an item by doing nothing.  Used when it's
     * convenient to select an action without actually handling
     * the item, so that all other actions remain available
     * and can be selected later on.
     */
    static class NoAction extends HandlerAction {
	public NoAction() {
	    shortDescription = "No Action";
	}
	public void handle(AgendaItem item) {
	    // N.B. no status change.
	    Debug.noteln("No action for", item);
	}
    }

    /**
     * Handles an issue by invoking one of the agent-level
     * issue-handlers.
     */
    public static class Automatic extends HandlerAction {

	ItemHandler handler;

	public Automatic(ItemHandler h) {
	    this(h, /* "Do it using " + */ h.getActionDescription());
	}

	public Automatic(ItemHandler h, String descr) {
	     this.handler = h;
	     this.shortDescription = descr;
	}

	public void handle(AgendaItem item) {
	    Debug.noteln("Giving issue to handler", handler);
	    handler.handle(item);
	    // N.B. It is up to that handler to do:
	    //   i.setStatus(STATUS_COMPLETE);
	    // or whatever other status changes are required.
	}

    }

    /**
     * Indicates that the item has been handled by the user.
     * Note that an item may have more than one Manual action,
     * with different descriptions that act as reminders of
     * different ways the user might act.
     */
    static class Manual extends HandlerAction {
	public Manual() {
	    shortDescription = "Done";
	}
	public Manual(String howDescription) {
	    shortDescription = "Done using " + howDescription;
	}
    }

    /**
     * Used to indicate the the item does not apply in the current
     * situation.
     */
    static class NotApplicable extends Manual {
	public NotApplicable() {
	    shortDescription = "N/A";
	}
    }

}

// Issues:
// * The approach taken by isStillValid() and computeStatus() is
//   obviously less than ideal, because something outside the object
//   has to know when to call the method; but it will do for now.
