/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Thu Jun 13 18:25:05 2002 by Jeff Dalton
 * Copyright: (c) 2001 - 2002, AIAI, University of Edinburgh
 */

package ix.ip2;

import javax.swing.*;

import java.util.*;

import java.io.File;

import javax.swing.JPanel;

import ix.ip2.event.AgendaListener;

import ix.iface.domain.DomainParser;
import ix.iface.util.LogoPanel;

import ix.ispace.*;

import ix.icore.domain.*;
import ix.icore.domain.event.*;
import ix.icore.*;

import ix.icore.domain.Constraint;
import ix.ichat.ChatMessage;

import ix.util.*;
import ix.util.lisp.*;
import ix.util.xml.*;


/** 
 * The generic I-P2 class and application main program 
 */

public class Ip2 extends IXAgent {

    protected Domain domain = new Domain();

    protected PanelController controller = new PanelController(this);

    protected Ip2Frame frame;

    boolean classic;		// use "old" viewers?

    AgendaViewer activityViewer;
    AgendaViewer issueViewer;
    StateViewer stateViewer;

    // Some customizable fields
    protected String logoLine1 = "I-X Process Panel";
    protected String logoLine2 = "Based on I-X Technology";
    protected String logoImage = "ip2-logo.gif";

    protected String domainName = null;

    public Ip2() {
	super();

	// Change some defaults
	displayName = "Process Panel";
    }

    /**
     * Main program.
     */
    public static void main(String[] argv) {

	Util.printGreeting("I-P2");

	new Ip2().mainStartup(argv);

    }

    /**
     * Command-line argument processing for arguments used by all
     * versions of I-P2.
     */
    protected void processCommandLineArguments() {
	super.processCommandLineArguments();

	logoLine1   = Parameters.getParameter("logo-line-1", logoLine1);
	logoLine2   = Parameters.getParameter("logo-line-2", logoLine2);
	logoImage   = Parameters.getParameter("logo-image", logoImage);

	classic = Parameters.getBoolean("classic", true);

	if (Parameters.haveParameter("domain"))
	    domainName = Parameters.getParameter("domain");

	if (Parameters.haveParameter("test-menu"))
	    // Avoids warning about unused parameter. /\/
	    Debug.noteln("Test menu file",
			 Parameters.getParameter("test-menu"));
    }

    /**
     * Completes basic I-P2 setup and initialization.
     */
    public void startup() {
	// Viewers
	activityViewer = makeActivityViewer();
	issueViewer = makeIssueViewer();
	stateViewer = makeStateViewer();

	// Controller and Agendas

	controller.addActivityListener((AgendaListener)activityViewer);
	controller.addIssueListener((AgendaListener)issueViewer);

  	addIssueHandlers();

  	controller.setDomain(domain);
	readDomain(domain);

	// State viewer
	// /\/: Should this go via the controller?
	controller.modelManager
	    .addProcessStatusListener(stateViewer);

	// Frame
	frame = classic
	    ? new Ip2Frame(this)
	    : new NewIp2Frame(this);
	frame.setUp();
	frame.becomeVisible();
    }

    /**
     * Called to create the activity viewer.  This method can be
     * redefined in subclass that want to instantiate a different
     * viewer class.
     */
    protected AgendaViewer makeActivityViewer() {
	if (classic)
	    return new ActivityViewTable(this);
	else
	    return new ActivityTableViewer(this);
    }

    /**
     * Called to create the issue viewer.  This method can be
     * redefined in subclass that want to instantiate a different
     * viewer class.
     */
    protected AgendaViewer makeIssueViewer() {
	if (classic)
	    return new IssueViewTable(this);
	else
	    return new IssueTableViewer(this);
    }

    /**
     * Called to create the state viewer.  This method can be
     * redefined in subclass that want to instantiate a different
     * viewer class.
     */
    protected StateViewer makeStateViewer() {
	return new StateViewTable(this);
    }

    /**
     * Read in any default domain description.
     */
    protected void readDomain(Domain domain) {
	if (domainName != null) {
	    try {
		DomainParser.makeParser(domainName).readDomain(domain);
	    }
	    catch (Exception e) {
		Debug.noteException(e);
	    }
	}
    }

    /**
     * Makes the LogoPanel for the application's main frame.
     * This method is in this class to make it easier to define
     * versions that have a different logo panel.
     */
    protected JPanel makeLogoPanel() {
	ImageIcon icon = Parameters.haveParameter("logo-image")
	    ? new ImageIcon(logoImage)
	    : Util.resourceImageIcon(logoImage);
	return new LogoPanel(symbolName, logoLine1, logoLine2, icon);
    }

    /**
     * Called when the main frame is set up to add items to the frame
     * menu bar's "Test" menu.  This method is in this class, rather than
     * in the frame class, to make it easier to define versions that have
     * different test items.
     */
    protected void addTestMenuItems() {
	if (Parameters.haveParameter("test-menu"))
	    addTestMenuItems(Parameters.getParameter("test-menu"));
    }

    protected void addTestMenuItems(String filename) {
	try {
	    List items = (List)XML.objectFromFile(new File(filename));
	    if (items != null) {
		for (Iterator i = items.iterator(); i.hasNext();) {
		    TestItem item = (TestItem)i.next();
		    frame.addTest(item);
		}
	    }
	}
	catch (Exception e) {
	    Debug.noteException(e);
	    Util.displayMessage(null,
	        Strings.foldLongLine
		  ("Problem parsing " + Strings.quote(filename) + ": " + e));
	}
    }

    /**
     * Called to restore the initial state.
     */
    public void reset() {
	// /\/: Things should register for reset instead of being known here.
	controller.reset();
	activityViewer.reset();
	issueViewer.reset();
	stateViewer.reset();
	frame.validate();
    }

    /**
     * Handles new issues from external sources.
     */
    public void handleNewIssue(Issue issue) {
	controller.addIssue(issue);
    }

    /**
     * Handles new activities from external sources.
     */
    public void handleNewActivity(Activity activity) {
	controller.addActivity(activity);
    }

    /**
     * Handles new constraints from external sources.
     */
    public void handleNewConstraint(Constraint constraint) {
	// /\/: Should this go via the controller?
	// /\/: Or should the model manager be known at this level?
	controller.modelManager.addConstraint(constraint);
    }

    /**
     * Handles new reports from external  sources.
     */
    public void handleNewReport(Report report) {
	controller.newReport(report);
    }

    /**
     * Handles new chat messages.
     */
    public void handleNewChatMessage(ChatMessage message) {
	frame.ensureChat();
	frame.chatFrame.newMessage(message);
    }

    /**
     * Install any built-in issue and activity handlers.
     */ 
    protected void addIssueHandlers() {

	// Handler that lets issues be converted to activities
	controller
	    .addIssueHandler(new IssueToActivityHandler());

	// Handler that provides forwarding to agents that have
	// required capabilities
	controller
	    .addItemHandler(new InvokeHandler(this));

	// Escalate to superior, do not require report back
	addForwardingHandler
	    ("Escalate", AgentRelationship.SUPERIOR, false);

	// Pass to peer, with and without report back
	addForwardingHandler
	    ("Pass with report-back", AgentRelationship.PEER, true);
	addForwardingHandler
	    ("Pass without report-back", AgentRelationship.PEER, false);

	// Delegate to subordinate, require report back
	addForwardingHandler
	    ("Delegate", AgentRelationship.SUBORDINATE, true);

	// Connecting an agent
	controller
	    .addItemHandler(new ConnectHandler());

	// Querying an agent
//  	controller
//  	    .addActivityHandler(new QueryHandler());

    }

    public void addForwardingHandler(String verb,
				     AgentRelationship rel,
				     boolean reportBack) {
//  	List data = contactManager.getAgentData(rel);
//  	if (data != null) {
//  	    List names = (List)Collect
//  		.map(data, Fn.accessor(AgentData.class, "getName"));
//  	    controller.addItemHandler
//  		(new ForwardingHandler(verb, names, reportBack));
//  	}
	controller.addItemHandler
	    (new ForwardingHandler(this, verb, rel, reportBack));
    }


    /*
     * Some useful handler classes
     */

    public class IssueToActivityHandler extends IssueHandler {
	public IssueToActivityHandler() {
	    super("Perform as activity");
	}
	public void handle(AgendaItem item) {
	    Issue issue = (Issue)item.getAbout();
	    Activity activity = new Activity(issue.getPattern());
	    // Take across some other fields ...
	    activity.setSenderId(issue.getSenderId());
	    activity.setPriority(issue.getPriority());
	    activity.setReportBack(issue.getReportBack());
	    activity.setRef(issue.getRef());
	    // ... Any others ...?  /\/
	    // Remove the issue item
	    controller.getIssueAgenda().removeItem(item);
	    // Add the activity
	    controller.addActivity(activity);
	}
    }

    public class ConnectHandler extends ActivityHandler {
	public ConnectHandler() {
	    super("Connect agent");
	}
	public boolean appliesTo(AgendaItem item) {
	    LList pattern = item.getPattern();
	    return pattern.length() == 3
		&& pattern.get(0) == S_CONNECT;
	}
	public void handle(AgendaItem item) {
	    // (connect ?agentName ?Relationship)
	    LList pattern = item.getPattern();
	    String name = pattern.get(1).toString();
	    String relName = pattern.get(2).toString();
	    AgentRelationship rel = AgentRelationship.valueOf(relName);
	    AgentData data = contactManager.getAgentData(name);
	    if (data == null)
		contactManager.addAgent(name, rel);
	    else if (data.getRelationship() != rel)
		contactManager.changeRelationship(name, rel);
	    item.setStatus(Status.COMPLETE);	    			  
	}
	Symbol S_CONNECT = Symbol.intern("connect");
    }

}

