/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Tue May 14 20:07:29 2002 by Jeff Dalton
 * Copyright: (c) 2001 - 2002, AIAI, University of Edinburgh
 */

package ix.ip2;

import javax.swing.*;		// but shouldn't act as part of UI /\/

import java.util.*;

import ix.icore.*;
import ix.icore.process.*;
import ix.icore.domain.*;
import ix.util.*;
import ix.util.lisp.*;
import ix.util.match.*;

public class Ip2ModelManager extends ProcessModelManager {

    public static final Symbol
	S_WORLD_STATE = Symbol.intern("world-state"),
	S_CONDITION   = Symbol.intern("condition"),
	S_EFFECT      = Symbol.intern("effect");

    protected MultiMap nodeToConditions = new MultiHashMap();
    protected MultiMap nodeToEffects = new MultiHashMap();

    protected WorldState worldState = new WorldState();

    public Ip2ModelManager() {
	super();
    }

    public void reset() {
	super.reset();
	nodeToConditions.clear();
	nodeToEffects.clear();
	worldState.clear();
    }

    // /\/: Needs a different name (addBindings?) so it won't be confused
    // with the listener method; or maybe the listener method should be called
    // "newBindingsAdded"?
    public void newBindings(Map bindings) {
	for (Iterator i = bindings.entrySet().iterator(); i.hasNext();) {
	    Map.Entry e = (Map.Entry)i.next();
	    Variable var = (Variable)e.getKey();
	    Object val = e.getValue();
	    var.setValue(val);
	}
	fireNewBindings(bindings);
	runSentinels();
    }

    public void addConstraint(Constraint c) {
	Debug.assert(c.getType() == S_WORLD_STATE);
	Symbol relation = c.getRelation();
	if (relation == S_EFFECT) {
	    PatternAssignment pv = (PatternAssignment)c.getParameter(0);
	    handleEffects(Lisp.list(pv));
	}
	else
	    throw new RuntimeException ("Can't handle " + c);
    }

    protected void addConstraint(PNode node, Constraint c) {
	Debug.assert(c.getType() == S_WORLD_STATE);
	Symbol relation = c.getRelation();
	PatternAssignment pv = (PatternAssignment)c.getParameter(0);
	if (relation == S_CONDITION) {
	    nodeToConditions.addValue(node, pv);
	}
	else if (relation == S_EFFECT) {
	    nodeToEffects.addValue(node, pv);
	}
	else
	    throw new RuntimeException ("Can't handle " + c + " at " + node);
    }

    public MatchEnv evalFilters(List conds, MatchEnv env) {
	Debug.assert(!conds.isEmpty());
	LList cl = LList.newLList(conds);
	return filter((PatternAssignment)cl.car(), cl.cdr(), env);
    }

    protected MatchEnv filter(PatternAssignment pv, LList conds,
			      MatchEnv baseEnv) {
	Debug.noteln("PMM filter", pv);
	MatchEnv result = filter1(pv, conds, baseEnv);
	Debug.noteln("result:", result);
	return result;
    }

    private MatchEnv filter1(PatternAssignment pv, LList conds,
			     MatchEnv baseEnv) {
	for (Iterator w = worldState.entrySet().iterator(); w.hasNext();) {
	    Map.Entry entry = (Map.Entry)w.next();
	    MatchEnv env = matchFilter(pv, entry, baseEnv);
	    if (env == null)
		continue;
	    if (conds.isEmpty())
		return env;
	    env = filter((PatternAssignment)conds.car(), conds.cdr(), env);
	    if (env != null)
		return env;
	}
	return null;
    }

    protected MatchEnv matchFilter(PatternAssignment pv,
				   Map.Entry entry,
				   MatchEnv baseEnv) {
	Debug.noteln("Matching " + pv + ", " + entry + ", " + baseEnv);
	MatchEnv env = new MatchEnv(baseEnv); // copy
	env = Matcher.match(pv.getPattern(), entry.getKey(), env);
	if (env == null)
	    return null;
	env = Matcher.match(pv.getValue(), entry.getValue(), env);
	return env;
    }

    public void statusChanged(PNode node) {
	Debug.noteln("PMM sees new status of", node);
	if (node.getStatus() == Status.COMPLETE) {
	    handleCompletion(node);
	}
    }

    protected void handleCompletion(PNode node) {
	List effects = (List)nodeToEffects.get(node);
	if (effects == null)
	    return;
	Set vars = getVarsFromConstraints(effects);
	Set unbound = Variable.unboundVarsIn(vars);
	if (!unbound.isEmpty()) {
	    Debug.noteln("Unbound vars on completion of", node);
	    Debug.noteln("Vars are ", unbound);
	    JOptionPane.showMessageDialog(null,
	        new Object[] {"Unbound vars on completion of " + node,
			      "Vars are: " + unbound},
		"Warning",
		JOptionPane.WARNING_MESSAGE);
	    node.setStatus(Status.IMPOSSIBLE);
	    addSentinel(new BindingSentinel(node, unbound));
	}
	else
	    handleEffects(node, effects);
    }

    protected void handleEffects(PNode node, List effects) {
	handleEffects(effects);
    }

    protected void handleEffects(List effects) {
	// N.B. All Variables in the effects must be bound.
	// /\/: Takes a list of PatternAssignments, NOT a list of Constraints.
	Map delta = new HashMap();
	for (Iterator i = effects.iterator(); i.hasNext();) {
	    PatternAssignment pv = (PatternAssignment)i.next();
	    LList pattern = pv.getPattern();
	    Object value = pv.getValue();
	    Debug.assert(Variable.isFullyBound(pattern));
	    Debug.assert(Variable.isFullyBound(value));
	    worldState.put(pattern, value);
	    delta.put(pattern, value);
	}
	fireStateChange(delta);
	runSentinels();
    }

    protected Set getVarsFromConstraints(List pv_pairs) {
	// /\/: Takes a list of PatternAssignments, NOT a list of Constraints.
	Set result = new HashSet();
	for (Iterator i = pv_pairs.iterator(); i.hasNext();) {
	    PatternAssignment pv = (PatternAssignment)i.next();
	    result.addAll(pv.getVariables());
	}
	return result;
    }

    /**
     * A sentinel that lets an activity become complete when all the
     * variables in its effects have values.
     */
    protected class BindingSentinel implements Sentinel {

	PNode node;
	Set unbound;

	BindingSentinel(PNode node, Set unbound) {
	    this.node = node;
	    this.unbound = new HashSet(unbound); // copy
	}

	public boolean isReady() {
	    Debug.noteln("Testing BindingSentinel for", node);
	    Set stillUnbound = Variable.unboundVarsIn(unbound);
	    return stillUnbound.isEmpty();
	}

	public void run() {
	    Debug.noteln("Able to complete", node);
	    Debug.noteln("Because all variables bound.");
	    node.setStatus(Status.COMPLETE);
	}

    }

}
