/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Thu May 30 01:52:11 2002 by Jeff Dalton
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 */

package ix.ispace;

import java.util.*;

import ix.ispace.event.*;
import ix.util.*;

/**
 * Keeps track of known relationships between agents
 * and agent capabilities.
 */
public class ContactManager {

    protected List listeners = new LinkedList();

    protected List agentData = new LinkedList();

    protected MultiMap relMap = new MultiHashMap();

    protected List groups = new LinkedList(); // remembers order added
    protected Map nameToGroup = new HashMap();

    public ContactManager() {
	addStandardAgentGroups();
    }

    public void addAgent(String name) {
	addAgent(name, AgentRelationship.CONTACT);
    }

    public void addAgent(String name, AgentRelationship rel) {
	Debug.assert(getAgentData(name) == null, "adding agent twice", name);
	AgentData data = new AgentData(name, rel);
	agentData.add(data);
	relMap.addValue(rel, data);
	fireNewContact(data);
    }

    public void addAgents(List names, AgentRelationship rel) {
	for (Iterator i = names.iterator(); i.hasNext();) {
	    addAgent((String)i.next(), rel);
	}
    }

    /**
     * Record the agent as a contact if it is not already known.
     * This makes it easier to reply to agents that have sent things
     * to us, e.g. by putting their name in menus of agents that
     * might be sent to.
     */
    public void noteAgent(String name) {
	noteAgent(name, AgentRelationship.CONTACT);
    }

    public void noteAgent(String name, AgentRelationship assumedRel) {
	AgentData data = getAgentData(name);
	if (data == null)
	    addAgent(name, assumedRel);
    }

    public void changeRelationship(String name, AgentRelationship newRel) {
	AgentData oldData = getAgentData(name);
	AgentRelationship oldRel = oldData.getRelationship();

	Debug.noteln("Changing " + name + " from " + oldRel + " to " + newRel);

	AgentData newData = new AgentData(oldData);
	newData.setRelationship(newRel);

	Collect.replaceFirst(oldData, newData, agentData);
	relMap.removeValue(oldRel, oldData);
	relMap.addValue(newRel, newData);
	fireContactChange(oldData, newData);
    }

    public void deleteAgent(String name) {
	AgentData oldData = getAgentData(name);
	AgentRelationship oldRel = oldData.getRelationship();

	agentData.remove(oldData);
	relMap.removeValue(oldRel, oldData);

	fireContactDeleted(oldData);
    }

    public List getAgentData() {
	return agentData;
    }

    public AgentData getAgentData(String name) {
	for (Iterator i = agentData.iterator(); i.hasNext();) {
	    AgentData data = (AgentData)i.next();
	    if (data.getName().equals(name))
		return data;
	}
	return null;
    }

    public List getAgentData(AgentRelationship rel) {
	return (List)relMap.get(rel);
    }

    public List getAgentData(Capability c) {
	List result = new LinkedList();
	for (Iterator i = agentData.iterator(); i.hasNext();) {
	    AgentData data = (AgentData)i.next();
	    if (data.hasCapability(c))
		result.add(data);
	}
	return result;
    }

    public List getSortedNameList() {
	List names = (List)Collect
	    .map(agentData, Fn.accessor(AgentData.class, "getName"));
	Collections.sort(names);
	return names;
    }


    /*
     * AgentGroups
     */

    public Collection getAgentGroups() {
	// return nameToGroup.values();
	return groups;
    }

    public AgentGroup getAgentGroup(String name) {
	return (AgentGroup)nameToGroup.get(name);
    }

    public void addAgentGroup(AgentGroup group) {
	groups.add(group);
	nameToGroup.put(group.getName(), group);
    }

    protected void addStandardAgentGroups() {

	// All known agents
	addAgentGroup(new AbstractAgentGroup("all", this) {});

	// A group for each relationship
	for (Iterator i = AgentRelationship.values().iterator();
	     i.hasNext();) {
	    AgentRelationship rel = (AgentRelationship)i.next();
	    if (rel == AgentRelationship.NONE) // /\/
		continue;
	    addAgentGroup(new RelationshipGroup(rel, this));
	}

    }


    /**
     * Process command-line arguments that are about our relationships
     * with other agents.
     *
     * <p>For each AgentRelationship, R, R+"s" is treated as a parameter
     * whose value should be a comma-separated list of names.
     *
     * <p>Simple capabilities can be given by
     * <pre>
     *    -external-capabilities=<i>ipcName</i>:<i>verb</i>, ...
     * </pre>
     *
     * @see ix.ispace.AgentRelationship
     */
    public void processCommandLineArguments() {
	// AgentRelationships
	for (Iterator i = AgentRelationship.values().iterator();
	     i.hasNext();) {
	    AgentRelationship rel = (AgentRelationship)i.next();
	    // /\/: Assumes +"s" plurals
	    List names = Parameters.getList(rel + "s");
	    addAgents(names, rel);
	}
	// external-capabilities=ipcName:verb,ipcName:verb,...
	String caps = Parameters.getParameter("external-capabilities", "");
	for (Iterator i = Strings.breakAt(",", caps).iterator();
	     i.hasNext();) {
	    String cap = (String)i.next();
	    String[] parts = Strings.breakAtFirst(":", cap);
	    String ipcName = parts[0];
	    String verb = parts[1];
	    noteAgent(ipcName, AgentRelationship.NONE);
	    AgentData data = getAgentData(ipcName);
	    data.addCapability(new VerbCapability(verb));
	}
    }

    public void addContactListener(ContactListener listener) {
	listeners.add(listener);
    }

    public void fireNewContact(AgentData data) {
	fireContactEvent(new ContactEvent(this, null, data));
    }

    public void fireContactDeleted(AgentData data) {
	fireContactEvent(new ContactEvent(this, data, null));
    }

    public void fireContactChange(AgentData oldData, AgentData newData) {
	fireContactEvent(new ContactEvent(this, oldData, newData));
    }

    public void fireContactEvent(ContactEvent event) {
	Debug.noteln("fireContactEvent", event);
	for (Iterator i = listeners.iterator(); i.hasNext();) {
	    ContactListener listener = (ContactListener)i.next();
	    listener.contactChange(event);
	}
    }

}
