/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Thu May 30 04:56:24 2002 by Jeff Dalton
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 */

package ix.itest;

import java.awt.Color;
import java.awt.GridLayout;
import java.awt.FlowLayout;
import java.awt.Component;
import java.awt.event.*;

import javax.swing.*;

import java.util.*;

import ix.iface.util.RadioButtonBox;
import ix.iface.util.ComboChoice;
import ix.iface.util.GridColumn;
import ix.iface.util.CatchingActionListener;

import ix.util.xml.XML;		// for debugging output

import ix.ispace.AgentData;
import ix.ispace.GroupSender;
import ix.ispace.event.*;
import ix.ichat.*;

import ix.icore.*;
import ix.icore.domain.Constraint;
import ix.icore.domain.PatternAssignment;

import ix.iface.domain.LTF_Parser; // for constraint parsing
import ix.iface.domain.SyntaxException;

import ix.util.*;
import ix.util.lisp.*;


/**
 * I-Test panel for sending messages.
 */
public class SendPanel extends JPanel
             implements ActionListener, ContactListener {

    IXAgent agent;

    ActionListener actionListener
	= CatchingActionListener.listener(this);

    JTextArea contentText = new JTextArea(5, 40);

    RadioButtonBox typeSelector = makeTypeControl();
    RadioButtonBox prioritySelector = makePriorityControl();

    ComboChoice destinationChoice;

    JComboBox reportTypeChoice
        = new JComboBox(ReportType.values().toArray());

    JCheckBox reportCheck = new JCheckBox("Report Back - Ref =");

    ComboChoice refChoice = new ComboChoice();

    public SendPanel(IXAgent agent) {
	this.agent = agent;
	agent.getContactManager().addContactListener(this);
	setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
	setBorder(BorderFactory.createTitledBorder("Compose Message"));
	add(new JScrollPane(contentText));
	add(makeControlPanel());
    }

    public void actionPerformed(ActionEvent event) {
	String command = event.getActionCommand();
	Debug.noteln("I-TEST send action:", command);
	if (command.equals("Send")) {
	    sendTo(destinationChoice.getSelectedString());
	}
	else
	    Debug.noteln("Nothing to do for", command);
    }

    public void initReport(TaskItem item) {
	// Called to set up to report back about an item.
	Name ref = item.getRef();
	Name reportTo = item.getSenderId();
	Status status = item.getStatus(); // not yet used /\/

	if (!contentText.getText().equals("")) {
	    Object[] message = {
		"You have asked to send a report when there",
		"is unsent text in the display.  Do you want",
		"to clear the text and set up to send the report?"
	    };
	    switch(JOptionPane.showConfirmDialog
		      (null, message, "Confirm", JOptionPane.YES_NO_OPTION)) {
	    case JOptionPane.YES_OPTION:
		break;
	    case JOptionPane.NO_OPTION:
		return;
	    }
	}

	typeSelector.setSelection("Report");
	contentText.setText("");
	reportTypeChoice.setSelectedItem(ReportType.PROGRESS);
	if (reportTo != null)
	    destinationChoice.setSelectedItem(reportTo.toString());
	if (ref != null)
	    refChoice.setSelectedItem(ref.toString());
    }

    public void contactChange(ContactEvent e) {
	if (e.isDeletion()) {
	    AgentData data = e.getOldData();
	    destinationChoice.removeItem(data.getName());
	}
	else {
	    // Remember the name of any agent that sent to us.
	    AgentData data = e.getNewData();
	    String sender = data.getName();
	    if (!destinationChoice.hasItem(sender)) {
		Debug.noteln("New contact", sender);
		destinationChoice.addItem(sender);
	    }
	}
    }

    /**
     * Construct an object from the current state of the GUI
     * and send it to the specified destination.
     *
     * @see ix.util.IPC#exceptionlessSend(Object, Object)
     */
    protected void sendTo(String destination) {
	Sendable contents = (Sendable)objectFromControls();
	send(destination, contents);
	contentText.setText("");
    }

    public void send(String destination, Sendable contents) {

	Debug.noteln("Sending to", destination);
	Debug.noteln("Contents as XML:",
		     XML.objectToXMLString(contents));

	if (destination.equalsIgnoreCase("me")) {
	    // Send to self without going through IPC.
	    agent.handleInput(new IPC.BasicInputMessage(contents));
	}
	else if (destination.equals("a group ...")) {
	    sendToGroup(contents);
	}
	else
	    // IPC.exceptionlessSend(destination, contents);
	    IPC.sendObject(destination, contents);
    }

    public void sendCopy(String destination, Sendable contents) {
	// /\/: Making the copy is a pain.
	Sendable copy;
	try {
	    copy = (Sendable)contents.clone();
	}
	catch (CloneNotSupportedException ex) {
	    Debug.noteException(ex);
	    throw new Error("Can't clone " + contents);
	}
	send(destination, copy);
    }

    protected void sendToGroup(Sendable contents) {
	new GroupSender(agent, contents);
    }

    /**
     * Construct an object from the current state of the GUI.
     * The object will be an Issue, Activity, or Report.
     */
    protected Object objectFromControls() {
	String type = typeSelector.getSelection();
	if (type.equals("Issue"))
	    return issueFromControls();
	else if (type.equals("Activity"))
	    return activityFromControls();
	else if (type.equals("Constraint"))
	    return constraintFromControls();
	else if (type.equals("Report"))
	    return reportFromControls();
	else if (type.equals("Message"))
	    return messageFromControls();
	else
	    throw new Error("Unexpected message type " + type);
    }

    /**
     * Construct an issue from the current state of the GUI.
     */
    protected Issue issueFromControls() {
	return (Issue)fillTaskItemFromControls(new Issue());
    }

    /**
     * Construct an activity from the current state of the GUI.
     */
    protected Activity activityFromControls() {
	return (Activity)fillTaskItemFromControls(new Activity());
    }

    /**
     * Fills in the contents of an Issue or Activity.
     */
    protected TaskItem fillTaskItemFromControls(TaskItem item) {
	String text = contentText.getText().trim();
	String priority = prioritySelector.getSelection();
	String ref = refChoice.getSelectedString();
	boolean reportBack = reportCheck.isSelected();

	item.setPattern(Lisp.elementsFromString(text));
	item.setSenderId(Name.valueOf((String)agent.getAgentIPCName()));
	item.setPriority(priorityFromControls());
	if (!ref.equals(""))
	    item.setRef(Name.valueOf(ref));
	if (reportBack)
	    item.setReportBack(YesNo.YES);
	return item;
    }

    /**
     * Construct a constraint from the current state of the GUI.
     */
    protected Constraint constraintFromControls() {
	String text = contentText.getText().trim();
	String[] parts = Strings.breakAtFirst("=", text);
	String pat = parts[0].trim();
	String val = parts[1].trim();
	if (val.equals(""))
	    val = "true";
	String source = "(world-state effect (" + pat + ") = " + val + ")";
	LList spec = (LList)Lisp.readFromString(source);
	Constraint c = constraintParser.parseConstraint(spec);
	if (c == null)
	    throw new SyntaxException("Invalid constraint: "
				      + Strings.quote(text));
	c.setSenderId(Name.valueOf((String)agent.getAgentIPCName()));
	return c;
    }

    LTF_Parser constraintParser = new LTF_Parser();

    /**
     * Construct a report from the current state of the GUI.
     */
    protected Report reportFromControls() {
	String text = contentText.getText().trim();
	String ref = refChoice.getSelectedString();
	ReportType type = (ReportType)reportTypeChoice.getSelectedItem();

	Report report = new Report(text);
	report.setSenderId(Name.valueOf((String)agent.getAgentIPCName()));
	if (!ref.equals(""))
	    report.setRef(Name.valueOf(ref));
	if (!type.equals(""))
	    report.setReportType(type);

	return report;
    }

    /**
     * Construct a chat message from the current state of the GUI.
     */
    protected ChatMessage messageFromControls() {
	String text = contentText.getText();
	String sender = (String)agent.getAgentIPCName();
	return new ChatMessage(text, sender);
    }

    /**
     * Obtain a priority from the current state of the GUI.
     */
    protected Priority priorityFromControls() {
	String selected = prioritySelector.getSelection();
	// This is a bit messy.  /\/
	Debug.assert(selected.endsWith(" Priority"));
	String p = Strings.beforeFirst(" ", selected);
	return Priority.valueOf(p.toLowerCase());
    }


    /*
     * Details of GUI construction
     */

    protected JPanel makeControlPanel() {
	JPanel panel = new JPanel();
	panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS));

	panel.add(typeSelector);
	panel.add(prioritySelector);

	panel.add(Box.createHorizontalGlue());

	panel.add(makeSendControls());

	return panel;
    }

    protected RadioButtonBox makeTypeControl() {
	RadioButtonBox box = RadioButtonBox.createVerticalBox();

	JRadioButton issueButton = new JRadioButton("Issue");
	JRadioButton activityButton = new JRadioButton("Activity");
	JRadioButton constraintButton = new JRadioButton("Constraint");
	JRadioButton reportButton = new JRadioButton("Report");
	JRadioButton messageButton = new JRadioButton("Message");

	reportButton.setSelected(true);

	box.add(issueButton);
	box.add(activityButton);
	box.add(constraintButton);
	box.add(reportButton);
	box.add(messageButton);

	box.add(Box.createVerticalGlue());

	return box;	
    }

    protected RadioButtonBox makePriorityControl() {
	RadioButtonBox box = RadioButtonBox.createVerticalBox();

	box.add(new JRadioButton("Highest Priority"));
	box.add(new JRadioButton("High Priority"));
	box.add(new JRadioButton("Normal Priority", true));
	box.add(new JRadioButton("Low Priority"));
	box.add(new JRadioButton("Lowest Priority"));

	box.add(Box.createVerticalGlue());

	return box;
    }

    protected JPanel makeSendControls() {
	// /\/: Or just have a class and call new?

	JPanel panel = new JPanel();
	panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS));

	GridColumn left = new GridColumn();
	GridColumn right = new GridColumn();

	panel.add(left);
	panel.add(right);

	List names = agent.getContactManager().getSortedNameList();
	names.add(0, "me");
	names.add(1, "a group ...");
	destinationChoice = new ComboChoice(names.toArray());

	reportCheck.setHorizontalAlignment(SwingConstants.TRAILING);
	left.add(reportCheck);
	right.add(refChoice);

	JLabel typeLabel = new JLabel("Report Type =");
	typeLabel.setForeground(Color.black); 		// metal theme? /\/
	typeLabel.setHorizontalAlignment(SwingConstants.TRAILING);
	left.add(typeLabel);
	right.add(reportTypeChoice);

	JLabel toLabel = new JLabel("Recipient =");
	toLabel.setForeground(Color.black);
	toLabel.setHorizontalAlignment(SwingConstants.TRAILING);
	left.add(toLabel);
	right.add(destinationChoice);

	JButton sendButton =
	    new JButton("Send", Util.resourceImageIcon("ip2-send.gif"));
	sendButton.setHorizontalTextPosition(SwingConstants.LEFT);
	sendButton.addActionListener(actionListener);
	left.add(new JLabel(""));
	right.add(sendButton);
		 
	return makeFlowPanel(panel);

    }

    /**
     * Puts a component in a JPanel that has a flow layout to protect it
     * from the layout manager of the container to which the component
     * would otherwise have been added directly.
     */
    static public JPanel makeFlowPanel(Component c) {
	// FlowLayout is the default, but we want to specify some
	// things about it.
	JPanel panel = new JPanel();
	// Use constructor FlowLayout(int align, int hgap, int vgap).
	panel.setLayout(new FlowLayout(FlowLayout.RIGHT, 0, 5));
	panel.add(c);
	return panel;
    }
	
}
