/****************************************************************************
 * An editor/viewer panel for refinement structures
 *
 * @author Jussi Stader
 * @version 2.0
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 *
 *****************************************************************************
 */
package ix.iview;

import java.lang.Math;
import java.util.*;
import javax.swing.*;       
//import java.awt.*;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.*;
import ix.*;
import ix.util.*;
import ix.util.lisp.*;
import ix.icore.*;
import ix.icore.domain.*;
import ix.icore.domain.event.*;
import ix.iface.ui.*;
import ix.iface.ui.table.*;
import ix.iface.ui.util.*;
import ix.iview.*;
import ix.iview.igraph.*;
import ix.iview.table.*;
import ix.iview.domain.*;
import ix.iview.domain.event.*;
import ix.iface.domain.*;

/****************************************************************************
 * An editor/viewer panel for refinement structures.
 *
 * The editor has a flag that specifies whether it is in advanced mode (true)
 * or in simple mode (false).<p>
 *
 * The editor is created with an advanced flag (true for advanced).
 * Subsequently, the refinement can be set using the setRefinement method.<p>
 *
 * The editor will not let you update a current refinement yet, because we have
 * not yet worked out how to manage the impact this would have on other 
 * modules that use the domain. 
 * The editor will let you add new refinements (if you have changed an old one,
 * change the name to save it).<p>
 *
 * The refinement name, pattern, and comments are currently simply strings. They
 * can all be edited by typing into their fields. The pattern should not stay
 * as a string - it should change into something a bit more constrained.
 * This will be changed when the refinement models have been sorted out.<p>
 * 
 * Sub-activities and constraints can be added, deleted, and cleared, either
 * using the buttons or using keys (enter or insert for add, delete for
 * delete).<p>
 *
 * Sub-activities are simple LLists for the moment. Adding a sub-activity
 * expects the user to type in a pattern (currently just a string) and it will 
 * generate a number for the activity. Deleting sub-activities will also
 * delete any constraints in which the deleted activity participates.<p>
 *
 * The only constraints currently known are ordering constraints. Adding
 * constraints will bring up an editor that lets the user select activities
 * and time-points. For the moment, these "proper" constraints are translated
 * into simple versions because other modules cannot handle anything more
 * complex. A simple constraint "(<before> <after>)" stands for
 * "(<before> End Before Begin <after>). The constraint editor and the
 * DomainEditor are ready to provide and manage "proper" constraints, but we
 * will not provide them until other modules can deal with them.<p>
 *
 * Example code for using the ActionEditorPanel:
 *<PRE><code>
 *    ...
 *    editorPanel = new ActionEditorPanel(this,isAdvanced);
 *    mainJPanel.add(editorPanel); 
 *    ...
 *</code></PRE>
 *****************************************************************************
 */
public class ActionEditorPanel extends AConstructFramePanel 
  implements ConstructFraming, ConstraintListener, ActionListener,
  UIDomainListener
{

  //---------------------Fields-------------------------------------------
  /** The panels parent - a DomainEditor */
  private IXEditorPanel miniConstraintsPanel;
  private IXEditorPanel nodesTextPanel;
  private IXEditorPanel conditionsPanel;
  private boolean advanced = false;
  private IXRenderer advancedNodeRenderer = new FullNodeRenderer(false);
  private IXRenderer simpleNodeRenderer = new NodeNameRenderer(false);
  //private IXRenderer advancedIssueRenderer = new FullIssueRenderer(false);
  private IXRenderer advancedIssueRenderer = new IssueRenderer(false);
  private IXRenderer simpleIssueRenderer = new IssueRenderer(false);
  private boolean sequence = true;
  /** An editor for temporal relations */
  public OrderingEditor orderingEditor;
  public ConditionEffectEditor conditionEditor;
  public SpecEditor specEditor;
  /** An editor for temporal relations */
  //public ListRelationEditor relationEditor;
  //private JCheckBox sequenceCB;

  private JSplitPane actionSplit = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
  private JScrollPane scrollTree = new JScrollPane();
  private IXTreeTable actionTree;
  private ActionTreeTableModel treeModel;

  IFormModel formModel = new ActionFormModel();

  ConstructEditing currentPanel;

  private boolean graphMode = false;
  GExpansionEditor graphPanel;
	
  String lineSeparator = System.getProperty("line.separator");


  private UIRefinement uiConstruct;  //this FramePanel needs it


  //---------------------Constructors----------------------------------------

  /**
   * Creates the editor panel with the given parent in advanced/simple mode.
   * Creates all panel components, the mouse listener, and the relation editor.
   *
   * @param parent the DomainEditor frame that the panel belons to. Used 
   *   mainly to keep toggle buttons and menus in synch
   * @param isAdvanced a flag specifying whether the editor is to display 
   *   its advanced features or not.
   */    
  public ActionEditorPanel(DomainEditor theParent, boolean isAdvanced){
    //make all things (some are made during initialisation)
    super(theParent);
    currentPanel = (ActionFormPanel)formPanel;
    setupActionTree();
    advanced = isAdvanced;
    this.enableAdvanced();
    String[] names = {"name"};
    formModel.setNameFields(names);
    //formModel.setDisplayFields(displays);//done in enableAdvanced()
    formPanel.setModel(formModel);
    MouseListener ml = new EdMouseAdapter();
    formPanel.addFormMouseListener(ml);
    newConstruct();
  }

  protected JPanel setupMainPanel() {
    formPanel = new ActionFormPanel(this);

    Debug.noteln("made form panel for SEP");
    graphPanel = new GExpansionEditor(this);
    return formPanel;
  }

  protected void setupActionTree() {
    actionSplit.setRightComponent(scrollPane);
    actionSplit.setLeftComponent(scrollTree);
    add(actionSplit, BorderLayout.CENTER);
    JPanel treePanel = new JPanel();
    scrollTree.add(treePanel);
    treeModel = new ShortActionTreeTableModel(getUIDomain());
    actionTree = new IXTreeTable(treeModel);
    actionTree.addMouseListener(new ActionTableMouser());
    actionTree.setPreferredScrollableViewportSize(new Dimension(50, 500));
    scrollTree.getViewport().add(actionTree);
    actionSplit.setOneTouchExpandable(true);
    //actionSplit.setDividerLocation(0.33);
    //actionSplit.setResizeWeight(0.33);
    actionSplit.setDividerSize(5);
    actionSplit.setDividerLocation(170);


  }

  protected void setPanel(ConstructEditing panel) {
    //Dimension vSize = scrollPane.getViewport().getSize();
    Dimension pSize = parent.getSize();
    //Debug.noteln("AEP: size is", pSize);
    if (currentPanel != null) {
      ((JPanel)currentPanel).setVisible(false);
      scrollPane.getViewport().remove((JPanel)currentPanel);
    }
    scrollPane.getViewport().add((JPanel)panel);
    currentPanel = panel;
    mainPanel = (JPanel)panel;
    ((JPanel)panel).setVisible(true);
    scrollPane.invalidate();
    //scrollPane.getViewport().setSize(vSize);
    parent.setSize(pSize);
    parent.validate();
  }


  //---------------------Field access----------------------------------------

  /**
   * @return the current setting of the advanced flag
   */
  public boolean isAdvancedSet(){ return advanced;}
  /**
   * Sets the editor to advanced (true) or simple (false) mode.
   * If the editor was in graph mode
   * If the editor is already in the given mode, nothing is done.
   * For advanced mode, a constraint panel appears with all constraints
   * listed and buttons for adding constraings.
   * For simple mode, the constraint panel only contains a radio button for
   * sequencing nodes and a button to set the mode to advanced.
   */
  public void setAdvanced(boolean newAdvanced) {
    if (!beforeSwitching()) return;

    //coming from graph mode so change panel and if diff adv fields
    if (graphMode) {
      //noteConstruct();
      this.setPanel((ActionFormPanel)formPanel);
      graphMode = false;
      if (newAdvanced != advanced) {
	advanced = newAdvanced;
	setAdvancedFields(newAdvanced, uiConstruct);
      }
      else {
	((ActionFormPanel)formPanel).setUIConstruct(uiConstruct);
	invalidate();
	validate();
      }
    }
    else if (newAdvanced != advanced) {
      advanced = newAdvanced;
      this.enableAdvanced();
    }
    /*
    if (newAdvanced != advanced) {
      advanced = newAdvanced;
      this.enableAdvanced();
    }
    if (graphMode) {
      noteConstruct();
      this.setPanel((ActionFormPanel)formPanel);
      graphMode = false;
      ((ActionFormPanel)formPanel).setUIConstruct(uiConstruct);
    }
    */
  }

  /**
   * @return the current setting of the sequence flag
   */
  public boolean isSequenceSet() { 
    return ((ActionFormPanel)formPanel).isSequenceSet();
  }
  /**
   * Sets the flag to put nodes in sequence (true) or in parallel (false)
   * If the state of the checkbox does not correspond to the new setting,
   * it is changed.
   * If the editor is already in the given mode, nothing is done.
   * Currently, nothing else changes in the panel
   */
  public void setDefaultSequence(boolean newSequence) {
    if (newSequence == sequence) return;
    else sequence = newSequence;
  }



  /** 
   * Checks whether the orderings are a sequence of the nodes.
   * @param nodes the nodes that can be part of the ordering.
   * @param orderings the orderings to be tested.
   */
  public static boolean orderingsAreSequential(LList nodes, LList orderings) {
    HashSet sequence = new HashSet(makeSequence(nodes));
    return new HashSet(orderings).equals(sequence);
    }

  public void setOrdering(boolean isSequence) {
    //Debug.noteln("AEP: Setting ordering to " + isSequence);
    if (isSequence) makeSequence();
    else makeParallel();
  }

  /**
   * Puts the nodes of the current refinement into sequence.
   */
  public void makeSequence() {
    // make all current sub-nodes sequential
    try {
      uiConstruct.setOrderings(makeSequence(uiConstruct.getNodes()));
      this.setOrderings();      
    }
    catch (Exception e) {
      Debug.noteException(e);
    }
  }
  /**
   * Puts the given nodes into sequence.
   */
  public static List makeSequence(List nodes) {
    //Debug.noteln("ActionEditorPanel - makeSequenceNodes");
    // make all current sub-nodes sequential
    NodeSpec previous = null;
    NodeSpec current;
    End tpBefore = End.END;
    End tpAfter = End.BEGIN;
    List orderings = new LListCollector();
    for (Iterator i = nodes.iterator(); i.hasNext(); ) {
      current = (NodeSpec)i.next();
      //Debug.noteln(" AEP: working on node", current.getId());
      if (previous != null) {
	Ordering ordering = 
	  new Ordering(new NodeEndRef(tpBefore, previous.getId()), 
		       new NodeEndRef(tpAfter, current.getId()));
	orderings.add(ordering);
	//Debug.noteln(" AEP: Orderings are", orderings);
      }
      previous = current;
    }
    return orderings;
  }


  /**
   * Puts the nodes of the current refinement in parallel.
   */
  public void makeParallel() {
    clearOrderings();
  }


  //---------------------Toggles----------------------------------------

  /*
   * If the mode is not advanced, disable constraints and other useful things
   * Advanced:
   *  sub-act renderer is FullNodeRenderer
   *  uses constraintsPanel
   *  uses nodesPanel
   * Simple:
   *  sub-act renderer is NodeNameRenderer
   *  uses miniConstraintsPanel
   *  uses nodesTextPanel 
   */
  private void enableAdvanced() {
    setAdvancedFields(advanced, uiConstruct);
    /*
    UIRefinement uir = null;
    if (uiConstruct != null) uir = (UIRefinement)noteConstruct();
    setAdvancedFields(advanced, uir);
    */
  }

  /**
   * Sets the fields of the form panel to advanced/simple mode.
   * Also sets UIConstruct (if not null) and invalidates
   */
  private void setAdvancedFields(boolean advanced, UIRefinement uir) {
    if (advanced) {
      String[] displays = {"pattern","issues","nodes","orderings",
			   //"conditions", 
			   "comments"};
      formPanel.unsetSimpleField("nodes");
      formPanel.unsetSimpleField("issues");
      formPanel.setRenderer("nodes", advancedNodeRenderer);
      formPanel.setRenderer("issues", advancedIssueRenderer);
      formModel.setDisplayFields(displays);
    }
    else {
      String[] displays = {"pattern","issues","nodes","miniConstraints",
			   "comments"};
      formPanel.setSimpleField("nodes");
      formPanel.setSimpleField("issues");
      formPanel.setRenderer("nodes", simpleNodeRenderer);
      formPanel.setRenderer("issues", simpleIssueRenderer);
      formModel.setDisplayFields(displays);
    }
    if (uir != null) ((ActionFormPanel)formPanel).setUIConstruct(uir);
    uiConstruct = uir;
    this.invalidate();
    validate();
  }

  //---------------------Window building-------------------------------------

  public void ensureRelationEditor(){
    if (orderingEditor == null){
      orderingEditor = new OrderingEditor();
      orderingEditor.addConstraintListener(this);
      orderingEditor.setVisible(false);
    } 
  }
  public void ensureConditionEditor(){
    if (conditionEditor == null){
      conditionEditor = new ConditionEffectEditor();
      conditionEditor.addConstraintListener(this);
      conditionEditor.setVisible(false);
    } 
  }
  public void ensureSpecEditor(String type){
    if (specEditor == null){
      specEditor = new SpecEditor(type);
      specEditor.addConstraintListener(this);
      specEditor.setVisible(false);
    } 
    specEditor.setType(type);
  }
  public void ensureSpecEditor(String type, String init) {
    ensureSpecEditor(type);
    specEditor.setMessage(init);
  }


  public boolean beforeClosing() {
    if (specEditor != null) specEditor.setVisible(false);
    if (orderingEditor != null) orderingEditor.setVisible(false);
    if (conditionEditor != null) conditionEditor.setVisible(false);
    return beforeSwitching();
  }
  

  //---------------------Public services--------------------------------------

  /**
   * Sets the given UI construct to be viewed/edited. 
   * Also sets the uiConstruct in the current editorPanel
   *
   * @param object the UI object to be viewed/edited
   */    
  public void setUIConstruct(UIObject object) {
    //noteConstruct();
    if (beforeSwitching()) {
      uiConstruct = (UIRefinement)object;
      //set the object in the actual panel
      currentPanel.setUIConstruct(uiConstruct);
    }
  }
  /**
   * Sets the given UI construct to be viewed/edited without noting the 
   * previous construct.
   * Also sets the uiConstruct in the current editorPanel
   *
   * @param object the UI object to be viewed/edited
   */    
  public void setUIConstructOnly(UIObject object) {
    uiConstruct = (UIRefinement)object;
    //set the object in the actual panel
    //currentPanel.setUIConstructOnly(uiConstruct);
  }
  /**
   * Clears the current construct and the panel.
   * Also sets the uiConstruct in the current editorPanel
   */    
  public void clearUIConstruct() {
    uiConstruct = new UIRefinement(parent.uiDomain);
    //set the object in the actual panel
    ((ConstructEditing)graphPanel).setUIConstruct(uiConstruct);
    ((ConstructEditing)formPanel).setUIConstruct(uiConstruct);
  }

  public UIRefinement makeNewRefinement(String name, LList pattern) {
    
    UIRefinement newR = new UIRefinement(parent.getUIDomain());
    newR.setName(name);
    newR.setPattern(pattern);

    return newR;
  }


  public ArrayList getAllExpansions(NodeSpec node) {
    if (parent.uiDomain == null) return null;
    LList pattern = node.getPattern();
    Set exps = parent.uiDomain.getMatchingRefinements(pattern);
    return new ArrayList(exps);
  }

  //---------------------Private services--------------------------------------

  /**
   * Note the settings in the editor in a UIObject. Do not save into doamin.
   */
  protected UIObject noteConstruct() {
    //make sure there is a current construct
    if (uiConstruct == null) uiConstruct = new UIRefinement(parent.uiDomain);
    currentPanel.saveToObject(uiConstruct);

    uiConstruct.noteChange();
    //make sure the refinement is noted into the right domain
    if (!parent.getUIDomain().equals(uiConstruct.getUIDomain())) 
      uiConstruct.moveToUIDomain(parent.getUIDomain());
    parent.getUIDomain().updateConstruct(uiConstruct);
    //Debug.noteln("AEP: Noted construct");
    //Debug.noteln(" ",uiConstruct.print());
    return uiConstruct;
  }

  //---------------------filling the panel details

  private void setChildren() {
    formPanel.displayFieldData("nodes");
  }
  
  private void setIssues(){
    formPanel.displayFieldData("issues");
  }
  
  private void setOrderings() {
    if (isAdvancedSet())
      formPanel.displayFieldData("orderings");
    else 
      formPanel.displayFieldData("miniConstraints");
  }
  private void setConstraints() {
    formPanel.displayFieldData("constraints");
  }
  
  /**
   * Same as setUIConstruct, but remembers the setting of the miniConstraints
   */
  protected void resetUIConstruct(UIObject object) {
    if (!advanced) {
      MiniConstraintPanel mcp = 
	(MiniConstraintPanel)formPanel.getFieldBit("miniConstraints");
      AbstractButton lastB = mcp.getLast();
      setUIConstruct(object);
      formPanel.setFieldData("miniConstraints", lastB.getText());
    }
    else setUIConstruct(object);
  }

  //---------------------User services----------------------------------------

  //************** these requests may come from the frame (menu selections)
  /**
   * User request to check the refinement - not implemented.
   */
  public boolean checkConstruct(){
    //notImp("Checking activities");
    //userAddCondition();
    return true;
  }

  /**
   * User request to make a new action.
   */
  public void newConstruct() {
    //noteConstruct();
    //if (beforeSwitching()) done in setUIConstruct
    setUIConstruct(new UIRefinement(parent.uiDomain)); 
  }
  /**
   * User request to make a new action.
   */
  public UIObject makeNewConstruct() {
    return new UIRefinement(parent.uiDomain); 
  }
  /**
   * User request to save an action into the draft domain.
   */
  public void saveConstruct(){
    noteConstruct();
    currentPanel.setUIConstruct(uiConstruct);
    //uiConstruct.saveToDomain();
  }
  /**
   * User request to revert an action to when it was last saved in 
   * the draft domain.
   */
  public void revertConstruct(){
    uiConstruct.loadFromDomain();
    currentPanel.setUIConstruct(uiConstruct);
  }
  /**
   * User request to revert an action to when it was last saved in 
   * the draft domain.
   */
  public void revertOConstruct(){
    uiConstruct.loadFromOriginal();
    currentPanel.setUIConstruct(uiConstruct);
    //uiConstruct.saveToDomain();
  }
  public List getAllConstructs() {
    if (parent.uiDomain == null) return null;
    List constructs = parent.uiDomain.getAllRefinements();
    return new ArrayList(constructs);
  }
  public Named getNamedConstruct(String name) {
    if (parent.getUIDomain() == null) return null;
    Object c = parent.getUIDomain().getNamedRefinement(name);
    if ((c != null) && (UIObject.class.isInstance(c)))
      return (Named)c;
    else return parent.getUIDomain().newUIObject((IXObject)c);
  }


  /**
   *  User request to put the nodes of the current action into sequence.
   */
  public void userMakeSequence(){
    makeSequence();
  }
  /**
   *  User request to put the nodes of the current action in parallel.
   */
  public void userMakeParallel(){
    makeParallel();
  }


  //************** service sub-activity buttons (add, edit, delete)
  private void userAddChild() {
    //Debug.noteln("Adding children for refinement", uiConstruct);
    this.ensureSpecEditor("sub-activity",
			  "Please enter the pattern of the sub-activity.");
    specEditor.start();
  }
  private void userAddIssue(){
    //Debug.noteln("Adding children for refinement", uiConstruct);
    this.ensureSpecEditor("issue",
			  "Please enter the pattern of the issue.");
    this.specEditor.start();
  }

  /**
   * Parse the nodes from the text area and store them in the current
   * refinement.
   */
  public void noteTextNodes() {
    //Debug.noteln("Noting nodes from text");
    JTextArea nt = (JTextArea)nodesTextPanel.getItemComponent();
    List nodes = UIRefinement.parseNodes(getUIDomain(), nt.getText());
    uiConstruct.setNodes(nodes); 
    if (isSequenceSet()) makeSequence();
  }

  private void userDeleteChild(){
    IXEditorPanel nodesPanel = formPanel.getFieldBit("nodes");
    JList list = (JList) nodesPanel.getItemComponent();
    if (list.isSelectionEmpty()) {
      String message = "Please select a sub-activity to delete first.";
      JOptionPane.showMessageDialog(this,message);} 
    else {
      Object[] values = list.getSelectedValues();
      for (int i = 0; i < values.length; i++){
	Object child = values[i];
        uiConstruct.deleteNode((NodeSpec)child);
      }
      list.clearSelection();
      this.setChildren();
      this.setOrderings();
    }
  }
  private void userDeleteIssue(){
    IXEditorPanel issuesPanel = formPanel.getFieldBit("issues");
    JList list = (JList) issuesPanel.getItemComponent();
    if (list.isSelectionEmpty()) {
      String message = "Please select an issue to delete first.";
      JOptionPane.showMessageDialog(this,message);} 
    else {
      Object[] values = list.getSelectedValues();
      for (int i = 0; i < values.length; i++){
	Object child = values[i];
        uiConstruct.deleteIssue((Issue)child);
      }
      list.clearSelection();
      this.setIssues();
    }
  }
  private void userEditChild(){
    //notImp("editing nodes");
    IXEditorPanel nodesPanel = formPanel.getFieldBit("nodes");
    JList list = (JList) nodesPanel.getItemComponent();
    if (list.isSelectionEmpty()) {
      String message = "Please select a sub-activity to edit first.";
      JOptionPane.showMessageDialog(this,message);} 
    else {
      Object[] values = list.getSelectedValues();
      ensureSpecEditor("sub-activity",
		       "Please enter the new pattern of the sub-activity.");
      specEditor.start(uiConstruct, values[0]);
    }
  }
  private void userEditIssue(){
    //notImp("editing issues");
    IXEditorPanel issuesPanel = formPanel.getFieldBit("issues");
    JList list = (JList) issuesPanel.getItemComponent();
    if (list.isSelectionEmpty()) {
      String message = "Please select an issue to edit first.";
      JOptionPane.showMessageDialog(this,message);} 
    else {
      Object[] values = list.getSelectedValues();
      ensureSpecEditor("issue",
		       "Please enter the new pattern of the issue.");
      specEditor.start(uiConstruct, values[0]);
    }
  }
  private void userEditConstraint(){
    //notImp("editing issues");
    IXEditorPanel cPanel = formPanel.getFieldBit("orderings");
    JList list = (JList) cPanel.getItemComponent();
    if (list.isSelectionEmpty()) {
      String message = "Please select a constraint to edit first.";
      JOptionPane.showMessageDialog(this,message);} 
    else {
      Object[] values = list.getSelectedValues();
      ensureRelationEditor();
      orderingEditor.start(uiConstruct, values[0]); 
    }
  }

  private void userClearChildren(){
    // if the refinement has no children, just reset the field to be safe
    List nodes = uiConstruct.getNodes();
    if ((nodes == null) || (nodes.isEmpty())) {
      this.setChildren();
      return;
    }
    // if the action has children, get the user to confirm
    String[] message = {"This will clear all sub-activities.",
			" Are you sure you want to do this?"};
    int answer = JOptionPane.showConfirmDialog(this, message, "Warning", 
					       JOptionPane.OK_CANCEL_OPTION, 
					       JOptionPane.WARNING_MESSAGE);
    if (answer == JOptionPane.OK_OPTION) {
      uiConstruct.setNodes(Lisp.NIL);
      uiConstruct.setOrderings(Lisp.NIL);
      this.setChildren();
      this.setOrderings();
    }
  }

  //************** service constraints buttons (add, edit, delete)
  private void userAddConstraint(){
    //Debug.noteln("Adding constraints for refinement",uiConstruct);
    ensureRelationEditor();
    orderingEditor.start(uiConstruct);
  }
  private void userAddCondition(){
    Debug.noteln("Adding condition or effect for refinement",uiConstruct);
    ensureConditionEditor();
    conditionEditor.start(uiConstruct);
  }

  private void userDeleteConstraint(){
    IXEditorPanel cPanel = formPanel.getFieldBit("orderings");
    JList list = (JList) cPanel.getItemComponent();
    if (list.isSelectionEmpty()) {
      String message = "Please select a constraint to delete first.";
      JOptionPane.showMessageDialog(this,message);} 
    else {
      Object[] values = list.getSelectedValues();
      for (int i = 0; i < values.length; i++){
	Object relation = (Object) values[i];
	List orderings = uiConstruct.getOrderings();
	orderings.remove(relation);
	uiConstruct.setOrderings(orderings);
      }
      list.clearSelection();
      this.setOrderings();
    }
  }

  private boolean userClearConstraints() {
    return clearOrderings();
  }

  protected boolean clearOrderings() {
    // if the action has no constraints, just reset the field to be safe
    List orderings = uiConstruct.getOrderings();
    if ((orderings == null) || (orderings.isEmpty())) {
      this.setOrderings();
      return true;
    }
      // if the action has constraints, get the user to confirm
    String[] msg = {"This will clear all orderings between sub-activities.",
		    " Are you sure you want to do this?"};
    int answer = JOptionPane.showConfirmDialog(this, msg, "Warning", 
					       JOptionPane.OK_CANCEL_OPTION, 
					       JOptionPane.WARNING_MESSAGE);
    if (answer == JOptionPane.OK_OPTION) {
      uiConstruct.setOrderings(Lisp.NIL);
      this.setOrderings();
      return true;
    }
    return false;
  }
  /*
  private void userDeleteConstraint(){
    notImp("Deleting constraints");
  }
  */



  //---------------------UIDomainListener things------------------------------

  /** 
   * Called when a UI construct is added. 
   * Use getSource() to get at the domain (UI/IX) in which the event occurred.
   * Use getObject() to get at the object (UI/IX) that has been added.
   */
  public void constructAdded(UIDomainEvent event){}
  public void constructAdded(RefinementEvent event){}
  /** 
   * Called when a UI construct is edited. 
   * Use getSource() to get at the UIDomain in which the event occurred.
   * Use getObject() to get at the UIObject that has been edited.
   */
  public void constructEdited(UIDomainEvent event){}
  /** 
   * Called when a UI construct is removed. 
   * Use getSource() to get at the UIDomain in which the event occurred.
   * Use getObject() to get at the UIObject that has been removed.
   */
  public void constructRemoved(UIDomainEvent event) {   
    if (event.getObject().equals(uiConstruct)) 
      clearUIConstruct();
  }
  public void domainCleared(UIDomainEvent event) {
    //Debug.noteln("AEP: got cleared domain");
    clearUIConstruct();
  }
  public void domainEdited(UIDomainEvent se) {
  }
  public void domainSet(UIDomainEvent event) {
    //Debug.noteln("AEP: got set domain");
    clearUIConstruct();
  }

  //---------------------Actions----------------------------------------

  //************** relay buttons to their services *************************
  class EdMouseAdapter extends java.awt.event.MouseAdapter 
  implements SchemaTerms {
    public void mouseClicked(MouseEvent event){
      Component component = event.getComponent();
      component.requestFocus();
      //Debug.noteln("ActionEditorPanel: mouse from ", component.getClass());
      if (!AbstractButton.class.isAssignableFrom(component.getClass())) return;
      
      AbstractButton button = (AbstractButton) component;
      //Debug.noteln("AEP: Button command " + button.getText());
      IXEditorPanel buttonParent = 
	(IXEditorPanel)button.getParent().getParent();
      //Debug.noteln(" Parent name " + buttonParent.getName());

      if (buttonParent.getName().equals("Nodes")) {
	if (button.getText() == "Add") 
	  ActionEditorPanel.this.userAddChild();
	else if (button.getText() == "Delete") 
	  ActionEditorPanel.this.userDeleteChild();
	else if (button.getText() == "Edit") 
	  ActionEditorPanel.this.userEditChild();
	else 
	  ActionEditorPanel.this.notImp(button.getText() + " for nodes");
	return;
      }
      else if (buttonParent.getName().equals("Issues")) {
	if (button.getText() == "Add") 
	  ActionEditorPanel.this.userAddIssue();
	else if (button.getText() == "Delete") 
	  ActionEditorPanel.this.userDeleteIssue();
	else if (button.getText() == "Edit") 
	  ActionEditorPanel.this.userEditIssue();
	else 
	  ActionEditorPanel.this.notImp(button.getText() + " for issues");
	return;
      }
      else if ((buttonParent.getName().equals("Ordering")) 
	       || (buttonParent.getName().equals("Orderings"))) {
	//Debug.noteln("AEP: got a constraints panel button");
	if (button.getText() == "Add") 
	  ActionEditorPanel.this.userAddConstraint();
	else if (button.getText() == "Delete")
	  ActionEditorPanel.this.userDeleteConstraint();
	else if (button.getText() == "Edit") 
	  ActionEditorPanel.this.userEditConstraint();
        else if (button.getText().equals(O_PARALLEL)) {
	  ((MiniConstraintPanel)buttonParent).setLast(button);
	  ActionEditorPanel.this.setOrdering(false);
	}
	else if (button.getText().equals(O_SEQUENCE)) {
	 ((MiniConstraintPanel)buttonParent).setLast(button);
	  ActionEditorPanel.this.setOrdering(true);
	}
	else if (button.getText() == "other") {}
	else if (button.getText() == "Comprehensive") 
	  ActionEditorPanel.this.parent.setAdvanced(true);
	else
	  ActionEditorPanel.this.notImp(button.getText() + " for constraints");
	return;
      }     
      else if (buttonParent.getName().equals("Conditions")) {
	ActionEditorPanel.this.notImp("Command " + button.getText() + " " +
				    conditionsPanel.getName());
	return;
      }
      else ActionEditorPanel.this.notImp("Command " + button.getText());
    }
  }

  /**
   * Wakes up the ConstraintListener with a newly defined ordering.
   * This is called when a new ordering has been defined by the component
   * with which the ConstraintListener registered its interest.
   *
   * @param o the object to which the constraint applies
   * @param constraint the new constraint
   */
  public void gotConstraint(Object o, Object constraint) {
    UIRefinement uir = (UIRefinement)o;
    if ((uir != null) && (!uir.equals(uiConstruct))) {
      String message = "Trying to edit constraints of refinement " + 
	uiConstruct.getName() + " with constraints of refinement " +
	uir.getName();
      JOptionPane.showMessageDialog(this, message);
      return;
    }
    else if (constraint == null) {
      //either an empty spec or an edited one. Make sure to update panel.
      this.setOrderings();
      this.setChildren();
      this.setIssues();
      validate();
      return;
    }

    else if (Ordering.class.isInstance(constraint)) {
      uiConstruct.getOrderings().add(constraint);
      this.setOrderings();
    }
    else if (Constraint.class.isInstance(constraint)) {
      uiConstruct.getConstraints().add(constraint);
      this.setConstraints();
    }
    else if (UISpec.class.isInstance(constraint)) {
      UISpec spec = (UISpec)constraint;
      LList pattern = spec.getSpec();
      if (pattern.isEmpty()) return;
      else { 
	String type = spec.getType();
	if (type.equals("sub-activity")) {
	  //make a new child, add it to the refinement. Then update display.
	  Name newId = uiConstruct.getNewNodeId();
	  uiConstruct.addNode(new UINodeSpec(getUIDomain(),newId,pattern));
	  this.setChildren();
	  if ((isSequenceSet()) && (!isAdvancedSet())) makeSequence(); 
	}
	else if (type.equals("issue")) {
	  //make a new issue, add it to the refinement. Then update display.
	  uiConstruct.addIssue(new Issue(pattern));
	  this.setIssues();
	} 
      }
    } 
    else Debug.noteln("AEP: ****Got a class", constraint.getClass().getName());

  }
        
  /**
   * Recognises panel specific events and processes them.
   * This is a separate method because it must let other event handlers know
   * whether it recognised the event.
   *
   * @param event the event to be processed.
   * @return true if the event was recognised, false if not.
   */
  public boolean frameActionPerformed(ActionEvent event){
    //deal with commands
    String command = event.getActionCommand();
    if (command == "newEdit") newConstruct();
    else if (command == "editConstruct") {
      //get the construct name from the item text and show it
      JMenuItem item = (JMenuItem)event.getSource();
      String name = item.getText().trim();
      setUIConstruct((UIObject)getNamedConstruct(name));  
    }
    else if (command == "modify") saveConstruct();
    else if (command == "revert") revertConstruct();
    else if (command == "revertO") revertOConstruct();
    else if (command == "check") checkConstruct();

    else if (command == "makeSequence") userMakeSequence();
    else if (command == "makeParallel") userMakeParallel();
    else if (command == "defSequence") {
      Object object = event.getSource();
      parent.setDefaultSequence(((JCheckBoxMenuItem)object).getState());
    }
    else if (command == "graphical") {
      //notImp("The graphical activity editor");
      if (!graphMode) {
	if (beforeSwitching()) {
	  this.setPanel(graphPanel);
	  graphMode = true;
	  parent.setGraphical();
	  parent.setOriginalStyle();
	  graphPanel.setUIConstruct(uiConstruct);
	}
      }
    }
    else if (command == "minimal") parent.setAdvanced(false);
    else if (command == "comprehensive") parent.setAdvanced(true);
    else return false;
    return true;

  }

  /**
   * Wakes up the ActionListener with a user action.
   * This is called when a KeyStroke or other panel event happens in which
   * the ActionListener registered its interest.
   */
  public void actionPerformed(ActionEvent ae){
    // give the method above a chance to recognise the event.
    if (frameActionPerformed(ae)) return;
    String command = ae.getActionCommand();
    Debug.noteln("ActionEditorPanel: Got action command", command);
    if (command.equals("add node")) userAddChild();
    else if (command.equals("delete node")) userDeleteChild();
    else if (command.equals("add ordering")) userAddConstraint();
    else if (command.equals("delete ordering")) userDeleteConstraint();
    else Debug.noteln("ActionEditorPanel cannot recognise command", command);
  }


  protected static void notImp(String message) {
    UIUtil.notImplemented(null,message);
  }


  private class ActionTableMouser extends MouseAdapter {
    /**
     * If the left button was clicked on a row (except column 0) put the action
     * in the editor.
     */
    public void mouseClicked(MouseEvent me) {
      //these clicks do not need the row to be selected - work on where clicked
      int column = actionTree.columnAtPoint(me.getPoint());
      if ((column == 0) || !SwingUtilities.isLeftMouseButton(me)) return;

      int row = actionTree.rowAtPoint(me.getPoint());
      //Component component = me.getComponent();
      //Debug.noteln("AEP: got component mouse", component);
      
      if ((row >= 0) && (row < actionTree.getRowCount())) {
	ActionTreeTableModel model = 
	  (ActionTreeTableModel)actionTree.getModel();
	Object construct = model.getRowObject(row);
	if (construct != null) {
	  if (UIObject.class.isInstance(construct)) 
	    setUIConstruct((UIObject)construct);
	  else if (IXObject.class.isInstance(construct)) 
	    setConstruct((IXObject)construct);
	  else if (UINodeSpec.class.isInstance(construct)) {
	    UIRefinement uir = new UIRefinement(getUIDomain());
	    uir.setPattern(((UINodeSpec)construct).getPattern());
	    uir.setName("Undefined");
	    setUIConstruct(uir); 
	  }
	}
      }
    }
  }


}

/*Issues***************************************************************
 *
 * NOTE: this does not do checkConsistency on new constructs anymore.
 *
/*Todos***************************************************************
 *
 ** edit pattern: patterns should have constraints on their structure. 
 *   e.g. <verb> <variables>|<other things>
 ** lists etc. should use MVC 
 *
 *
 ** simple/advanced/graphical should be done via "mode" switch
 ** documentation of simple/advanced/graphical and their ConstructPanels things
 *****************************************************************************
 */
