/****************************************************************************
 * A panel containing form-style sub-panels for viewing/editing an object
 *
 * @author Jussi Stader
 * @version 1.0
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 *
 *****************************************************************************
 */
package ix.iview;

import ix.*;
import ix.util.Debug;
import ix.util.lisp.*;
import ix.iface.ui.util.*;
import ix.iface.ui.*;
import ix.icore.*;
import ix.icore.domain.*;
import ix.iview.util.*;
import ix.iview.domain.*;

import java.util.*;
import java.lang.reflect.*;
import java.awt.event.*;
import javax.swing.*;

/**
 * A panel containing form-style sub-panels for viewing/editing an object
 *
 */
public class ActionFormPanel extends IFormPanel implements ConstructEditing {

  String lineSeparator = System.getProperty("line.separator");
  AConstructFramePanel parent;
  private UIObject uiConstruct;

  ActionFormPanel(AConstructFramePanel parent) {
    super();
    this.parent = parent;
  }

  //-----------------------IFormPanel things-------------------------------

   protected void setRenderers() {
    IXRenderer r = new OrderingRenderer(false);
    fieldRenderers.put("orderings", r);
  }
 

  /**
   * Render values for: 
   * pattern (list to display),
   * text nodes (write a line for each node)
   */
  protected Object renderValue(String field, Object value) {
    //Debug.noteln("Doing field with name " + field);
    if (value == null) return null;
    else if (field.equals("pattern")) {
      //Debug.noteln("making pattern value");
      return UIUtil.listToDisplay((LList) value);
    }
    else if (field.equals("nodes") && isSimpleField(field)) {
      //Debug.noteln("making simple nodes value");
      //Collection sorted = IViewUtil.sortNodeCollection((List) value);
      NodeSpec spec;
      LList childPattern;
      String text = "";
      for (Iterator i = ((List)value).iterator(); i.hasNext(); ) {
	spec = (NodeSpec)i.next();
	childPattern = spec.getPattern();
	text = text + Lisp.elementsToString(childPattern) + lineSeparator;
      }  
      return text;
    }
    else if (field.equals("issues") && isSimpleField(field)) {
      //Debug.noteln("making simple issues value");
      //Collection sorted = IViewUtil.sortNodeCollection((List) value);
      Issue issue;
      LList pattern;
      String text = "";
      for (Iterator i = ((List)value).iterator(); i.hasNext(); ) {
	issue = (Issue)i.next();
	pattern = issue.getPattern();
	text = text + Lisp.elementsToString(pattern) + lineSeparator;
      }  
      return text;
    }
    else return super.renderValue(field, value);
  }
  

  /**
   * De-render values for: 
   * pattern (lisp.elementsFromString),
   * nodes (parse if text, otherwise ignore because updates are immediate)
   * orderings (return als LList)
   */
  protected Object deRenderValue(String field, Object data) {
    //Debug.noteln("AFP: de-rendering field with name " + field);
    //Debug.noteln("AFP: Data is of class " + data.getClass().getName());
    if (field.equals("pattern")) {
      if (data == null) return Lisp.NIL;
      //Debug.noteln("making pattern value from data");
      return Lisp.elementsFromString((String)data);
    }
    else if (field.equals("nodes")) {
      if (data == null) return Lisp.NIL;
      if (data.getClass() == String.class) {
	//Debug.noteln("AFP: Parsing nodes from data: " + data);
	return UIRefinement.parseNodes(getUIDomain(), (String)data);
      }
      else {
	//nodes are noted as we go along, so look them up and forget data!
	//Debug.noteln(" AFP: getting nodes from refinement");
	UIRefinement uir = (UIRefinement)model.getObject();
	if (uir != null) 
	  return uir.getNodes();
	else return Lisp.NIL;
      }
    }
    else if (field.equals("issues")) {
      if (data == null) return Lisp.NIL;
      if (data.getClass() == String.class) {
	//Debug.noteln("AFP: Parsing issues from data: " + data);
	return UIRefinement.parseIssues(getUIDomain(), (String)data);
      }
      else {
	//issues are noted as we go along, so look them up and forget data!
	//Debug.noteln(" AFP: getting issues from refinement");
	UIRefinement uir = (UIRefinement)model.getObject();
	if (uir != null) 
	  return uir.getIssues();
	else return Lisp.NIL;
      }
    }
    else if (field.equals("orderings")) {
      if (data == null) return Lisp.NIL;
      java.util.List values = Arrays.asList((Object[])data);
      LListCollector orderings = new LListCollector(values);
      return orderings;
    }
    else return super.deRenderValue(field, data);
  }

  /**
   * make a special display bit for miniConstraints
   */
  protected IXEditorPanel makeDisplayBit(String field) {
    //Debug.noteln("AFP: Making display bit for " + field);
    if (field.equals("miniConstraints")) {
      //Debug.noteln("AFP: miniConstraint");
      String[] buttons = {"Comprehensive"};
      return new MiniConstraintPanel(this, "Orderings", buttons);
    }
    //patterns are lists, edited as strings
    if (field.equals("pattern")) {
      //Debug.noteln("AFP: pattern");
      return new ThingEditorPanel(this, model.getFieldName(field), 
				  new JTextField());
    }
    else return super.makeDisplayBit(field);
  }

  //-----------------------Action panel things-------------------------------

  /**
   * @return the current setting of the sequence flag
   */
  public boolean isSequenceSet() { 
    Object data = getFieldData("miniConstraints");
    //Debug.noteln("SEP: Mini constraings orderings are " + data);
    return (UIRefinement.O_SEQUENCE.equals(data));
  }

  private String getNameString() {
    Object data = getFieldData("name");
    if (data == null) return "";
    else return data.toString().trim();
  }
  private LList getPatternLList() {
    Object data = getFieldData("pattern");
    return Lisp.elementsFromString((String)data);
  }


  //-----------------------ConstraintEditing things-----------------------


  /** Gets the current UIDomain. */
  public UIDomain getUIDomain() { 
    return parent.getUIDomain();
  }
  /** Gets the current construct. */
  public IXObject getConstruct() { 
    return uiConstruct.getBaseObject(); 
  }
  /** Sets the current construct. */
  public void setConstruct(IXObject construct) {
    setUIConstruct(getUIDomain().getUIObject(construct));
  }
  /** Gets the current UI construct. */
  public UIObject getUIConstruct() { 
    return uiConstruct; 
  }

  public void setUIConstruct(UIObject uiObject) {
    //if (!(uiObject == null) && (uiObject.equals(uiConstruct))) return;
    uiConstruct = uiObject;
    getModel().setObject(uiObject);
  }

  public void setUIConstructOnly(UIObject uiObject) {
    //if (!(uiObject == null) && (uiObject.equals(uiConstruct))) return;
    uiConstruct = uiObject;
    getModel().setObject(uiObject);
  }

  public void loadFromObject() {
    displayModelData();
  }

  public void saveToObject(Object object) {
    UIRefinement uir = (UIRefinement) object;
    if (IVUtil.checkCopyObject("activity", uir, getNameString())) {
      //Debug.noteln("AFoP: copying object");
      uir = new UIRefinement(parent.getUIDomain());
      saveToGivenObject(uir);
      ((ActionEditorPanel)parent).setUIConstructOnly(uir);
    }
    else saveToGivenObject(uir);
  }
    
  private void saveToGivenObject(UIRefinement newUIR) {
    //Debug.noteln("AFoP: saving to refinement", newUIR.getName());
    super.saveToObject(newUIR);
    if (isSequenceSet()) {
      newUIR.setOrderings(((ActionEditorPanel)parent)
			  .makeSequence(newUIR.getNodes()));
    }
    //Debug.noteln("AFoP: saved to refinement", newUIR.getName());
  }

  
}
