/****************************************************************************
 * An abstract class with useful things for building constraint editors.
 *
 * @author Jussi Stader
 * @version 1.0
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 *
 *****************************************************************************
 */

package ix.iview;

import javax.swing.*;       
import java.util.ArrayList;
import java.awt.Color;
import java.awt.Component;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseListener;
import java.awt.event.MouseEvent;
import ix.*;
import ix.util.Debug;
import ix.util.lisp.LList;
import ix.util.lisp.Lisp;
import ix.icore.domain.*;
import ix.iview.domain.*;
import ix.iface.ui.*;

/****************************************************************************
 * An abstract class with useful things for building constraint editors.
 *
 * The editor is implemented as a frame using Boxes and JButtons.  <p>
 *
 * The editor can be created with no arguments or with a given object
 * which will be set as the current object to which the constraint applies.
 * Subsequently, the object in the editor can be set using the setObject
 * method. <p>
 *
 * Concrete editors that extend this class should implement the following 
 * methods:
 * <ul> makeRels() - makes all components needed to specify the constraint.
 * <li> initFromObject() - initialises components with information from the
 *      current object, e.g. lists of children are filled with those of the
 *      current refinement.
 * <li> setFromObject(Object original) - set the constraints using the given
 *      object (used when editing constraints).
 * <li> collectConstraint() - collects user input from the components and 
 *      creates (and returns) a Constraint object from that information.
 *      Also deals with edit vs add behaviour.
 * </ul>
 * The rest can be handled by the generic methods defined in this class.
 * For an example of an implementing class see TPRelationEditor.
 *****************************************************************************
 */
public abstract class ConstraintEditor extends JFrame implements MouseListener
{
  protected JPanel mainPanel = new JPanel();
  protected Component relsBox;
  protected Component butBox;
  //relation items -- to be defined by the concrete class
  //control items
  // protected JCheckBox requiredCheck = new JCheckBox("Required", true);
  protected JButton okBut = new JButton("Ok");
  protected JButton cancelBut = new JButton("Cancel");
  protected JButton moreBut = new JButton("Add More");

  protected JLabel label;

  protected ArrayList constraintListeners = new ArrayList();
  public Object currentObject;
  protected Object original; //used for "edit" rather than "add" behaviour

  
  /**
   * Creates an editor with the title for specifying constraints.
   * Creates space for components for specifying the constraints
   * and creates two buttons (Ok, Cancel)
   */
  public ConstraintEditor(String title, String message) {
    super(title);
    this.setSize(405,175);
    this.mainPanel.setLayout(new BoxLayout(this.mainPanel,BoxLayout.Y_AXIS));
    this.getContentPane().add(this.mainPanel);
    this.mainPanel.setBounds(0,0,400,170);
    relsBox = makeRels();
    butBox = makeButs();
    
    if (message != "") {
      label = new JLabel(message);
      this.mainPanel.add(label);
      mainPanel.add(Box.createGlue());
    }
    this.mainPanel.add(relsBox);
    mainPanel.add(Box.createGlue());
    mainPanel.add(new JSeparator());
    mainPanel.add(butBox);  
    
    this.setVisible(false);
  }
    
  /**
   * As above but with a default title.
   */
  public ConstraintEditor(String message) {
    this("Constraint Editor", message);
  }
    
  /**
   * Creates an editor for specifying time point relations of the given object.
   * As ConstraintEditor() but with a given object.
   * @param object the object whose sub-activities are to be related
   */
  public ConstraintEditor(String title, String message, Object object){
    this(title, message);

    try {
      this.setObject(object);
    } catch (NullPointerException e){
        JOptionPane.showMessageDialog(this, e.getMessage());
        return;
    }
    this.pack();
    this.setVisible(true);
  }
  
  
  /**
   * Makes and adds ok and cancel buttons.
   */
  private Component makeButs(){
    okBut.addMouseListener(this);
    cancelBut.addMouseListener(this);
    moreBut.addMouseListener(this);
    Box butBox = new Box(BoxLayout.X_AXIS);
    butBox.add(Box.createGlue());
    butBox.add(okBut);
    butBox.add(cancelBut);
    butBox.add(moreBut);
    butBox.add(Box.createGlue());
    return butBox;

  }




  public boolean isEditing() {
    return (original != null);
  }

  public void setMessage(String message) {
    if (label != null) label.setText(message);
  }

  /**
   * Sets the object whose components (children) are to be realted.
   * Fills the activity lists of the panel with the children of the object.
   * @param object the object whose children are to be related 
   */
  public void setObject(Object object) throws NullPointerException {
    //Debug.noteln("Object in ConstraintEditor is", object);
    if (object == null) return;
    this.currentObject = object;
    this.initFromObject();
  }

  /**
   * Set this to the original value (non-null) to get edit-bahaviour rather
   * than add-behaviour.
   */
  public void setOriginal(Object original) {
    this.original = original;
  }

  public void start() {
    setVisible(true);
  }
  public boolean start(Object object) {
    try {
      setObject(object); 
      start();
      return true;
    }
    catch (NullPointerException e) {
      JOptionPane.showMessageDialog(this, e.getMessage());
      Debug.noteException(e);
      return false;
    }
  }
  public void start(Object object, Object original) {
    if (start(object)) {
      setOriginal(original);
      setFromObject(original);
      moreBut.setEnabled(false);
    }
  }


  /**
   * Fills the controls with relevant object-sensitive information.
   * To be implemented by the sub-class. E.g. fill ordering editor with the
   * sub-activities of the refinement.
   */


  public void addConstraintListener(ConstraintListener listener){
    this.constraintListeners.add(listener);
  }
  public void removeConstraintListener(ConstraintListener listener){
    while (this.constraintListeners.contains(listener)){
      this.constraintListeners.remove(listener);
    }
  }

  /**
   * Reads the selected relation and let interested parties know.
   * Checks that selections have been made.
   */
  private void processConstraint() {
    Object constraint = collectConstraint();
    //if (constraint == null) return;
    if (this.constraintListeners == null) return;
    for (int i=0; i<this.constraintListeners.size(); i++) {
      ConstraintListener l = 
	(ConstraintListener) this.constraintListeners.get(i);
      l.gotConstraint(this.currentObject, constraint);
    }
  }


  public void mouseClicked(MouseEvent e){
    if (e.getSource() == this.cancelBut) {
      this.closeEditor();
      return;
    }
    if (e.getSource() == this.okBut) {
      this.processConstraint();
      this.closeEditor();
      return;
    }
    if (e.getSource() == this.moreBut) {
      this.processConstraint();
      initFromObject();
      return;
    }
  }
  public void mouseEntered(MouseEvent e){}
  public void mouseExited(MouseEvent e){}
  public void mousePressed(MouseEvent e){}
  public void mouseReleased(MouseEvent e){}

  public void closeEditor() {
    setVisible(false);
    original = null;
    currentObject = null;
    moreBut.setEnabled(true);
  }

  /**
   * Does nothing for the abstract class.
   * For a concrete class this makes all items on the panel used to specify
   * parts of the constraint.
   */
  abstract protected Component makeRels();
  abstract protected void initFromObject();
  abstract protected void setFromObject(Object original);
  abstract protected Object collectConstraint();

}

