/****************************************************************************
 * An editor/viewer frame for domain details.
 *
 * @author Jussi Stader
 * @version 2.0
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 *
 *****************************************************************************
 */

package ix.iview;

import java.util.*;

import java.io.File;
import java.io.FileNotFoundException;
import javax.swing.*;       
import javax.swing.event.*;       
import java.awt.GridBagConstraints;
import java.awt.Component;
import java.awt.Insets;
import java.awt.Dimension;
import java.awt.event.*;
import ix.*;
import ix.util.Debug;
import ix.util.Util;
import ix.util.lisp.*;
import ix.icore.*;
import ix.icore.domain.*;
import ix.icore.domain.event.*;
import ix.iface.ui.*;
import ix.iface.ui.util.*;
import ix.iface.domain.*;
import ix.iview.util.*;
import ix.iview.domain.*;
import ix.iview.domain.event.*;
import ix.iview.igraph.*;

/****************************************************************************
 * An editor/viewer frame for domain details.
 *
 * The editor can be created with a mandatory advanced mode (true for advanced,
 * false for simple) and an optional domain or an optional title.
 * Most of the window buidling goes on in the ConstructFramePanels and their
 * ConstructEditorPanels, except for the menubar and the toolbar.<p>
 *
 * Super-class JFrameMT defines useful menu/toolbar making things.
 * Most of the action processing goes on in the ConstructFramePanels, except
 * for domain specific things (read...) and window things (toggles).<p>
 *
 * Example for using DomainEditor:
 * <PRE><code>
 *   ...
 *   File modelDirectory = new File("domain-library");
 *   boolean isAdvanced = false;
 *   DomainEditor frame = new DomainEditor(isAdvanced);
 *   frame.setModelDirectory(modelDirectory);
 *   frame.setVisible(true);
 *   ...
 * </code></PRE>
 *
 *****************************************************************************
 */
public class DomainEditor extends JFrameMTP implements ChangeListener,
  DomainListener, UIDomainListener, InternalDomainEditor
{

  //---------------------Fields-------------------------------------------

  protected IXAgent agent;
  //------------------------------------Window things
  protected GlobalFramePanel domainPanel;
  protected ActionEditorPanel actionPanel;
  protected AROEditorPanel aroPanel;
  protected GrammarEditorPanel grammarPanel;
  protected AConstructFramePanel lexiconPanel;
  protected AConstructFramePanel testPanel;

  protected OrderingEditor orderingEditor;
  private JFileChooser fileChooser;

  //protected ImageIcon checkIcon = new ImageIcon(imageDir + "ide-cross.gif");
  protected ImageIcon checkIcon = Util.resourceImageIcon("ide-check.gif");
  protected ImageIcon revertOIcon = 
    Util.resourceImageIcon("ide-revert-orig.gif");
  protected ImageIcon modifyIcon = 
    Util.resourceImageIcon("ide-modify-draft.gif");
  protected ImageIcon publishIcon = 
    Util.resourceImageIcon("ide-draft-publish.gif");
  protected ImageIcon simpleIcon = 
    Util.resourceImageIcon("ide-view-minimal.gif");
  protected ImageIcon advancedIcon = 
    Util.resourceImageIcon("ide-view-comprehensive.gif");
  protected ImageIcon graphicalIcon = 
    Util.resourceImageIcon("ide-view-graphical.gif");

  protected HelpFrame helpFrame;
  private String helpFile = "ide-help.html";
  private ImageIcon helpFrameIcon = 
    new ImageIcon(imageDir + "ide-help-icon.gif");

  private JMenu editConstructMenu;
  private JMenu orderMenu;
  /*
  private JMenuItem saveItem;
  private JMenuItem revertItem;
  private JMenuItem revertOItem;
  private JMenuItem newItem;
  private JMenuItem checkItem; // check construct
  private JMenuItem cn; // new construct
  private JMenuItem cs; // save construct
  private JMenuItem exitItem; // exit system
  */
  //private JMenuItem Item;
  private JRadioButtonMenuItem simpleMI;
  private JRadioButtonMenuItem advancedMI;
  private JRadioButtonMenuItem graphicalMI;
  private JButton simpleTB;
  private JButton advancedTB;
  private JButton graphicalTB;
  private JCheckBoxMenuItem defSequenceMI;

  protected IXToolItem publishItem =
    new IXToolItem(fileMenu, toolBar, this, "publish", "Publish draft", 
		   publishIcon, 'P', "Publish draft to application");
  protected IXToolItem insertItem =
    new IXToolItem(fileMenu, this, "insert", "Insert into draft domain...",'I',
		   "Insert all constructs from an existing domain " +
		   "into the draft");
  protected IXToolItem clearItem =
    new IXToolItem(fileMenu, this, "clearDom", "Clear draft domain", 'C',
		   "Clear draft domain");
  protected IXToolItem revertDItem =
    new IXToolItem(fileMenu, this, "revertDom", 
		   "Discard changes to draft (Revert)", 'R',
		   "Discard changes to draft domain (Revert)");

  protected IXToolItem newEditItem =
    new IXToolItem(editMenu, toolBar, this, "newEdit", "New", newIcon, 'N',
		   "Create new item");
  protected IXToolItem revertOItem =
    new IXToolItem(editMenu, toolBar, this, "revertO", "Revert to published", 
		   revertOIcon, "Revert to last published version");
  protected IXToolItem modifyItem =
    new IXToolItem(editMenu, toolBar, this, "modify", "Modify draft", 
		   modifyIcon, 'M', "Note changes from panel into draft");
  protected IXToolItem simpleItem =
    new IXToolItem(viewMenu, toolBar, this, "minimal", "Minimal view", 
		   "Minimal", simpleIcon, "Switch to minimal view");
  protected IXToolItem advancedItem =
    new IXToolItem(viewMenu, toolBar, this, "comprehensive", 
		   "Comprehensive view", "Comprehensive", advancedIcon, 
		   "Switch to comprehensive view");
  protected IXToolItem graphicalItem =
    new IXToolItem(viewMenu, toolBar, this, "graphical", "Graphical view", 
		   "Graphical", graphicalIcon, "Switch to graphical view");

  protected IXToolItem checkItem =
    new IXToolItem(toolsMenu, toolBar, this, "check", "Check", checkIcon, 'C',
		   "Check the current construct");
  /*
  protected IXToolItem Item =
    new IXToolItem(editMenu, toolBar, this, "", "", Icon, '',
		   "");
   */

  //------------------------------------Modes and toggles
  private String originalStyle = "tabbed";
  private boolean advanced;
  private boolean standAlone = false;
  private ImageIcon frameIcon = new ImageIcon(imageDir + "ide-icon.gif");
  private ImageIcon logo = new ImageIcon(imageDir + "ide-logo.gif");
  private String title = "I-X Domain Editor";

  //------------------------------------"Globals"
  protected File modelDirectory = new File("domain-library");
  /** The current domain in the editor. May be null */
  //public Domain currentDomain = new Domain();
  public UIDomain uiDomain = new UIDomain(new Domain(), this);

    
  private boolean adjustingDomain = false;

  //---------------------Constructors-------------------------------------

  /**
   * Makes the editor with the default (simple) advanced mode.
   * Subsequently, the domain can be set with the setDomain method or it can
   * be loaded via the menu. The default window title is "IX - Domain Editor".
   */
  public DomainEditor(){this(false, true);}


  /**
   * Makes the editor with the given advanced mode.
   * Subsequently, the domain can be set with the setDomain method or it can
   * be loaded via the menu. The default window title is "IX - Domain Editor".
   *
   * @param isAdvanced advanced mode: true for advanced, false for simple
   */
  public DomainEditor(boolean isAdvanced, boolean standAlone){
    super("tabbed");
    //this.setPreferredSize(new Dimension(405,605)); //not defined for frames
    setSize(650,550);
    setVisuals();

    this.standAlone = standAlone;
    this.advanced = isAdvanced;
    makeDefaultMenuToolbar();
    makeToolbar();
    // make a gap - after this there may be panel-specific buttons

    actionPanel = new ActionEditorPanel(this,isAdvanced);
    //aroPanel = new AROEditorPanel();
    //aroPanel.add(new JLabel("Editing of activity relatable objects (AROs)" +
    //		    " is not yet supported."));
    domainPanel = new GlobalFramePanel(this);
    grammarPanel = new GrammarEditorPanel(this);
    //testPanel = new TestEditorPanel();
    //lexiconPanel = new SimpleEditorPanel();
    //lexiconPanel.add(new JLabel("Editing of the lexicon" +
    //		       " is not yet supported."));
    addPanels();
    if (!isAdvanced) setPanelStyle("simple");

    fillWindowsMenu();
    setCurrentPanel(actionPanel);

    advancedMI.setSelected(isAdvanced);
    simpleMI.setSelected(!isAdvanced); 

    orderingEditor = actionPanel.orderingEditor;

    this.addWindowListener(new IXWindow());

    //setDomain(new Domain()); //heads off into a loop somehow! set in field.
    setUIDomainListeners(getUIDomain());
    populateEditMenu();
    validate(); //***
    }

  /**
   * Makes the editor with the given domain - simple, stand-alone by default.
   * The default window title is "IX - Domain Editor".
   * 
   * @param domain the domain to use in the editor
   */
  public DomainEditor(Domain domain){
    this(domain, false, true);
  }

  /**
   * Makes the editor with the given advanced mode and load the given domain.
   * The default window title is "IX - Domain Editor".
   * 
   * @param domain the domain to use in the editor
   * @param isAdvanced advanced mode: true for advanced, false for simple
   */
  public DomainEditor(Domain domain, boolean isAdvanced, boolean standAlone){
    this(isAdvanced, standAlone);
    this.setDomain(domain);
  }

  /**
   * Makes the editor with the given domain - simple by default.
   * The default window title is "IX - Domain Editor".
   * 
   * @param domain the domain to use in the editor
   */
  public DomainEditor(IXAgent theAgent, Domain domain){
    this(domain, false, false);
    this.agent = theAgent;
    this.title = theAgent.getAgentSymbolName().toString() + " Domain Editor";
    this.standAlone = false;
    exitItem.hide();
    setVisuals();
  }

  /** Adds all panels to the frame. */
  private void addPanels() {
    addPanel((JPanel)domainPanel, "Domain", 
	     "Viewer/editor for global domain constraints and issues");
    addPanel((JPanel)actionPanel, "Activity", "Activity viewer/editor");
    addPanel((JPanel)grammarPanel, "Grammar", "Grammar viewer/editor");
    /*
    addPanel((JPanel)testPanel, "Test", "Test panel");
    addPanel((JPanel)graphPanel, "Expansion", 
    	  "Graphical expansion viewer/editor");
    addPanel((JPanel)aroPanel, "Object",
         "Activity Relatable Objects viewer/editor." +
         "This will cover objects, resources, and performers of the" +
         " domain and include specification of types and object" +
         " constraints.");
    addPanel((JPanel)lexiconPanel, "Lexicon", "Lexicon viewer/editor");
    */
  }

  /** Clears all panels of the frame. */
  private void clearPanels() {
    domainPanel.clear();
    actionPanel.clear();
    grammarPanel.clear();
    //testPanel.clear();
    //aroPanel.clear();
    //lexiconPanel.clear();
  }

  /** Closes all panels of the frame. */
  private void closePanels() {
    Debug.noteln("DE: Closing panels");
    domainPanel.beforeClosing();
    actionPanel.beforeClosing();
    grammarPanel.beforeClosing();
    //testPanel.beforeClosing();
    //aroPanel.beforeClosing();
    //lexiconPanel.beforeClosing();
  }


  /**
   * Changing current panel from old to new.
   */
  protected boolean currentPanelChanging(JPanel oldPanel, JPanel newPanel) {
    if (oldPanel == null) 
      return ((AConstructFramePanel)newPanel).beforeOpening();
    //do nothing but make sure the panel is changed to (important for cancel)
    if (oldPanel.equals(newPanel)) return true; 
    //Debug.noteln("DE: current panel changing");
    AConstructFramePanel oldACFP = (AConstructFramePanel)oldPanel;
    AConstructFramePanel newACFP = (AConstructFramePanel)newPanel;
    if (oldACFP.beforeClosing()) {
      if (newACFP.beforeOpening()) {
	//Debug.noteln("DE: current panel changing done");
	return true;
      }
    }
    //Debug.noteln("DE: CANCEL current panel changing");
    return false;
  }

  protected void setVisuals() {
    setTitle(title);
    setIconImage(frameIcon.getImage());
    //testPanel.setAgent(agent);
  }

  public void ensureHelpFrame(){
    if (helpFrame == null){
      helpFrame = new HelpFrame(helpFile);
      helpFrame.setIconImage(helpFrameIcon.getImage());
      helpFrame.setVisible(false);
    } 
  }
  //---------------------Field access-------------------------------------  

  /**
   * @return the current setting of the advanced flag
   */
  public boolean isAdvancedSet(){ return this.advanced;}
  /**
   * Sets the editor to advanced (true) or simple (false) mode.
   * If the state of the menu item does not correspond to the new setting,
   * it is changed.
   * If the editor is already in the given mode, nothing is done, except 
   * ensuring that the menu item is set to the corresponding state.
   * Currently, nothing else changes in the frame, so the command is passed on 
   * to the ActionEditorPanel.
   */
  public void setAdvanced(boolean newAdvanced){ 
    //Debug.noteln("setting advanced in domain editor from " + 
    //	 advanced + " to " + newAdvanced);
    if ((advancedMI != null) && (advancedMI.isSelected() != newAdvanced)) {
      //command came from elsewhere, so make menu item correspond to new value
      advancedMI.setSelected(newAdvanced); 
      simpleMI.setSelected(!newAdvanced); 
    }
    if (newAdvanced != this.advanced) { //not already in the desired state
      this.advanced = newAdvanced;
      this.enableAdvanced();
    }
    // pass it on to panel
    this.actionPanel.setAdvanced(this.advanced); 
  }

  /**
   * If the state of the menu item does not correspond to the new setting,
   * it is changed.
   */
  public void setGraphical(){ 
    if ((graphicalMI != null) && !graphicalMI.isSelected()) {
      //command came from elsewhere, so make menu item correspond to new value
      graphicalMI.setSelected(true); 
      simpleMI.setSelected(false); 
      advancedMI.setSelected(false); 
    }
  }

  /**
   * Sets the directory used to load/save models.
   * This method just notes the given value which is used when needed
   * for the file selectors for loading and saving models
   */
  public void setModelDirectory(File directory){
    this.modelDirectory = directory;
  }
  /**
   * @return the directory used to load/save models.
   */
  public File getModelDirectory(){
    return this.modelDirectory;
  }



  //---------------------Toggles-----------------------------------------

  private void enableAdvanced(){ 
    Dimension size = getSize();
    //change whether default sequence is enabled
    //ensure the relevant advanced bit is selected in the menu
    if (defSequenceMI != null) {
      if (this.advanced) {
	this.defSequenceMI.setEnabled(false);
	setOriginalStyle();
      }
      else {
	this.defSequenceMI.setEnabled(true);
      }
    }
    this.setSize(size);
    validate();
  }

  /**
   * If the editor is brought up in simple mode, style is changed to 
   * simple too - this should move to tabbed style when first advanced
   * or graphical is selected (do only once!)
   */
  public void setOriginalStyle() {
    if ((originalStyle != "") &&  !(getPanelStyle().equals(originalStyle))) {
      setPanelStyle(originalStyle);
      originalStyle = "";
    }
  }

        
  //---------------------Window building------------------------------------

  public void adaptDefaultMenuToolbar() {
    adaptMenuToolbarToAction();
  }
  private void adaptMenuToolbarToAction() {
    optionsMenu.addSeparator();
    defSequenceMI = makeCheckMenuItem(optionsMenu,"defSequence",
				   "New Sub-Activities are in Sequence");
    defSequenceMI.setState(true);
    //defSequenceMI.setVisible(false);

    orderMenu = this.makeMenu(editMenu, "Set Constraints");
    makeMenuItem(orderMenu,"makeSequence","Put Sub-Activities in Sequence");
    makeMenuItem(orderMenu,"makeParallel","Put Sub-Activities in Parallel");
    //orderMenu.setVisible(false);
    populateEditMenu();    

    viewMenu.addSeparator();
    ButtonGroup bg = new ButtonGroup();
    simpleMI = simpleItem.makeRadioItem();
    advancedMI = advancedItem.makeRadioItem();
    graphicalMI = graphicalItem.makeRadioItem();
    bg.add(simpleMI);
    bg.add(advancedMI);
    bg.add(graphicalMI);
    /*
    toolBar.addSeparator();
    toolBar.addSeparator();
    toolBar.addSeparator();
    toolBar.addSeparator();
    */
    simpleItem.showItem();
    simpleItem.showButton();
    advancedItem.showItem();
    advancedItem.showButton();
    graphicalItem.showItem();
    graphicalItem.showButton();

    //graphicalMI.setEnabled(false);

  }
  /**
   * Makes a the file menu and toolbar buttons. Replaces that of the superclass
   * Menu items: new, open, save, save as, S, Options, S, close, exit
   * Toolbar buttons: new, open, save
   */
  public JMenu makeFileMenuTool(JComponent parentMenu) {
    //fileMenu = this.makeMenu(parentMenu,"File");
    
    openItem.setItemLabel("Open as draft domain...");
    openItem.setToolTip("Open an existing file as a draft domain");
    openItem.showItem();
    insertItem.showItem();
    clearItem.showItem();
    revertDItem.showItem();
    saveItem.setButtonLabel("Save draft");
    saveItem.setItemLabel("Save draft to file");
    saveItem.showItem();
    publishItem.showItem();
    //publishItem.showButton();
    
    //toolBar.addSeparator();
    fileMenu.addSeparator();
    makeOptionsMenuTool(fileMenu);

    makeFileMenuToolExits(fileMenu);
    return fileMenu;
  }

  /**
   * Makes default close/exit menu options and and toolbar buttons.
   * Menu items: S, close, exit
   * Toolbar buttons: none
   */
  public void makeFileMenuToolExits(JMenu fileMenu) {
    fileMenu.addSeparator();
    /*
    makeMenuItem(fileMenu,"close","Close");
    if (standAlone) exitItem = makeMenuItem(fileMenu,"exit","Exit");
    */
    closeItem.showItem();
    if (standAlone) exitItem.showItem();
  }
  /**
   * Makes an edit menu
   * <nl><em>Menu items:</em> edit construct
   * <nl><em>Toolbar buttons:</em> none
   */
  public JMenu makeEditMenuTool(JComponent parentMenu) {
    //editMenu = makeMenu(parentMenu,"Edit"); 
    //Debug.noteln("DomainEditor: Setting edditConstruct to Edit Construct");

    newEditItem.setLabel("New construct");
    newEditItem.setToolTip("New construct");
    newEditItem.show();
    editConstructMenu = makeMenu(editMenu,"Edit Construct");
    editConstructMenu.setEnabled(false);

    modifyItem.showItem();
    modifyItem.showButton();
    revertItem.showItem();
    revertItem.showButton();
    revertOItem.showItem();
    editMenu.addSeparator();
    //saveItem.setEnabled(false);
    return editMenu;
  }
  /**
   * Makes a tools menu and toolbar buttons.
   * <nl><em>Menu items:</em> none
   * <nl><em>Toolbar buttons:</em> none
   */
  public JMenu makeToolsMenuTool(JComponent parentMenu) {
    //toolsMenu = makeMenu(parentMenu,"Tools");
    /*
    checkItem = makeMenuItem(toolsMenu,"checkConstruct",
			     "Check the current construct");
    */
    checkItem.showItem();
    checkItem.setEnabled(false);
    return toolsMenu;
  }

  protected void makeToolbar() {
    toolBar.removeAll();

    /*
    newEditItem.showButtonText(true);
    newEditItem.showButton();
    revertItem.showButtonText(true);
    revertItem.showButton();
    modifyItem.showButtonText(true);
    modifyItem.showButton();
    toolBar.addSeparator();

    saveItem.showButtonText(true);
    saveItem.showButton();
    publishItem.showButtonText(true);
    publishItem.showButton();
    */

    saveItem.showButtonText(true);
    saveItem.showButton();
    publishItem.showButtonText(true);
    publishItem.showButton();
    toolBar.addSeparator();

    newEditItem.showButtonText(true);
    newEditItem.showButton();
    revertItem.showButtonText(true);
    revertItem.showButton();
    modifyItem.showButtonText(true);
    modifyItem.showButton();


    toolBar.addSeparator();
    toolBar.addSeparator();
    toolBar.addSeparator();
    toolBar.addSeparator();
    toolBar.addSeparator();
    //showButtonTexts(true);
  }

  public void showButtonTexts(boolean show) {
  }
 
  /**
   * Called whenever the current panel is updated
   */
  protected void adaptMenusToPanel(JPanel panel){  
    if (panel == null) return;

    //things to do for all construct panels
    String name = panel.getName();
    /*
    newItem.setText("New " + name);
    saveItem.setText("Note " + name + " into draft domain");
    revertItem.setText("Revert " + name + " to draft");
    revertOItem.setText("Revert " + name + " to published");
    checkItem.setText("Check " + name);
    //Debug.noteln("DomainEditor: Setting edditConstruct to Edit " + name);
    */
    editConstructMenu.setText("Edit " + name);
    newEditItem.setLabel("New " + name);
    modifyItem.setItemLabel("Modify draft (Note " + name + " into draft)");
    checkItem.setItemLabel("Check current " + name);
    
    newEditItem.setEnabled(true);
    modifyItem.setEnabled(true);
    checkItem.setEnabled(true);
    populateEditMenu();

    //things to do for each individual panel (have an if-else per panels)
    if (panel.equals(actionPanel)) {
      //switch things on
      optionsMenu.add(defSequenceMI);
      /*
      viewMenu.add(simpleMI);
      viewMenu.add(advancedMI);
      viewMenu.add(graphicalMI);
      //toolBar.add(simpleTB);
      //toolBar.add(advancedTB);
      //toolBar.add(graphicalTB);
      simpleTB.setEnabled(true);
      simpleTB.setVisible(true);
      advancedTB.setEnabled(true);
      advancedTB.setVisible(true);
      graphicalTB.setEnabled(true);
      graphicalTB.setVisible(true);
      */
      simpleItem.show();
      advancedItem.show();
      graphicalItem.show();
      editMenu.add(orderMenu);
    }
    else {      //switch things off
      optionsMenu.remove(defSequenceMI);
      /*
      viewMenu.remove(simpleMI);
      viewMenu.remove(advancedMI);
      viewMenu.remove(graphicalMI);
      //toolBar.remove(simpleTB);
      //toolBar.remove(advancedTB);
      //toolBar.remove(graphicalTB);
      simpleTB.setEnabled(false);
      simpleTB.setVisible(false);
      advancedTB.setEnabled(false);
      advancedTB.setVisible(false);
      graphicalTB.setEnabled(false);
      graphicalTB.setVisible(false);
      */
      simpleItem.hide();
      advancedItem.hide();
      graphicalItem.hide();
      editMenu.remove(orderMenu);
    }
  }

  /**
   * Fill the edit menu with constructs from the domain.
   */
  protected void populateEditMenu(){
    AConstructFramePanel panel = (AConstructFramePanel)getCurrentPanel();
    //Debug.noteln("DomainEditor: populating edit menu for panel" );
    //make the list of objects to go in
    if (panel != null) {
      Collection constructs = panel.getAllConstructs();
      List elements;
      if (constructs != null) {
	constructs = IVUtil.sortNamedCollection(constructs);
	elements = new ArrayList(constructs);
      }
      else  elements = new ArrayList();
      UIUtil.populateMenu(this,editConstructMenu,"editConstruct",elements);
    }
  }


  //  public void setVisible(boolean v) {
  //  }


  //---------------------Public Services--------------------------------------

  /**
   * The name of this is a hangover from something Jeff did. This is meant
   * to set the given refinement into the editor ready to be changed.
   */
  public void saveExpansion(Refinement data) {
    setConstruct(data);
  }


  /**
   * Sets the given construct as the current one and displays it for editing.
   * All work is done in the current panel.
   *
   * @param construct the original domain object to be viewed/edited
   */
  public void setConstruct(IXObject construct) {
    //if (Refinement.class.isInstance(construct))
    //  setCurrentPanel(actionPanel);
    setUIConstruct(uiDomain.getUIObject((IXObject)construct));
  }
  /**
   * Sets the given UIObject as the current one and displays it for editing.
   * All work is done in the current panel.
   *
   * @param construct the UI object to be viewed/edited
   */
  public void setUIConstruct(UIObject construct) {
    //find a suitable panel for the object and set it
    if (UIRefinement.class.isInstance(construct))
      setCurrentPanel(actionPanel);
    else if (UIDomain.class.isInstance(construct))
      setCurrentPanel(domainPanel);
    //set the construct in the current panel
    AConstructFramePanel panel = (AConstructFramePanel)getCurrentPanel();
    if (panel != null) panel.setUIConstruct((UIObject)construct);
  }

  /**
   * Reads the descriptions in the given domain file.
   * Displays a message if the file cannot be found.
   * Sets the domain as current in this.
   *
   * @param domainFile the file from which to read the domain
   */
  public void setDomain(File domainFile) {
    //Debug.noteln("DE: Setting domain from file", domainFile.toString());
    try {
      DomainParser parser = DomainParser.makeParser(domainFile.getPath());
      this.setDomain(parser.readDomain(uiDomain.getDomain()));
    }
    catch (FileNotFoundException e) {
      JOptionPane.showMessageDialog(this, "Can't find file " + 
				    Util.quote(domainFile.getPath()),
				    "File not found", 
				    JOptionPane.ERROR_MESSAGE);}
  }

  /**
   * Sets the given domain as current both in this frame and the domain
   * editor panel.
   *
   * @param domain the domain to use
   */
  public void setDomain(Domain domain) {
    if (adjustingDomain) return;
    adjustingDomain = true;
    //Debug.noteln("DE: Setting domain", domain.toString());
    if (domain.equals(getDomain())) return;
    UIDomain newUID = uiDomain.setDomain(domain);
    //Debug.noteln(" set uiDomain", newUID.print());
    setUIDomain(newUID);
    //populateEditMenu();  //done by setUIDomain
    adjustingDomain = false;
  } 

  /**
   * Sets the given UIDomain as current both in this frame and the domain
   * editor panel.
   *
   * @param uiDomain the UIDomain to use
   */
  public void setUIDomain(UIDomain uiDomain) {
    if (this.uiDomain.equals(uiDomain)) return;
    this.uiDomain = uiDomain;
    setUIDomainListeners(uiDomain);
    //global editing panel is added during init of global frame panel
    //populateEditMenu();
  } 
  private void setUIDomainListeners(UIDomain uiDomain) {
    //Debug.noteln("DE: Adding listeners to UI domain", uiDomain.getName());
    uiDomain.addUIDomainListener(this);
    uiDomain.addUIDomainListener(grammarPanel);
    uiDomain.addUIDomainListener(actionPanel);
    UIDomainEvent event = new UIDomainEvent(uiDomain);
    //this seems to be needed - don't know why
    uiDomain.fireDomainSet(uiDomain.getDomain());
  }

  public UIDomain getUIDomain() { return uiDomain; }
  public Domain getDomain() { return uiDomain.getDomain(); }

  /**
   * Sets the current panel and sets the domain in the panel
   *
   * @param domain the domain to use
   */
  public void updateCurrentPanel(AConstructFramePanel panel) {
    super.updateCurrentPanel(panel);
    //Debug.noteln("DE: Current panel is now", getACurrentPanel().getName());
    //Debug.noteln("DE: populating edit menu");
    populateEditMenu();
  } 

  /**
   * Sets the toggle for default sub-activity sequencing behaviour.
   *
   * @param on true: put new sub-activities in sequence, false: put them in
   *   parallel.
   */
  public void setDefaultSequence(boolean on){
    if (defSequenceMI != null) defSequenceMI.setState(on);
    this.actionPanel.setDefaultSequence(on);
  }


  //---------------------Private Services--------------------------------------

  //---------------------User Services---------------------------------------

  /**
   * Clears the current domain.
   * Gets confirmation from the user if the domain is not empty.
   */
  protected void userClearDomain() {
    int y = JOptionPane.YES_OPTION;
    if (!uiDomain.isEmpty()) {
      String[] message = {"This will clear all constructs in the draft.",
			  "Are you sure you want to do this?"};
      y = JOptionPane.showConfirmDialog(this, message, "Confirm",
					JOptionPane.YES_NO_OPTION);
    }
    if (y == JOptionPane.YES_OPTION) {
      uiDomain.clearDraft();
      this.clearPanels();
      //***should be a domain event
      //((GlobalFormPanel)domainPanel.formPanel).domainCleared(); 
      this.setUIDomain(uiDomain);  //make changes take effect
    }
  }
  /**
   * Reverts the whole draft domain to the last published (or original) one.
   * Gets confirmation from the user if the domain has changed.
   */
  protected void userRevertDomain() {
    //UIUtil.notImplemented(null, "Reverting the whole domain to original");

    int y = JOptionPane.YES_OPTION;
    if (uiDomain.hasChangedConstructsFromOriginal()) {
      String[] message = {"This will clear all changes to the draft" +
			  " (including open/insert)",
			  "since it was last published.",
			  "If the domain has not been published yet," +
			  " it will revert to the original domain.",
			  "Are you sure you want to do this?"};
      y = JOptionPane.showConfirmDialog(this, message, "Confirm",
					JOptionPane.YES_NO_OPTION);
    }
    if (y == JOptionPane.YES_OPTION) {
      this.clearPanels();
      uiDomain.revertToOriginal();
      this.setUIDomain(uiDomain);  //make changes take effect
    }

  }
  /**
   * Gets a domain file name, reads it, and replaces the contents of the
   * current domain with the read ones.
   */
  protected void userOpenDomain() {
    int y = JOptionPane.YES_OPTION;
    if (!uiDomain.isEmpty()) {
      String[] message = {"This will clear all current domain constructs.",
			  "Are you sure you want to do this?",
			  "(Use 'Insert domain' to add to the current" +
			  " constructs)"};
      y = JOptionPane.showConfirmDialog(this, message, "Confirm",
					JOptionPane.YES_NO_OPTION);
    }
    if (y == JOptionPane.YES_OPTION) 
      uiDomain.loadDomain();
  }

  /**
   * Gets a domain file name and reads it.
   */
  protected void userInsertDomain() {
    uiDomain.insertDomain();
    //this.setDomain(uiDomain.getDomain());
  }
  protected void userSaveDomain() {
    //***this first line is a hack - should just noteConstruct
    ((AConstructFramePanel)getACurrentPanel()).beforeClosing();
    uiDomain.saveDomain();
  }
  protected void userSaveDomainAs() {
    ((AConstructFramePanel)getACurrentPanel()).beforeClosing();
    uiDomain.saveDomain(null);
  }
  protected void userPublishDomain() {
    ((AConstructFramePanel)getACurrentPanel()).beforeClosing();
    uiDomain.publishToDomain();
  }

  public void refinementAdded(RefinementEvent e) {
    //Debug.noteln("DomainEditor: got refinement-added event");
    populateEditMenu();
  }
  public void constructAdded(RefinementEvent e) {
    //Debug.noteln("DomainEditor: got construct-added event");
    populateEditMenu();
  }
  public void constructAdded(UIDomainEvent e) {
    //Debug.noteln("DomainEditor: got construct-added event");
    populateEditMenu();
  }
  public void constructEdited(UIDomainEvent e) {
    //Debug.noteln("DomainEditor: got construct-edited event");
    populateEditMenu();
  }
  public void constructRemoved(UIDomainEvent e) {
    //Debug.noteln("DomainEditor: got construct-removed event");
    populateEditMenu();
  }
  public void domainCleared(UIDomainEvent e) {
    //Debug.noteln("DomainEditor: got domain-cleared event");
    populateEditMenu();
  }
  public void domainEdited(UIDomainEvent se) {
    //Debug.noteln("DomainEditor: got domain-edited event");
    populateEditMenu();
  }
  public void domainSet(UIDomainEvent se) {
    Debug.noteln("DomainEditor: got domain-set event");
    populateEditMenu();
  }

  //---------------------Actions-------------------------------------------

  /**
   * Implements the ActionListener interface.
   * If this cannot recognise the event, it passes it to the current 
   * AConstructFramePanel. If the panel does not recognise the event either,
   * this passes it to its superclass (JFrameMTP, which will pass the
   * event to the current panel first). <p>
   *
   * Catch specific menu/toolbar events here. Let the panels process their
   * own events, let the superclass process panel-swapping events (inc. style)
   * and, via its superclass, general window events (close, exit, ...)
   */
  public void actionPerformed(ActionEvent event) {
      Object object = event.getSource();
      String command = ((AbstractButton)object).getActionCommand();
      //Debug.noteln("action object is",object);
      Debug.noteln("Domain Editor: action command is",command);
      if (command == "close") this.userClose();
      else if (command == "clearDom") this.userClearDomain();
      else if (command == "open") this.userOpenDomain();
      else if (command == "revertDom") this.userRevertDomain();
      else if (command == "insert") this.userInsertDomain();
      else if (command == "save") this.userSaveDomain();
      else if (command == "saveAs") this.userSaveDomainAs();
      else if (command == "publish") this.userPublishDomain();

      //else if (command == "toolbar") 
      //  super.toggleToolbar(((JCheckBoxMenuItem)object).getState());

      else if (command == "manual") this.userManual(); 
      else if (command == "help") this.userHelp(); 

      else {
	//see whether the super classes JFrameMT(Panels) can do it
	super.actionPerformed(event);
      }
  }

  private boolean userCheckConstruct() {
    AConstructFramePanel panel = (AConstructFramePanel)getCurrentPanel();
    if (panel != null) return panel.checkConstruct();    
    else return true;
  }

  public void userClose() {
    if (helpFrame != null) helpFrame.setVisible(false);
    closePanels();
    if (standAlone) System.exit(0); 
    else setVisible(false);
  }

  private void userManual() {
    userHelp();
  }
  
  private void userHelp() {
    ensureHelpFrame();
    helpFrame.setVisible(true);
  }
  

  //---------------------Inner Classes----------------------------------------
    
  class IXWindow extends java.awt.event.WindowAdapter
  {
    public void windowClosing(WindowEvent e) { 
      userClose();
    }	  
  }
  

  //---------------------Stand Alone----------------------------------------

  static public void main(String args[])
  {
    File modelDirectory = new File("domain-library"); 
    //set to true to make an advanced editor
    boolean isAdvanced = false;
    
    //make a stand-alone editor
    DomainEditor frame = new DomainEditor(isAdvanced, true);
    frame.setModelDirectory(modelDirectory);
    
    frame.setVisible(true);
  }


}
