/****************************************************************************
 * Mirrors the Domain for safe editing and adds UI things and change management
 *
 * @author Jussi Stader
 * @version 2.0
 * Copyright: (c) 2002, AIAI, University of Edinburgh
 *
 *****************************************************************************
 */

package ix.iview.domain;

import java.util.*;
import javax.swing.JOptionPane;

import java.io.File;
import ix.iview.*;
import ix.iview.util.*;
import ix.iview.domain.event.*;
import ix.icore.domain.event.*;
import ix.icore.*;
import ix.icore.domain.*;
import ix.iface.domain.*;
import ix.iface.ui.util.*;
import ix.util.*;
import ix.util.match.*;
import ix.util.lisp.*;

/**
 * Mirrors the Domain for safe editing and adds UI things and change
 * management.
 * A UIDomain maintains the link to a base domain. It keeps track of UIObjects
 * that have been edited/added/removed and can publish these changes in the 
 * base domain. UIDomain also contains a link to the DomainEditor which is
 * the one that keeps track of which uiDomain is current.
 */

public class UIDomain implements UIObject, Named, DomainListener {
  private Domain currentDomain;
  private Domain draftDomain;
  private File domainFile;

  public boolean loading = false;
  private Domain saving = null;

  public String name = "Unnamed UI Domain";
  protected String baseReference = ""; /** name of the original domain */
  public Annotations annotations;

  /** list of graphs that apply to Refinements; should be UIRefinement slot */
  public HashMap graphs = new HashMap();
  /** list of UIObjects added to the UIDomain (unless removed).*/
  private HashSet added = new HashSet();
  /** list of UIObjects edited from the domain (unless added). */
  private HashSet edited = new HashSet();
  /**
   * list of UIObjects removed from the domain (unless added).
   * Contains UIObjects and IXObjects.
   */
  private HashSet removed = new HashSet();
  /** domain object to uiObject */
  private HashMap touched = new HashMap();
  /** name to uiObject */
  //  private HashMap uiObjects = new HashMap();

  private HashMap classes = new HashMap();

  private HashSet listeners = new HashSet();

  private DomainEditor editor;

  public UIDomain(Domain baseDomain, DomainEditor editor) {
    super();
    this.editor = editor;
    currentDomain = baseDomain;
    draftDomain = cloneDomain(baseDomain);
    loadFromDomain();
    classes.put(Refinement.class, UIRefinement.class);
    classes.put(Domain.class, UIDomain.class);
    //classes.put(.class, UI.class);
    touched.put(baseDomain, this);
    setBaseReference(baseDomain.getName());
    currentDomain.addDomainListener(this);
  }

  public UIDomain(DomainEditor editor) {
    this(new Domain(), editor);
  }

  public UIDomain(String name, DomainEditor editor) {
    this(new Domain(), editor);
    this.name = name;
  }

  
  //--------------------------Handling changes--------------------------------
  public void addConstruct(UIObject object) {
    if (!object.isEmpty()) {
      //Debug.noteln("UID: addConstruct", object.getName());
      //Debug.noteln(" Added now: ", UIUtil.listToDisplay(added));
      if (UIRefinement.class.isInstance(object)) {
	boolean toAdd = handleOverwrite(object);
	if (toAdd) {
	  addConstruct(object);
	  return;
	}
	else fireUIConstructEdited(object);
      }
      if (added.add(object)) fireUIConstructAdded(object);
      //else Debug.noteln("UID--object already added");
    }
  }
  /**
   * @returns true if the object needs to be updated/added again, false if
   * the object should be processed as normal
   */
  private boolean handleOverwrite(UIObject uiObject) {
    //Debug.noteln("UID: handleOverwrite");
    Object oldObject = getNamedRefinement(uiObject.getName());    
    if ((oldObject != null) && //there is an old object
	(!oldObject.equals(uiObject)) && //not the new one
	(!removed.contains(oldObject))) { //not removed
      String[] message = {"There already is a refinement called " + 
			  uiObject.getName() + ". Do you want to ",
			  "Replace the existing refinement? (Yes)",
			  "Change the name of the new refinement? (No)"};
      int y = JOptionPane.showConfirmDialog(null, message, 
					    "WARNING: Name clash",
					    JOptionPane.YES_NO_OPTION);
      if (y == JOptionPane.YES_OPTION) {
	//remove the old one and do original plan with the new one
	if (IXObject.class.isInstance(oldObject)) 
	  removeConstruct((IXObject)oldObject);
	else removeConstruct((UIObject)oldObject);
	return true;
      }
      else {
	//change the name then do original plan with the new one
	String answer = 
	  JOptionPane.showInputDialog("Please enter the new name");
	uiObject.setName(answer);
	return true;
      }
    }
    return false;
  }
  
  public void updateConstruct(UIObject object) {
    if (!object.isEmpty()) {
      //Debug.noteln("UID: updateConstruct");
      if (UIRefinement.class.isInstance(object)) {
	boolean toAdd = handleOverwrite(object);
	if (toAdd) { 
	  updateConstruct(object);
	  return;
	}
      }
    }
    //if it has been added, ignore edits, else remember them
    if (object.hasOriginal()) {
      if (!added.contains(object)) edited.add(object);
      //Debug.noteln("UID: Edited construct", object.getName());
      //back to same as original, forget changes!
      if (!object.hasChangedFromOriginal()) edited.remove(object);
      fireUIConstructEdited(object);
    }
    else addConstruct(object);
  }

  public void removeConstruct(UIObject object) {
    //if it has been added, forget this and be done (do fire removed)
    if (added.contains(object)) added.remove(object);
    //if it is empty and has no original - just forget it.(don't fire removed)
    else if (!object.hasOriginal() && object.isEmpty()) return;
    //else forget all edits (if any) and remember that it is removed
    else {
      edited.remove(object);
      removed.add(object);
    }
    fireUIConstructRemoved(object);
  }

  public void removeConstruct(IXObject object) {
    UIObject uiObject = findUIObject(object);
    if (uiObject == null) 
      removed.add(object);
    else removeConstruct(uiObject);
  }

  /** 
   * Finds out whether the domain has changed from the given one.
   * The domain has not changed if the domain's own fields (name) have not
   * changed.
   */
  public boolean hasChanged(Domain domain) {
    if (((name == "") || (name == null)) && (annotations == null) 
	&& ((domain == null) || domain.isEmpty())) 
      //was and is empty
      return false;
    else if (domain == null) {
      Debug.noteln("UID has changed: was empty, not now");
      return true;
    }      
    else if (!nameHasChanged(domain.getName())) {
      //same name - check annotations
      if (((annotations == null) && (domain.getAnnotations() == null))
	  || annotations.equals(domain.getAnnotations())) 
	return false;
      else {
	Debug.noteln("UID has changed annotations");
	return true;
      }
    }
    else {
      //Debug.note("UID has changed name from " + domain.getName());
      //Debug.noteln(" to", name + ".");
      return true;
    }
  }
  /** 
   * Finds out whether the domain has changed from the draft.
   */
  public boolean hasChanged() {
    return hasChanged(draftDomain);
  }
  /** 
   * Finds out whether the domain has changed from the published/original one.
   */
  public boolean hasChangedFromOriginal() {
    return hasChanged(currentDomain);
  }


  public boolean nameHasChanged(Object nameThing) {
    String original = (String) nameThing;
    if ((((name == null) || name.equals("")) &&
	 ((original == null) || (original.equals("")))) ||
	((name != null) && name.equals(original)))
      return false;
    else return true;
  }
  

  /** 
   * Finds out whether anything in the domain (inc. constructs) have changed
   * from the given one.
   * The domain has not changed if there are no added/edited/removed notes
   * and the domain's own fields have not changed.
   */
  public boolean hasChangedAll(Domain domain) {
    if (hasChanged(domain)) return true;  //domain itself changed
    else return hasChangedConstructs(domain);
  }

  /** 
   * Finds out whether any constructs in the domain have changed
   * from the given one.
   * The domain has not changed if there are no added/edited/removed notes
   */
  public boolean hasChangedConstructs(Domain domain) {
    if (added.isEmpty() && removed.isEmpty() && edited.isEmpty())
      return false;
    else return true;				   
  }

  /** 
   * Finds out whether the domain and its constructs have changed from the
   * draft.
   */
  public boolean hasChangedAll() {
    return hasChangedAll(draftDomain);
  }
  /** 
   * Finds out whether the domain and its constructs have changed from the
   * original.
   */
  public boolean hasChangedAllFromOriginal() {
    return hasChangedAll(currentDomain);
  }

  /** 
   * Finds out whether the domain and its constructs have changed from the
   * original.
   */
  public boolean hasChangedConstructsFromOriginal() {
    return hasChangedConstructs(currentDomain);
  }


  private void setSaving(Domain domain) {
    saving = domain;
  }
   private void unsetSaving() {
    saving = null;
  }
 private boolean isSaving(Domain domain) {
    return domain.equals(saving);
  }


  public void refinementAdded(RefinementEvent e) {
    Domain source = (Domain)e.getSource();
    //Debug.noteln("UID: Refinement added", e.getRefinement().getName());
    //Debug.noteln(" domain is", source);
    // ignore own updates
    if (!isSaving(source)) fireConstructAdded(e);
  }


  //--------------------------Getting information-----------------------------

  public List getAllRefinements() {
    List list;
    if (currentDomain == null) {
      //added only - removed ones that had been added are gone already
      list = new LListCollector(added);
    }
    else {
      //original, filter with UI ones (use UI ones in preference)
      // + UI-added (removed have already been dropped from added)
      // - removed (removed originals)
      //get added
      HashSet uiRefs = new HashSet(added);
      //get originals
      HashSet original = new HashSet(currentDomain.getRefinements());
      //use ui ones of originals and take out removed ones
      for (Iterator i = original.iterator(); i.hasNext(); ) {
	Refinement r = (Refinement)i.next();
	UIObject uio = findUIObject(r);
	if (uio != null) {
	  if (!removed.contains(uio)) 
	    //ui version exists and is not removed so forget the domain version
	    uiRefs.add(uio);
	}
	else if (!removed.contains(r)) uiRefs.add(r);
      }
      list = new LListCollector(uiRefs);
    }
    //Debug.noteln("AllRefs are:", IVUtil.namedListToDisplay(list));
    return list;
  }

  public List getAllConstructs() {
    return getAllRefinements();
  }

  public UIObject newUIObject(IXObject object) {
    if (object == null) return null;
    Class ixClass = object.getClass();
    UIObject uiObject = null;
    if (Refinement.class.isAssignableFrom(ixClass))
      uiObject = new UIRefinement(this, (Refinement)object);
    else if (Domain.class.isAssignableFrom(ixClass))
      uiObject = new UIDomain((Domain)object, editor);
    else return null;

    addedObject(uiObject, object);
    return uiObject;
  }

  /**
   * @return a UIRefinement, a Refinement, or null - depending on which exists.
   */
  public Object getNamedRefinement(String name) {
    Object refinement = findUIObject(name);
    if (refinement == null) 
      refinement = currentDomain.getNamedRefinement(name);
    return refinement;
  }

  public UIObject getUIObject(IXObject object) {
    //find out whether there is a UI version for this object
    UIObject uio = findUIObject(object);
    if (uio == null) return newUIObject(object);
    else return uio;
  }

  public UIObject findUIObject(IXObject object) {
    //Debug.noteln("UID: touched length:-------",touched.size());
    Object o = touched.get(object);
    if (o != null) {
      //Debug.note("UID: found object for " + ((Named)object).getName());
      //Debug.noteln(" -- object is",((Named)o).getName());
    }
    return (UIObject)o;
  }
  public UIObject findUIObject(String name) {
    //Debug.noteln("UID: Looking for named object",name);
    UIObject found = findUIObject(name, added);
    if (found == null)
      found = findUIObject(name, touched.values());
    return found;
  }

  private static UIObject findUIObject(String name, Collection collection) {
    for (Iterator i = collection.iterator(); i.hasNext(); ) {
      try {
	Named uio = (Named)i.next();
	//Debug.noteln(" checking", uio.getName());
	if (uio.getName().equals(name)) return (UIObject)uio;
      }
      catch (Exception e) {} //dont worry if the object is not named
    }
    return null;
  }

  public void addedObject(UIObject uio, IXObject original) {
    if (original == null) 
      addConstruct(uio);
    else {
      //Debug.note("UID: touched object " + ((Named)original).getName());
      //Debug.noteln(" - with object",((Named)uio).getName());
      touched.put(original, uio);
      //Debug.noteln("UID:              touched length:", touched.size());
      //fireUIConstructAdded(uio);
    }
  }

  public Set getMatchingRefinements(LList pattern) {
    //List dMatches = currentDomain.getMatchingRefinements(pattern);
    List dMatches = currentDomain.getRefinements();
    HashSet uiMatches = getMatchingUIRefinements(pattern, dMatches);
    return uiMatches;
  }

  public HashSet getMatchingUIRefinements(LList pattern, List domainMatches) {
    HashSet uiMatches = new HashSet(domainMatches);
    //check domain matches for removed or edited ones
    for (Iterator i = domainMatches.iterator(); i.hasNext(); ) {
      Refinement r = (Refinement)i.next();
      UIObject uio = findUIObject(r);
      if (uio != null)  //ui version exists, so forget the domain version
	uiMatches.remove(r); 
      else if (isMatchingRefinement(pattern, r))
	uiMatches.add(r);
    }
    //now add the UI ones that match
    for (Iterator i = touched.values().iterator(); i.hasNext(); ) {      
      UIRefinement uir = (UIRefinement)i.next();
      if (isMatchingRefinement(pattern, uir)) 
	uiMatches.add(uir);
    }
    return uiMatches;
  }

  /**
   * Checks whether the given refinement matches the pattern.
   * If it has been removed, it never matches.
   * If it has been added or edited, it matches if the patterns match.
   */
  protected boolean isMatchingRefinement(LList pattern, UIRefinement uir) {
    //removed, so no valid match
    if (removed.contains(uir)) return false;
    //added or edited, check if it matches
    if (edited.contains(uir) || added.contains(uir)) 
      return uir.matchesPattern(pattern);
    else return false; //should not happen
  }
  /**
   * Checks whether the given refinement matches the pattern.
   * It matches if the patterns match.
   */
  protected boolean isMatchingRefinement(LList pattern, Refinement r) {
    if (((pattern == null) && (r.getPattern() == null)) ||
	(pattern.equal(r.getPattern())) || 
	(pattern.equals(r.getPattern())))
      return true;
    else return (null != SimpleMatcher.match(pattern, r.getPattern()));
  }
  

  //--------------------------Managing domain---------------------------------
  public void saveToDomain() {
    if (currentDomain == null) currentDomain = (Domain)makeBaseObject();
    if (draftDomain == null) draftDomain = cloneDomain((Domain)currentDomain);
    saveToDomain(draftDomain);
  }
  public void fullSaveToDomain() {
    if (currentDomain == null) currentDomain = (Domain)makeBaseObject();
    if (draftDomain == null) draftDomain = cloneDomain((Domain)currentDomain);
    fullSaveToDomain(draftDomain);
  }

  public void saveToDomain(Object domainObject) {
    Domain domain = (Domain)domainObject;
    domain.setName(name);
    domain.setAnnotations(annotations);
  }
  public void fullSaveToDomain(Object domainObject) {
    Domain domain = (Domain)domainObject;
    setSaving(domain);
    domain.setName(name);

    Debug.noteln("UID-Added:", UIUtil.listToDisplay(added));
    Debug.noteln("UID-Edited:", UIUtil.listToDisplay(edited));
    Debug.noteln("UID-Removed:", UIUtil.listToDisplay(removed));

    for (Iterator i=added.iterator(); i.hasNext();) {
      Object o = i.next();
      //Debug.noteln("UID: adding object", o.toString());
      if ((UIObject.class.isInstance(o)) && (!((UIObject)o).isEmpty()))
	((UIObject)o).addToDomain(domain);
    }
    for (Iterator i=edited.iterator(); i.hasNext();) {
      Object o = i.next();
      //Debug.noteln("UID: updating object", o.toString());
      if (UIObject.class.isInstance(o))
	((UIObject)o).updateInDomain(domain);
    }
    for (Iterator i=removed.iterator(); i.hasNext();) {
      Object o = i.next();
      //Debug.noteln("UID: removing object", o.toString());
      if (UIObject.class.isInstance(o))
	((UIObject)o).removeFromDomain(domain);
      else removeFromDomain(domain, o);
    }
    added.clear();
    edited.clear();
    removed.clear();
    //Debug.noteln("UID - finished saveToDomain(domain)");
    unsetSaving();
  }
  public void publishToDomain() {
    if (currentDomain == null) currentDomain = new Domain();
    fullSaveToDomain(currentDomain);
    draftDomain = cloneDomain((Domain)currentDomain);
  }
  /** Copy the name of the original domain */
  public void loadFromDomain() {
    loadFromDomain(draftDomain);
  }
  public void loadFromDomain(Object domainObject) {    
    name = ((Domain)domainObject).getName();
    annotations = ((Domain)domainObject).getAnnotations();
    //Debug.noteln("UID: got name", name);
    fireDomainEdited();
  }
  public void loadFromOriginal() { 
    loadFromDomain(currentDomain);
  }



  /** 
   * Reverts the domain and all its construct to the last published domain.
   * If the domain has not been published, reverts to the original (usually
   * an empty domain).
   */
  public void revertToOriginal() {
    added.clear();
    edited.clear();
    removed.clear();
    touched.clear();
    graphs.clear();
    draftDomain = cloneDomain(currentDomain);
    touched.put(currentDomain, this);
    loadFromDomain(currentDomain);
    fireDomainEdited();
  }

  public void removeFromDomain(Domain domain, Object o) {
    if (Refinement.class.isInstance(o)) {
      domain.deleteRefinement((Refinement)o);
    }
  }

  //--------------------------Notifying listeners-----------------------------


  public void addUIDomainListener(UIDomainListener listener) {
    listeners.add(listener);
  }
  public void setUIDomainListeners(HashSet listeners) {
    this.listeners = listeners;
  }
  public void removeUIDomainListener(UIDomainListener listener) {
    listeners.remove(listener);
  }

  public void fireConstructAdded(RefinementEvent e) {
    //Debug.noteln("UID added construct");
    //(new Throwable()).printStackTrace();
    for (Iterator i = listeners.iterator(); i.hasNext();) {
      UIDomainListener listener = (UIDomainListener)i.next();
      listener.constructAdded(e);
    }
  }
  public void fireUIConstructAdded(UIObject object) {
    //Debug.noteln("UID added UIconstruct", object.getName());
    //(new Throwable()).printStackTrace();
    UIDomainEvent event = new UIDomainEvent(this, object);
    for (Iterator i = listeners.iterator(); i.hasNext();) {
      UIDomainListener listener = (UIDomainListener)i.next();
      listener.constructAdded(event);
    }
  }
  public void fireUIConstructEdited(UIObject object) {
    //Debug.noteln("UID edited UIConstruct");
    UIDomainEvent event = new UIDomainEvent(this, object);
    for (Iterator i = listeners.iterator(); i.hasNext();) {
      UIDomainListener listener = (UIDomainListener)i.next();
      listener.constructEdited(event);
    }
  }
  public void fireUIConstructRemoved(UIObject object) {
    //Debug.noteln("UID ******** removed UIConstruct", object.getName());
    UIDomainEvent event = new UIDomainEvent(this, object);
    for (Iterator i = listeners.iterator(); i.hasNext();) {
      UIDomainListener listener = (UIDomainListener)i.next();
      listener.constructRemoved(event);
    }
  }
  /**
   * The domain's own fields changed (currently only name)
   */
  public void fireDomainEdited() {
    //Debug.noteln("UID edited domain");
    UIDomainEvent event = new UIDomainEvent(this);
    for (Iterator i = listeners.iterator(); i.hasNext();) {
      UIDomainListener listener = (UIDomainListener)i.next();
      listener.domainEdited(event);
    }
  }
  public void fireDomainCleared() {
    //Debug.noteln("UID cleared domain");
    UIDomainEvent event = new UIDomainEvent(this);
    for (Iterator i = listeners.iterator(); i.hasNext();) {
      UIDomainListener listener = (UIDomainListener)i.next();
      listener.domainCleared(event);
    }
  }
  public void fireDomainSet(Domain domain) {
    //Debug.noteln("UID set domain");
    UIDomainEvent event = new UIDomainEvent(this);
    for (Iterator i = listeners.iterator(); i.hasNext();) {
      UIDomainListener listener = (UIDomainListener)i.next();
      listener.domainSet(event);
    }
  }



  //--------------------------UIObject things---------------------------------
  public Domain getDomain() {
    return currentDomain;
  }
  public UIDomain getUIDomain() {
    return this;
  }
  /** 
   * Makes a new uiDomain which points to the given domain.
   * Transfers its own listeners to the new uiDomain and points the editor
   * to the new uiDomain (which should free this domain to be deleted)
   */
  public UIDomain setDomain(Domain domain) {
    if (getDomain().equals(domain)) return this;
    else return setNewDomain(domain);
  }
  /** 
   * Makes a new uiDomain which points to the given domain.
   * Transfers its own listeners to the new uiDomain and points the editor
   * to the new uiDomain (which should free this domain to be deleted)
   */
  public UIDomain setNewDomain(Domain domain) {
    //Debug.noteln("UID: setting new UIDomain for", domain.toString());
    UIDomain newUID = new UIDomain(domain, editor);
    //Debug.noteln("  new UIDomain is", newUID.print());
    newUID.setUIDomainListeners(listeners);
    editor.setDomain(domain);

    //***What are the implications??
    //Debug.noteln("Cannot set the current domain yet.");
    
    fireDomainSet(domain);
    return newUID;
  }

  public boolean hasOriginal() {
    return (currentDomain != null);
  }
  
  public IXObject getBaseObject() {
    return currentDomain;
  }
  public IXObject makeBaseObject() {
    return new Domain();
  }
  public void setBaseObject(IXObject object) {
    currentDomain = (Domain)object;
  }
  public Object getBaseReference() {
    return baseReference;
  }
  public void setBaseReference(Object o) {
    baseReference = (String)o;
  }
  public Class getBaseClass() {
    return Domain.class;
  }




  //-------------------------- things---------------------------------

  public Object clone() throws CloneNotSupportedException {
    return super.clone();
  }
 
  public boolean isEmpty() {
    if (added.isEmpty() && ((currentDomain == null) || 
			    (currentDomain.isEmpty())))
      return true;
    else 
      return false;
  }

  public void clearThis() {
    name = "";
    annotations = null;
  }

  public void clear() {
    clearThis();
    added.clear();
    edited.clear();
    removed.clear();
    touched.clear();
    graphs.clear();
    draftDomain.clear();
    removed.addAll(currentDomain.getRefinements());
    saveToDomain(draftDomain);
    fireDomainCleared();
  }
  public void clearBase() {
    currentDomain.clear();
    clear();
    saveToDomain(currentDomain);
    /*
    added.clear();
    edited.clear();
    removed.clear();
    touched.clear();
    graphs.clear();
    currentDomain.clear();
    fireDomainCleared();
    */
  }
  /**
   * Clear all constructs in the draft domain. If there is a bas-domain, 
   * note that all its constructs are removed.
   */
  public void clearDraft() {
    clear();
  }

  public String getName() {
    return name;
  }
  public void setName(String name) {
    this.name = name;
  }
  public Annotations getAnnotations() {
    return annotations;
  }
  public void setAnnotations(Annotations annotations) {
    this.annotations = annotations;
  }

  public void setDomainFile(File file) {
    domainFile = file;
  }
  public File getDomainFile(File file) {
    return domainFile;
  }
  /**
   * Save the UIDomain to a local file - does not publish! 
   * If the domain has been saved before, it uses the previous file (warning
   * about overwriting). Otherwise it will ask the user for a file.
   */
  public void saveDomain() {
    saveDomain(domainFile);
  }
  /**
   * Save the UIDomain to a local file - does not publish! 
   * If the domain has been saved before, it uses the previous file (warning
   * about overwriting). Otherwise it will ask the user for a file.
   */
  public void saveDomainAs() {
    saveDomain(null);
  }
  /**
   * Save the UIDomain to the given local file - does not publish! 
   * Call this with a null argument for saveAs behaviour.
   */
  public void saveDomain(File file) {
    //save this domain into the draftDomain
    fullSaveToDomain(draftDomain);
    
    //save the draft domain to a suitable file (if there is a given one,
    // use that otherwise ask the user for one).
    if (file != null) 
      file = DomainWriter.saveDomain(editor, draftDomain, file); 
    //no file known or user decided not to overwrite existing one
    if (file == null) 
      file = DomainWriter.saveDomain(editor, draftDomain);
    //remember the file that was used
    if (file != null) domainFile = file;
  }


  public void analyseDomain() {
    //currentDomain.checkRefinementReferences();
  }

  private Domain cloneDomain(Domain original) {
    //Make a deep copy! otherwise the refinements will be same...
    try {
      Domain newDom = (Domain)original.clone();
      newDom.setRefinements(new ArrayList(original.getRefinements()));
      return newDom;
    }
    catch (CloneNotSupportedException e) {
      Debug.error("Cloning of the domain failed");
      return null;
    }
  }

  public String toString() { return name; }
  public String print() { 	
    List refinements = getAllRefinements();
    return "UIDomain[" + name + ", "  + refinements.size() + 
      " refinements]";
  }



  public void insertDomain() {
    loading = true;
    File file = DomainParser.loadDomain(editor, draftDomain);
    //Debug.noteln("UID - loading domain. Name is", draftDomain.getName());
    if (file != null) {
      domainFile = file;
      //copy the domain's new own fields
      loadFromDomain();
      //copy the new refinements
      List refinements = draftDomain.getRefinements();
      List newRefs = new ArrayList();
      for (Iterator i = refinements.iterator(); i.hasNext(); ) {
	IXObject ixo = (IXObject)i.next();
	addConstruct(getUIObject(ixo));
      }
      //Debug.noteln("UID: this",UIUtil.show(getAllRefinements()));
      //Debug.noteln(" draft",UIUtil.show(draftDomain.getRefinements()));
      //Debug.noteln(" current",UIUtil.show(currentDomain.getRefinements()));
    }
    loading = false;
    fireDomainEdited();
  }

  /**
   * Clears the draft domain (noting all of original domain as removed) 
   * and loads a domain-file into the draft domain.
   */
  public void loadDomain() {
    //clear added, edited, removed; remove all in draft domain
    clear();
    insertDomain();
  }





  public void addGraph(Named g) {
    graphs.put(g.getName(), g);
    fireGraphAdded(g);
  }

  public void fireGraphAdded(Object g) {
  }

  public LList getAllGraphs() {
    LListCollector list = new LListCollector(graphs.values());
    return list.contents();
  }
  public Named getNamedGraph(String name) {
    return (Named)graphs.get(name);
  }
  public void removeNamedGraph(String name) {
    graphs.remove(name);
  }


  //these do not make sense for the domain intself.
  public void addToDomain(Domain domain) {}
  public void updateInDomain(Domain domain) {}
  public void removeFromDomain(Domain domain) {}

  public void moveToUIDomain(UIDomain uiDomain) {}
  public void noteChange() {}
  
}


// Issues:
// * 
// * 
// * 
