/****************************************************************************
 * Mirrors the domain's Refinement and adds UI, Graph, and TreeNode things
 *
 * @author Jussi Stader
 * @version 2.0
 * Copyright: (c) 2002, AIAI, University of Edinburgh
 *
 *****************************************************************************
 */

package ix.iview.domain;

import java.util.*;
import javax.swing.tree.*;

import ix.util.*;
import ix.iface.ui.util.*;
import ix.util.lisp.*;
import ix.util.match.*;
import ix.icore.*;
import ix.icore.domain.*;
import ix.iview.igraph.*;
import ix.iview.*;

/**
 * Mirrors the domain's Refinement and adds UI, Graph, and TreeNode things
 */

public class UIRefinement 
implements UIObject, Named, SchemaTerms, TreeNode {
  public String name;
  public LList pattern = Lisp.NIL;
  public List nodes = new ArrayList();	// List of NodeSpecs
  public List orderings = new ArrayList();	// List of Orderings
  public List constraints = new ArrayList();
  public List issues = new ArrayList();
  public String comments = "";    // annotations
    
  protected UIDomain uiDomain;
  protected Refinement baseObject;
  protected String baseReference = ""; /** name of the original refinement */
  protected IGraph graph;


  public UIRefinement(UIDomain uiDomain) {
    super();
    this.uiDomain = uiDomain;
    //uiDomain.addConstruct(this);
  }

  public UIRefinement(UIDomain uiDomain, Refinement original) {
    //this(uiDomain);
    super();
    this.uiDomain = uiDomain;
    
    try {
      baseObject = (Refinement)original.clone();
    }
    catch (CloneNotSupportedException e) {
      baseObject = (Refinement)makeBaseObject();
    }
    //if (!original.isEmpty()) { //***put back in for efficiency!
      baseReference = original.getName();
      loadFromDomain();
      //}
    uiDomain.addedObject(this, original);
  }



  //-----------------------UIRefinement things--------------------------------


  public String getName() { return name; }
  public void setName(String name) {
    //if (uiDomain.
    this.name = name; 
  }
  
  public LList getPattern() {	return pattern; }
  public void setPattern(LList pattern) { this.pattern = pattern; }
  public void clearPattern() {pattern = Lisp.NIL;}

  public List getNodes() { return nodes; }
  public void setNodes(List nodes) { 
    this.nodes = IViewUtil.sortNodeCollection(nodes); 
  }
  public void clearNodes() {nodes = new ArrayList();}
  

  public List getOrderings() { return orderings; }
  public void setOrderings(List orderings) { 
    this.orderings = new ArrayList(orderings); 
  }
  public void clearOrderings() {orderings = new ArrayList();}

  public List getConstraints() { return constraints; }
  public void setConstraints(List cs) { 
    this.constraints = new ArrayList(cs); 
  }
  public void clearConstraints() {constraints = new ArrayList();}

  public List getIssues() { return issues; }
  public void setIssues(List issues) { 
    this.issues = IViewUtil.sortTICollection(issues); }
  //public void addIssue(Issue issue) { issues.add(issue); }
  public void deleteIssue(Issue issue) { issues.remove(issue); }
  public void clearIssues() {issues = new ArrayList();}

  public String getComments() { return comments; }
  public void setComments(String comments) { this.comments = comments; }

  //public String toString() { return name; }

  public String toString() {
    return "Refinement[" + name + " expands " + pattern + "]";
  }


  //-----------------------UIObject things------------------------------------
  public Domain getDomain() {
    return uiDomain.getDomain();
  }
  public UIDomain getUIDomain() {
    return uiDomain;
  }

  public void moveToUIDomain(UIDomain uiDomain) {
    if (this.uiDomain == null) this.uiDomain = uiDomain;
    else { //**this should be more difficult - what does it mean?
      this.uiDomain = uiDomain;
      uiDomain.addedObject(this, getBaseObject());
    }
  }


  public Object getBaseReference() {
    return baseReference;
  }
  public void setBaseReference(Object ref) {
    baseReference = (String) ref;
  }


  public IXObject getBaseObject() {
    return baseObject;
  }
  public IXObject makeBaseObject() {
    return new Refinement();
  }
  public void setBaseObject(IXObject object) {
    baseObject = (Refinement)object;
  }
  public Class getBaseClass() {
    return Refinement.class;
  }

  public boolean hasOriginal() {
    if ((getBaseReference() == null) || (getBaseReference().equals(""))) 
      return false;
    else return true;
  }


  /**
   * Saves this object into a suitable baseObject. Does not add the base object
   * to the domain.
   */
  public void saveToDomain() {
    if (baseObject == null) baseObject = (Refinement)makeBaseObject();
    //Debug.noteln("UIR-saveToDomain: ", baseObject.toString());
    saveToDomain(baseObject);
    //Debug.noteln(" nodes now:", UIUtil.listToDisplay(baseObject.getNodes()));
  }
  public void saveToDomain(Object domainObject) {
    Refinement refinement = (Refinement) domainObject;
    refinement.setName(name);
    refinement.setPattern(pattern);
    //refinement.setNodes(nodes);
    writeNodes(refinement);
    refinement.setOrderings(orderings);
    refinement.setConstraints(constraints);
    refinement.setIssues(issues);
    refinement.setComments(comments);
  }

  /**
   * Loads this object from its baseObject. 
   */
  public void loadFromDomain() {
    if (baseObject == null) {
      baseObject = new Refinement();
    }
    name = baseObject.getName();
    if (baseObject.getPattern() == null) pattern = Lisp.NIL;
    else pattern = baseObject.getPattern();
    makeNodes(baseObject.getNodes());
    if (baseObject.getOrderings() == null) orderings = new ArrayList();
    else orderings = baseObject.getOrderings();
    if (baseObject.getConstraints() == null)
      constraints = new ArrayList();
    else constraints = baseObject.getConstraints();
    if (baseObject.getIssues() == null) issues = new ArrayList();
    else setIssues(baseObject.getIssues());
    comments = baseObject.getComments();
  }

  public void loadFromDomain(Object domainObject) {
    baseObject = (Refinement)domainObject;
    loadFromDomain();
  }

  /**
   * Loads this object from its original in the domain if it has one, otherwise
   * clear the object. Replaces the base-object with a fresh clone of the 
   * original or with a new base-object.
   */
  public void loadFromOriginal() {
    if (hasOriginal()) {
      Object ref = getBaseReference();
      Refinement original = getDomain().getNamedRefinement((String)ref);
      try {
      setBaseObject((Refinement)original.clone());
      loadFromDomain();
      }
      catch (CloneNotSupportedException e) {
	Debug.noteException(e);
      }
    }
    else {
      setBaseObject(makeBaseObject());
      loadFromDomain();
    }
  }


  private void makeNodes(List originalNodes) {
    nodes = new ArrayList();
    if (originalNodes == null) return;
    for (Iterator i = originalNodes.iterator(); i.hasNext(); ) 
      nodes.add(new UINodeSpec(uiDomain, (NodeSpec)i.next()));
  }
  private void writeNodes(Refinement refinement) {
    if (nodes == null) refinement.setNodes(null);
    else {
      List adds = new LListCollector();
      for (Iterator i = nodes.iterator(); i.hasNext(); ) {
	NodeSpec ns = (NodeSpec)i.next();
	adds.add(new NodeSpec(ns.getId(), ns.getPattern()));
      }
      refinement.setNodes(adds);
    }
  }
  private boolean sameNodes(List nodes, List otherNodes) {
    if (nodes == null)
      return ((otherNodes == null) || (otherNodes.size() == 0));
    else if (otherNodes == null)
      return ((nodes == null) || (nodes.size() == 0));

    if (nodes.size() != otherNodes.size()) return false;
    else {
      //hash one list with id/pattern, then lookup nodes by ids
      Hashtable hNodes = new Hashtable();
      //Debug.noteln("UIR: nodes");
      for (Iterator i = nodes.iterator(); i.hasNext(); ) {
	NodeSpec ns = (NodeSpec)i.next();
	//Debug.note(" " + ns.getId() + UIUtil.listToDisplay(ns.getPattern()));
	hNodes.put(ns.getId(), ns.getPattern());
      }
      //Debug.noteln("UIR: other nodes");
      for (Iterator i = otherNodes.iterator(); i.hasNext(); ) {
	NodeSpec ns = (NodeSpec)i.next();
	//Debug.note(" " + ns.getId() + UIUtil.listToDisplay(ns.getPattern()));
	LList p1 = (LList)hNodes.get(ns.getId());
	LList p2 = ns.getPattern();
	if (p2 == null) {
	  if (p1 != null) return false;
	}
	else if (!p2.equal(p1)) return false;
      }
      return true;
    }
  }
  private boolean sameIssues(List issues, List otherIssues) {
    if (bothEmpty(issues, otherIssues)) return true;
    else if ((issues == null) || (otherIssues == null)) return false;

    //compare issues by patterns only!
    else if (issues.size() != otherIssues.size()) return false;
    else {
      //hashSet one list with patterns only, lookup other list's patterns
      HashSet hIssues = new HashSet();
      //Debug.noteln("UIR: issues");
      for (Iterator i = issues.iterator(); i.hasNext(); ) {
	Issue issue = (Issue)i.next();
	//Debug.note(" " + issue.getId());
	//Debug.note(" " + UIUtil.listToDisplay(issue.getPattern()));
	hIssues.add(issue.getPattern());
      }
      //Debug.noteln("UIR: other issues");
      for (Iterator i = otherIssues.iterator(); i.hasNext(); ) {
	Issue issue = (Issue)i.next();
	//Debug.note(" " + issue.getId());
	//Debug.note(" " + UIUtil.listToDisplay(issue.getPattern()));
	if (!hIssues.contains(issue.getPattern())) {
	  //Debug.note("UIR: Could not find issue pattern ");
	  //Debug.noteln(UIUtil.listToDisplay(issue.getPattern()));
	  return false;
	}
      }
      return true;
    }
  }
  private boolean sameList(List list, List otherList) {
    if (bothEmpty(list, otherList)) return true;
    else return ((list != null) && list.equals(otherList));
  }
  private boolean sameSet(Collection list, Collection otherList) {
    if (bothEmpty(list, otherList)) return true;
    if ((list == null) || (otherList == null)) return false;
    HashSet set = new HashSet(list);
    HashSet otherSet = new HashSet(otherList);
    return ((set != null) && set.equals(otherSet));
  }
  private boolean sameString(String thing, String other) {
    if (((thing == null) || thing.equals("")) && 
	((other == null) || other.equals(""))) return true;
    else return thing.equals(other);
  }
  private boolean bothEmpty(Collection list, Collection otherList) {
    if (((otherList == null) || (otherList.size() == 0)) &&
	((list == null) || (list.size() == 0)))
      return true;
    else return false;
  }



  //------------------------updating the original domain-------

  public void addToDomain() { addToDomain(getDomain()); }
  public void addToDomain(Domain domain) {
    //if there already is such an object, complain. ***do this?
    
    //If there is no base-object, make a new one.
    if (baseObject == null) baseObject = (Refinement)makeBaseObject();

    //copy this into the base object (to be safe) then pass it to the domain.
    saveToDomain();
    domain.addRefinement(baseObject);
  }
  public void updateInDomain() { updateInDomain(getDomain()); }
  public void updateInDomain(Domain domain) {
    //if there is no base object, complain (add instead?)
    String ref = (String)getBaseReference();
    Refinement r = getReferredObject(domain, ref, name);
    if (r == null) {
      Debug.noteln("Update refinement: cannot find base object called", ref);
      return;
    }
    //copy this into base object (to be safe) and pass it with its id to the
    //domain
    Debug.noteln("UIR: Update refinement base is ", ref);
    saveToDomain();
    domain.replaceNamedRefinement(ref, baseObject);
  }
  public void removeFromDomain() { removeFromDomain(getDomain()); }
  public void removeFromDomain(Domain domain) {
    //if there is no base object id, try to find the id in the domain
    //Complain if it cannot be found
    String ref = (String)getBaseReference();
    Refinement r = getReferredObject(domain, ref, name);
    if (r == null) {
      Debug.noteln("Remove refinement: Cannot find a base object called ",ref);
      return;
    }
    //give the id to the domain and ask for removal
    domain.deleteNamedRefinement(ref);
  }

  private Refinement getReferredObject(Domain domain, 
				       String reference, String name) {
    String ref;
    if ((reference == null) || (reference.equals(""))) 
      ref = name;
    else ref = reference;
    return domain.getNamedRefinement(ref);
  }

  /**
   * Get the UIDomain to note the change of this object
   */
  public void noteChange() {
    if (isEmpty() && !hasOriginal()) 
      uiDomain.removeConstruct(this);
    else if (hasChanged()) {
      //saving, so update the base object
      saveToDomain();

      //Debug.noteln("UIR: noting change");
      uiDomain.updateConstruct(this);
    }
    //if (isEmpty()) uiDomain.removeConstruct(this);
  }


  /**
   * A refinement is empty if all its data fields are empty.
   * Note: the refinement may not have started out as empty - check whether it
   * has a base-object for full emptiness.
   */
  public boolean isEmpty() {
    if ((name == null || (name.equals("")))
	&& (pattern == null || pattern.isEmpty())
	&& (nodes == null || nodes.isEmpty())
	&& (orderings == null || orderings.isEmpty())
	&& (issues == null || issues.isEmpty())
	&& (comments == null || (comments.equals("")))) {
      //Debug.noteln("UIRefinement is empty");
      return true;
    }
    //Debug.noteln("refinement is not empty");
    return false;
  }
   /**
   * A refinement is empty if all its data fields are empty.
   * Note: the refinement may not have started out as empty - check whether it
   * has a base-object for full emptiness.
   */
  public boolean isEmpty(Refinement refinement) {
    if (refinement == null) return true;
    if ((refinement.getName() == null || (refinement.getName().equals("")))
	&& (refinement.getPattern() == null || 
	    refinement.getPattern().isEmpty())
	&& (refinement.getNodes() == null || refinement.getNodes().isEmpty())
	&& (refinement.getOrderings() == null || 
	    refinement.getOrderings().isEmpty())
	&& (refinement.getIssues() == null || refinement.getIssues().isEmpty())
	&& (refinement.getComments() == null || 
	    refinement.getComments().equals(""))) {
      //Debug.noteln("refinement is empty");
      return true;
    }
    //Debug.noteln("refinement is not empty");
    return false;
  }
  
  /** 
   * Finds out whether the refinement has changed from the base-object 
   * (i.e. since the object was last saved to draft).
   */
  public boolean hasChanged() {
    return hasChanged((Refinement)getBaseObject());
  }
  /** 
   * Finds out whether the refinement has changed from the given object.
   * The refinement has not changed if the fields have not changed.
   * The refinement's fields are:
   * name, pattern, nodes, orderings, constraints, issues, comments
   * The refinement has also not changed if the original is null and it is
   * empty.
   */
  public boolean hasChangedNoDebug(Refinement original) {
    //created when "editing" a sub-node (undefined, pattern set, nothing else)
    if (isEmpty(original) 
	&& (name != null) && (name.equals("Undefined") 
			      || name.equals("undefined"))
	&& (nodes == null || nodes.isEmpty())
	&& (orderings == null || orderings.isEmpty())
	&& (issues == null || issues.isEmpty())
	&& (comments == null || (comments.equals(""))))
      return false;
    if (original == null) return !isEmpty();
    if (!nameHasChanged(original.getName()) && 
	sameList(pattern,original.getPattern()) && 
	sameNodes(nodes,original.getNodes()) && 
	sameSet(orderings, original.getOrderings()) && 
	sameList(constraints, original.getConstraints()) && 
	sameIssues(issues, original.getIssues()) && 
	sameString(comments,original.getComments()))
      return false;
    else return true;				   
  }
  /** 
   * Finds out whether the refinement has changed from the given object.
   * The refinement has not changed if the fields have not changed.
   * The refinement's fields are:
   * name, pattern, nodes, orderings, constraints, issues, comments
   * The refinement has also not changed if the original is null and it is
   * empty.
   */
  public boolean hasChanged(Refinement original) {
    //created when "editing" a sub-node (undefined, pattern set, nothing else)
    if (isEmpty(original)
	&& (name != null) && (name.equals("Undefined") 
			      || name.equals("undefined"))
	&& (nodes == null || nodes.isEmpty())
	&& (orderings == null || orderings.isEmpty())
	&& (issues == null || issues.isEmpty())
	&& (comments == null || (comments.equals("")))) {
      Debug.noteln("Empty sub-node editing");
      return false;
    }
    if (original == null) {
      //Debug.noteln("UIR hasChanged from null original");
      return !isEmpty();
    }
    if (!nameHasChanged(original.getName())) {
      if (sameList(pattern, original.getPattern())) {
	if (sameNodes(nodes,original.getNodes())) {
	  if (sameSet(orderings, original.getOrderings())) {
	    if (sameList(constraints, original.getConstraints())) {
	      if (sameIssues(issues, original.getIssues())) {	
		if (sameString(comments,original.getComments())) return false;
		else Debug.noteln("UIR hasChanged comments");
	      }
	      else Debug.noteln("UIR hasChanged issues");
	    }
	    else Debug.noteln("UIR hasChanged constraints");
	  }
	  else {
	    Debug.noteln("UIR hasChanged orderings");
	    Debug.noteln(UIUtil.listToDisplay(orderings));
	    Debug.noteln(UIUtil.listToDisplay(original.getOrderings()));
	  }
	}
	else Debug.noteln("UIR hasChanged nodes");
      }
      else Debug.noteln("UIR hasChanged pattern");
    }
    else Debug.noteln("UIR hasChanged name from", original.getName());
    
    return true;				   
  }
  /** 
   * Finds out whether the refinement has changed from the original.
   * The refinement has not changed if there is an original (baseReference)
   * and the refinement's fields have not changed.
   * The refinement's fields are:
   * name, pattern, nodes, orderings, constraints, issues, annotations
   */
  public boolean hasChangedFromOriginal() {
    //has no original so it is new (changed!)
    if (!this.hasOriginal()) return true;

    //has original, so compare the two
    Refinement original = getDomain().getNamedRefinement(baseReference);
    return hasChanged(original);
  }

  public boolean nameHasChanged(Object nameThing) {
    String original = (String) nameThing;
    if ((((name == null) || name.equals("")) &&
	 ((original == null) || (original.equals("")))) ||
	((name != null) && name.equals(original)))
      return false;
    else return true;
  }

  //-----------------------UI service things-----------------------------------


  /**
   * Builds a LList of nodes from a string that contains node descriptions.
   * 
   * @param expansion a string containing node descriptions (patterns)
   * seperated by new lines.
   * @return a LList of Nodes.
   */
  public static List parseNodes (UIDomain uiDomain, String expansion) {
    List nodes = new ArrayList();
    long i = 1;
    // For some reason, it doesn't work on Windows to
    // use lineSpearator here.
    while (!expansion.equals("")) {
      // String[] parts =
      //    Util.breakStringAtFirst(expansion, lineSeparator);
      String[] parts = UIUtil.breakStringAtFirst(expansion, "\n");
      Long index = new Long(i);
      LList childPattern = Lisp.elementsFromString(parts[0]);
      if (!childPattern.isNull()) {
	nodes.add(new UINodeSpec(uiDomain, index, childPattern));
	//Debug.noteln("got child", index + childPattern.toString());
	//nodes.addElement(new DefaultNode(index, childPattern));
	i++;
      }
      expansion = parts[1];
    }
    return nodes;
  }

  /**
   * Builds a LList of nodes from a string that contains node descriptions.
   * 
   * @param expansion a string containing node descriptions (patterns)
   * seperated by new lines.
   * @return a LList of Nodes.
   */
  public static List parseIssues (UIDomain uiDomain, String sIssues) {
    List issues = new ArrayList();
    long i = 1;
    // For some reason, it doesn't work on Windows to
    // use lineSpearator here.
    while (!sIssues.equals("")) {
      // String[] parts =
      //    Util.breakStringAtFirst(sIssues, lineSeparator);
      String[] parts = UIUtil.breakStringAtFirst(sIssues, "\n");
      LList pattern = Lisp.elementsFromString(parts[0]);
      if (!pattern.isNull()) {
	Issue newIssue = new Issue(pattern);
	issues.add(newIssue);
      }
      sIssues = parts[1];
    }
    return issues;
  }

  public void deleteNode(NodeSpec node) {
    deleteNodeConstraints(node);
    nodes.remove(node);
  }

  protected void deleteNodeConstraints(NodeSpec node) {
    Name nodeId = node.getId();
    //Debug.noteln("Removing orderings for node", nodeId);
    List os = new ArrayList();    
    for (Iterator i = orderings.iterator(); i.hasNext(); ) {
      Ordering constraint = (Ordering) i.next();
      if (!constraintMentionsNode(constraint, nodeId))
	os.add(constraint);
    }
    this.setOrderings(os);
  }

  public boolean matchesPattern(LList pattern) {
    if (((pattern == null) && (getPattern() == null)) ||
	(pattern.equal(getPattern())) || 
	(pattern.equals(getPattern())))
      return true;
    else return (null != SimpleMatcher.match(pattern, getPattern()));
  }

  public NodeSpec findNode(Name id) {
    for (Iterator i = nodes.iterator(); i.hasNext(); ) {
      NodeSpec node = (NodeSpec)i.next();
      if (id.equals(node.getId())) return node;
    }
    return null;
  }

  public static boolean constraintMentionsNode(Constraint c, Name nodeId) {
    Debug.noteln("*****UIRef: cannot see whether constraint mentions node");
    return false;
  }
  public static boolean constraintMentionsNode(Ordering o, Name nodeId) {
    NodeEndRef ref;
    ref = o.getFrom();
    if ((ref != null) && (nodeId.equals(ref.getNode()))) return true;
    else {
      ref = o.getTo();
      if ((ref != null) && (nodeId.equals(ref.getNode()))) return true;
      else return false;
    }
  }

  public Name getNewNodeId() {
    Name largest = getHighestNodeId();
    Integer l = new Integer(largest.toString());
    int i = l.intValue();
    i++;
    return Name.valueOf("" + i); 
  }
  public Name getHighestNodeId() {
    return getHighestNodeId(getNodes());
  }

  public static Name getHighestNodeId(List nodes) {
    Integer largest = new Integer(0);
    Name largestId = Name.valueOf(largest.toString());
    int c;
    Iterator inodes = nodes.iterator();
    while (inodes.hasNext()) {
      Name id = ((NodeSpec)inodes.next()).getId();
      Integer idInt = new Integer(id.toString());
      c = largest.compareTo(idInt);
      if (c <= 0) { largestId = id; largest = idInt; }
    }
    return largestId;
  }

  public void addNode(NodeSpec node) {
    //Debug.noteln("UIR: addNode - nodes list is class", nodes.getClass());
    nodes.add(node);
    setNodes(nodes); //setNodes sorts the nodes!
  }
  public void addIssue(Issue issue) {
    issues.add(issue);
    setIssues(issues); //setIssues sorts!
  }

  public void addConstraint(Object newValue) throws ConstraintTypeException {
    Constraint constraint = null;
    if (Constraint.class.isAssignableFrom(newValue.getClass())) 
      addConstraint((Constraint)newValue);
    else if (Ordering.class.isAssignableFrom(newValue.getClass())) 
      addOrdering((Constraint)newValue);
    else {
       String message = "Cannot recognise this as a constraint: " + '\n' +
	newValue.toString();
      throw new ConstraintTypeException(message);
    }
    /*
    Name cType = constraint.getType();
    if (cType.equals(C_ORDERING)) addOrdering(constraint);
    else if (cType == C_EFFECT) addEffect(constraint);
    else if (cType == C_CONDITION) addCondition(constraint);
    else if (cType == ) add(constraint);
    else {
      String m = "addConstraint found constraint with non-valid type" + cType;
      throw new ConstraintTypeException(m);
    }
    */
  }

  public void addOrdering(Constraint newValue) {
    orderings.add(newValue);
  }
  public void addOrdering(Ordering newValue) {
    orderings.add(newValue);
  }


  /**
   * Gets the refinement's ordering type (none, parallel, sequence, other).
   * @return <ul>
   * <li> none (undecided) if there are less than two nodes
   * <li> parallel if there are no orderings an 2 or more nodes
   * <li> sequence if there are 2 or more nodes and they are in sequence
   * <li> other otherwise.
   * </ul>
   */
  public String getOrderingType() {
    if ((nodes == null) || (nodes.size() < 2)) return O_NONE;
    else if ((orderings == null) || (orderings.isEmpty())) return O_PARALLEL;
    else if (orderingsAreSequential()) return O_SEQUENCE;
    else return O_OTHER;
  }

  /** 
   * Checks whether the orderings are a sequence of the nodes.
   */
  public boolean orderingsAreSequential() {
    HashSet sequence = new HashSet(makeSequence());
    return new HashSet(orderings).equals(sequence);
  }

  public List makeSequence() {
    //Debug.noteln("UIR - makeSequenceNodes");
    // make all current sub-nodes sequential

    List nodes = this.getNodes();
    Iterator i = nodes.iterator();
    NodeSpec previous = null;
    NodeSpec current = null;
    List orderings = new ArrayList();
    while (i.hasNext()) {
      current = (NodeSpec)i.next();
      if (previous != null) {
	Ordering refRel = 
	  new Ordering(new NodeEndRef(End.END ,previous.getId()), 
		       new NodeEndRef(End.BEGIN, current.getId()));
	orderings.add(refRel);
	//Debug.noteln(" Orderings are", orderings);
      }
      previous = current;
    }
    return orderings;
  }

  public class NodeTypeException extends Exception {
    public NodeTypeException() { super();}
    public NodeTypeException(String message) { super(message);}
  }
  public class ConstraintTypeException extends Exception {
    public ConstraintTypeException() { super();}
    public ConstraintTypeException(String message) { super(message);}
  }

  /**
   * Check whether the given object is a reference to the given node.
   * checks equals on node, node's toString, node's pattern
   * for NodeSpec refs checks node/ref's id, node/ref's pattern
   * for NodeEndRef checks isRef on ref's node
   */
  public static boolean isRef(NodeSpec node, Object o) {
    //Debug.noteln("Comparing " + node.toString() + " to " + o.toString());
    if ((node.equals(o)) 
	|| (node.getId().toString().equals(o.toString()))
	|| (node.getPattern().toString().equals(o.toString()))) return true;
    if ((NodeSpec.class.isAssignableFrom(o.getClass()))
	&& (node.getId().equals(((NodeSpec)o).getId()))
	&& (node.getPattern().equals(((NodeSpec)o).getPattern()))) return true;
    //Debug.noteln("false");
    if ((NodeEndRef.class.isAssignableFrom(o.getClass()))
	&& isRef(node, ((NodeEndRef)o).getNode())) return true;
    return false;
  }

  public String print() {
    String names = name + " " + pattern.toString();
    String ns = IViewUtil.printNodes(nodes);
    String os = IViewUtil.printOrderings(orderings);
    return "Refinement: " + names + "\n" +
      "  Nodes: " + ns +  "\n" +
      "  Orderings: " + os;
  }

  //-----------------------Graph things----------------------------------------

  public void setGraph(IGraph g) { graph = g; }
  public IGraph getGraph() { return graph; }
  public void removeGraph(IGraph g) { graph = null; }


  //-----------------------TreeNode things-------------------------------------

  private static final int NOFIND = -1;
  public Enumeration children() { 
    //nodes = this.getNodes();
    return new Enumeration() {
      /*
      int i = 0;
      public boolean hasMoreElements() { return i < nodes.size(); }
      public Object nextElement() { return nodes.get(i++); }
      */
      int i = nodes.size()-1;
      public boolean hasMoreElements() { return i >= 0; }
      public Object nextElement() { return nodes.get(i--); }
    };
  }
  public boolean getAllowsChildren() { return true; }
  public TreeNode getChildAt(int index) { 
    return null;//dummy
  }
  public int getChildCount() { return this.getNodes().size(); }
  public int getIndex(TreeNode node) {
    return NOFIND;//dummy
  }
  public TreeNode getParent() { return null; }//one level deep only
  public boolean isLeaf() { return (getChildCount() == 0);}

}

// Issues:
// * Ordering needs to sub-class Constraint
// * 

