/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Fri Feb 22 19:48:12 2002 by Jeff Dalton
 * Copyright: (c) 2002, AIAI, University of Edinburgh
 */

package ix.util;

import java.util.*;

/**
 * Converts between class names (Strings) and classes.  Note that a
 * single ClassFinder maps in both directions to make it easier to
 * create mappings that are consistent. <p>
 *
 * Explicit mappings and "imports" may be specified; otherwise
 * fully qualified names must be used. <p>
 *
 * Primitive types such as <code>byte</code>, <code>boolean</code>,
 * and <code>int</code> are matched with the corresponding <code>TYPE</code>
 * classes <code>Byte.TYPE</code>, <code>Boolean.TYPE</code>, etc. <p>
 *
 * <b>This class does not yet handle array classes.</b>
 */

public class ClassFinder {

    protected List imports = new LinkedList();
    protected Map nameToClass = new HashMap();
    protected Map classToName = new HashMap();
    protected Map nameToClassCache = new HashMap();
    protected Map classToNameCache = new HashMap();

    public ClassFinder() {
	addInitialNames();
	addInitialImports();
    }

    /**
     * Adds the initial set of name mappings.
     */
    protected void addInitialNames() {
	// Explicitly specify the mapping for primitive types,
	// because Class.forName(String) will not work.
	addName("byte",    Byte.TYPE);
	addName("char",    Character.TYPE);
	addName("short",   Short.TYPE);
	addName("int",     Integer.TYPE);
	addName("long",    Long.TYPE);
	addName("float",   Float.TYPE);
	addName("double",  Double.TYPE);
	addName("boolean", Boolean.TYPE);
	addName("void",    Void.TYPE);
    }

    /**
     * Adds the initial set of imports.  These are:
     * <pre>
     *   java.lang.*
     *   java.util.*
     * </pre>
     */
    protected void addInitialImports() {
	// Standard imports
	addImport("java.lang.*");
	addImport("java.util.*");
    }

    /**
     * Records an import that allows a class name or names to be used without
     * package-qualification.  The import is specified as either a fully
     * qualified class name or a package name suffixed by ".*", as in the
     * Java <code>import</code> statement.
     */
    public void addImport(String name) {
	imports.add(new Import(name));
	nameToClassCache.clear();         // Caches may no longer be valid
	classToNameCache.clear();
    }

    /**
     * Adds an explicit mapping in both directions between a name
     * and a class.  This takes precedence over all other ways of
     * determining those relationships.
     */
    public void addName(String name, Class c) {
	nameToClass.put(name, c);
	classToName.put(c, name);
	nameToClassCache.clear();         // Caches may no longer be valid
	classToNameCache.clear();
    }

    /**
     * Returns the class that corresponds to the name, as determined
     * by any explicit mappings and imports.  Fully qualified names
     * may also be used.
     *
     * @return a Class or null.
     */
    public Class classForName(String name) {
	Class c = (Class)nameToClassCache.get(name); // cached?
	if (c != null)
	    return c;
	if (c == null)
	    c = (Class)nameToClass.get(name);   // explicitly specified?
	if (c == null)
	    c = tryClassForName(name); 		// fully qualified?
	if (c == null)
	    c = findImportClass(name); 		// imported?
	if (c != null)
	    nameToClassCache.put(name, c);	// cache if found
	return c;
    }

    /**
     * Tries to find a class by asking each import in turn.
     *
     * @throws RuntimeException if more than one import can find a class.
     */
    protected Class findImportClass(String name) {
	Class result = null;
	Import source = null;
	for (Iterator i = imports.iterator(); i.hasNext();) {
	    Import imp = (Import)i.next();
	    Class c = imp.tryName(name);
	    if (c != null) {
		if (result == null) {
		    result = c;
		    source = imp;
		}
		else {
		    throw new RuntimeException
			("Ambiguous name " + Util.quote(name) +
			 " imported via both " + source + " and " + imp);
		}
	    }
	}
	return result;
    }

    /**
     * Like <code>Class.forName(String)</code> but returns null
     * if the class cannot be found.
     */
    public static Class tryClassForName(String name) {
	try { return Class.forName(name); }
	catch (ClassNotFoundException e) {
	    return null;
	}
    }

    /**
     * Returns the best name that this ClassFinder would map to the
     * specified class, where a plain class name counts as better than
     * a package-qualified one.
     */
    public String nameForClass(Class c) {
	// First see if cached.
	String name = (String)classToNameCache.get(c);
	if (name != null)
	    return name;

	// See if the mapping has been explicitly specified
	name = (String)classToName.get(c);
	if (name != null) {
	    Debug.assert(nameToClass.get(name) == c);
	    classToNameCache.put(c, name);	 // cache the result
	    return name;
	}

	// Otherwise, determine whether we need the fully qualified name
	// or can give a simple name because of imports.
	String fullname = c.getName();
	if (fullname.startsWith("[")) {
	    throw new RuntimeException("Can't handle array class" + fullname);
	}
	String shortname = Strings.afterLast(".", fullname);
	if (classForName(shortname) == c)
	    name = shortname;
	else
	    name = fullname;
	classToNameCache.put(c, name);	 // cache the result
	return name;
    }

    /**
     * Internal class that performs the name-to-class mapping that
     * corresponds to a single import specification.
     */
    protected static class Import {
	boolean star;
	String name;
	String prefix;

	public Import(String name) {
	    this.name = name;
	    star = name.endsWith("*");
	    if (star) {
		prefix = name.substring(0, name.length() - 1);
		Debug.assert(prefix.endsWith("."), "invalid import", name);
	    }
	    else
		prefix = name;
	}

	/**
	 * Tries to find a class for the name.
	 *
	 * @return a Class or null.
	 */
	public Class tryName(String name) {
	    if (star) 
		return tryClassForName(prefix + name);
	    else if (prefix.equals(name))
		return tryClassForName(prefix);	// in case name is fully qual
	    else if (prefix.endsWith("." + name))
		return tryClassForName(prefix); // explicit import
	    else
		return null;
	}

	public String toString() {
	    return "import[" + name + "]";
	}

    }

    // For testing ...

    /**
     * A test loop that asks the user for a name and prints
     * the corresponding class (or else null) as determined
     * by a plain ClassFinder.  When the result is not null,
     * it also prints the ClassFinder's "best" name for that
     * class.
     */
    public static void main(String[] argv) {
	ClassFinder cf = new ClassFinder();
	for (;;) {
	    String in = Util.askLine("Class name:");
	    if (in.equals("bye"))
		return;
	    Class c = cf.classForName(in);
	    System.out.println(c);
	    if (c != null)
		System.out.println("Best name is " + 
				   Util.quote(cf.nameForClass(c)));
	}
    }

}
