/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Thu May 30 03:24:48 2002 by Jeff Dalton
 * Copyright: (c) 2002, AIAI, University of Edinburgh
 */

package ix.util;

import java.util.*;

import ix.util.lisp.*;

/**
 * Collection and Map utilities
 *
 * @see Fn
 * @see Strings
 * @see Util
 */
public final class Collect {

    private Collect() { }	// block instantiation

    /**
     * Returns true iff the collection is either <code>null</code>
     * or empty.
     */
    public static final boolean isEmpty(Collection c) {
	return c == null || c.isEmpty();
    }

    /**
     * Returns true iff the Map is either <code>null</code>
     * or empty.
     */
    public static final boolean isEmpty(Map m) {
	return m == null || m.isEmpty();
    }

    /**
     * Returns the collections's iterator, treating <code>null</code>
     * as an empty list.
     */
    public static final Iterator iterator(Collection c) {
	return c == null ? Lisp.NIL.iterator() : c.iterator();
    }

    /**
     * Ensures that a value is a List object by returning an empty
     * list if the value is <code>null</code>.
     */
    public static List ensureList(List list) {
	// Can't return Collections.EMPTY_LIST, because it's of a
	// special class that Util.makeInstance and hence map can't handle.
	return list == null ? new ArrayList(0) : list;
    }

    /**
     * Replaces the first occurrence of an item in a list, returning
     * the modified list.  If the item does not occur, the list is
     * returned unchanged.
     */
    public static List replaceFirst(Object from, Object to, List source) {
	for (ListIterator i = source.listIterator(); i.hasNext();) {
	    if (i.next() == from) {
		i.set(to);
		break;
	    }
	}
	return source;
    }

    /**
     * Returns a new collection of the same class as the original
     * and with corresponding elements obtained by calling the
     * specified function.
     */
    public static Collection map(Collection c, Function1 f) {
	return map(c.getClass(), c, f);
    }

    /**
     * Returns a new collection of the specified result class and
     * with elements obtained by calling the specified function
     * on the corresponding elements of the original collection.
     */
    public static Collection map(Class resultClass,
				 Collection c,
				 Function1 f) {
	Debug.assert(Collection.class.isAssignableFrom(resultClass));
	Collection result = (Collection)Util.makeInstance(resultClass);
	for (Iterator i = c.iterator(); i.hasNext();) {
	    result.add(f.funcall(i.next()));
	}
	return result;
    }

    /**
     * Returns a new collection of the same class as the original
     * and containing only the elements for which the predicate
     * returned true;
     */
    public static Collection filter(Collection c, Predicate1 p) {
	return filter(c.getClass(), c, p);
    }

    /**
     * Returns a new collection of the specified result class and
     * and containing only the elements for which the predicate
     * returned true;
     */
    public static Collection filter(Class resultClass,
				    Collection c,
				    Predicate1 p) {
	Debug.assert(Collection.class.isAssignableFrom(resultClass));
	Collection result = (Collection)Util.makeInstance(resultClass);
	for (Iterator i = c.iterator(); i.hasNext();) {
	    Object elt = i.next();
	    if (p.trueOf(elt))
		result.add(elt);
	}
	return result;
    }

}
