/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Fri Feb 22 19:43:50 2002 by Jeff Dalton
 * Copyright: (c) 2002, AIAI, University of Edinburgh
 */

package ix.util;

import java.lang.reflect.*;
import java.util.*;

import java.io.PrintStream;

/**
 * Used to visit the fields of an object.  A FieldWalker contains the
 * information needed to process instances of a given class field-by-field.
 * It also embodies decisions about which fields should be accessible
 * in this way, how their values are obtained or set, and the order
 * in which they should be visited.
 */

public class FieldWalker {

    protected Class objClass;
    protected List fields;		// list of FieldInfo objects
    protected Map index;		// maps field name to FieldInfo

    /**
     * Constructs a FieldWalker for the given class.
     */
    public FieldWalker(Class objClass) {
	this.objClass = objClass;
	this.fields = new LinkedList();
	this.index = new HashMap();
	collectFieldInfo();
    }

    /**
     * Returns a list of field descriptions.
     * @see FieldInfo
     */
    public List getFieldInfo() {
	return fields;
    }

    /**
     * Returns the information for the specified field, or else
     * <code>null</code> if the field does not exist or cannot
     * be accessed via this FieldWalker.
     */
    public FieldInfo getFieldInfo(String fieldName) {
	return (FieldInfo)index.get(fieldName);
    }

    /**
     * Calls the <code>visitField</code> method for every field.
     */
    public void walkFields(Object object, Object result) {
	Debug.assert(objClass.isInstance(object));
	for (Iterator i = fields.iterator(); i.hasNext();) {
	    FieldInfo info = (FieldInfo)i.next();
	    visitField(object, info, result);
	}
    }

    /**
     * Called once for each field by the <code>walkFields</code> method.
     * This <code>visitField</code> method does nothing and is usually
     * overridden in subclasses.
     */
    public void visitField(Object object, FieldInfo info, Object result) {
    }

    /**
     * Calls the "set" method for the field.
     */
    public void setField(Object object, String fieldName, Object newValue)
	throws NoSuchFieldException,
	       IllegalAccessException,
	       IllegalArgumentException,
	       InvocationTargetException {
	Debug.assert(objClass.isInstance(object));
	FieldInfo info = (FieldInfo)index.get(fieldName);
	if (info == null)
	    throw new NoSuchFieldException
		("FieldWalker does not recognize field " +
		 Util.quote(fieldName) + " in " + objClass);
	info.setValue(object, newValue);
    }

    /**
     * Prints a description of this FieldWalker's view of its source class
     * on the <code>System.out</code>.
     */
    public void describeClass() {
	describeClass(System.out);
    }

    /**
     * Prints a description of this FieldWalker's view of its source class
     * on the designated PrintStream.
     */
    public void describeClass(PrintStream out) {
	out.println(objClass + " has "
		    + (fields.isEmpty() ? "no fields" : "fields:"));
	for (Iterator i = fields.iterator(); i.hasNext();) {
	    FieldInfo info = (FieldInfo)i.next();
	    out.println(info.type.getName() + " " + info.name);
	    out.println("   getter: " + info.getter);
	    out.println("   setter: " + info.setter);
	}
    }

    protected void collectFieldInfo() {
	List fieldNames = (List)collectFieldNames(objClass, new LinkedList());
	Method[] methods = objClass.getMethods();

	// Create FieldInfo list indexed by field names.
	for (Iterator i = fieldNames.iterator(); i.hasNext();) {
	    String name = (String)i.next();
	    FieldInfo info = new FieldInfo(name);
	    fields.add(info);
	    index.put(name, info);
	}

	// Add information from "get" and "set" methods
	for (int i = 0; i < methods.length; i++) {
	    Method meth = methods[i];
	    String methName = meth.getName();
	    if (isGetName(methName)) {
		String fieldName = fieldNameFromGetName(methName); 
		FieldInfo info = (FieldInfo)index.get(fieldName);
		// Process "get" method if there is a corresponding field
		if (info != null) {
		    Class[] ptypes = meth.getParameterTypes();
		    // Ignore if has > 0 parameters
		    if (ptypes.length == 0) {
			// Debug.noteln("Method", meth);
			info.processGetMethod(meth);
		    }
		}
	    }
	    else if (isSetName(methName)) {
		String fieldName = fieldNameFromSetName(methName);
		FieldInfo info = (FieldInfo)index.get(fieldName);
		// Process "set" method if there is a corresponding field
		if (info != null) {
		    Class[] ptypes = meth.getParameterTypes();
		    Class rtype = meth.getReturnType();
		    // Ignore if doesn't have 1 parameter.
		    // Need to check return type is void. /\/
		    if (ptypes.length == 1) {
			// Debug.noteln("Method", meth);
			info.processSetMethod(meth);
		    }
		}
	    }
	}

	// Remove fields that lack required information
	// and determine the type of the others.
	for (Iterator i = fields.listIterator(); i.hasNext();) {
	    FieldInfo info = (FieldInfo)i.next();
	    if (info.getter == null || info.setter == null)
		i.remove();
	    else
		info.determineType();
	}
    }

    public static final boolean isGetName(String name) {
	return name.length() > 3 && name.startsWith("get");
    }

    public static final boolean isSetName(String name) {
	return name.length() > 3 && name.startsWith("set");
    }

    public String fieldNameFromGetName(String name) {
	return Strings.uncapitalize(name.substring(3));
    }

    public String fieldNameFromSetName(String name) {
	return Strings.uncapitalize(name.substring(3));
    }

    protected Collection collectFieldNames(Class c, Collection names) {

	// Debug.noteln("Collecting names from", c);

	// Recursively process superclass, if it exists
	Class sup = c.getSuperclass();
	if (sup != null)
	    collectFieldNames(sup, names);

	// Recursively process interfaces
	Class[] interfaces = c.getInterfaces();
	for (int i = 0; i < interfaces.length; i++) {
	    collectFieldNames(interfaces[i], names);
	}

	// Non-inherited fields from this class.
	Field[] localf = c.getDeclaredFields();
	for (int i = 0; i < localf.length; i++) {
	    String name = localf[i].getName();
	    if (!names.contains(name))
		names.add(name);
	}
	return names;
    }

    public static class FieldInfo {
	public String name;
	public Class type;
	public Method getter;
	public Method setter;

	public FieldInfo(String name) {
	    this.name = name;
	}

	public Object getValue(Object obj)
	    throws IllegalAccessException,
		   IllegalArgumentException,
		   InvocationTargetException {
	    return getter.invoke(obj, new Object[0]);
	}

	public void setValue(Object obj, Object newVal)
	    throws IllegalAccessException,
		   IllegalArgumentException,
		   InvocationTargetException {
	    setter.invoke(obj, new Object[] { newVal });
	}

	protected void processGetMethod(Method meth) {
	    // The tricky part is that there might be > 1.
	    if (getter == null)
		getter = meth;
	    else
		getter = moreSpecificMethod(meth, getter);
	}

	protected void processSetMethod(Method meth) {
	    // The tricky part is that there might be > 1.
	    if (setter == null)
		setter = meth;
	    else
		setter = moreSpecificMethod(meth, setter);
	}

	protected Method moreSpecificMethod(Method m1, Method m2) {
	    // This comparison is not always meaningful, but since
	    // Java is essentially single-inheritance, we should
	    // eventually get a method that is more specific than
	    // all the others in a reasonably meaningful sense,
	    // at least for getters (because there can be only
	    // one per class).
	    Class d1 = m1.getDeclaringClass();
	    Class d2 = m2.getDeclaringClass();
	    if (d1.isAssignableFrom(d2)) {
		// N.B. for all we know, d2.isAssignableFrom(d1).
		return m2;
	    }
	    else {
		// N.B. for all we know, neither is assignable from
		// the other.
		return m1;
	    }
	}

	/**
	 * Tries to determine a field's type by looking at the
	 * get and set methods.
	 *
	 * @throws RuntimeException if the methods do not agree.
	 */
	protected void determineType() {
	    // At this point, we should have both a getter and a setter,
	    // and we insist that they agree on the type.
	    // /\/: Maybe instead find the most specific one that
	    // does have the right return type.
	    Class rtype = getter.getReturnType();
	    Class ptype = setter.getParameterTypes()[0];
	    if (rtype == ptype)
		type = rtype;
	    else
		// Example of class that fails here: javax.swing.JFrame
		throw new RuntimeException
		    ("Methods " + getter + " and " + setter +
		     " do not agree on field type");
	}

    }

    // For debugging ...

    /**
     * Test loop that repeatedly asks the user for a class name, 
     * constructs a FieldWalker for the class, and asks the
     * walker to print a class description.
     */
    public static void main(String[] argv) {
	final ClassFinder cf = new ClassFinder();
	for (;;) {
	    String in = Util.askLine("Class name:");
	    if (in.equals("bye"))
		return;

	    Class c = cf.classForName(in);
	    System.out.println(c);
	    if (c == null) continue;

	    FieldWalker w = new FieldWalker(c);
	    w.describeClass(System.out);
	    System.out.println("");
	}
    }

}
