/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Sat Jun  8 16:32:33 2002 by Jeff Dalton
 * Copyright: (c) 2002, AIAI, University of Edinburgh
 */

package ix.util;

import java.lang.reflect.*;
import java.util.*;

/**
 * Function utilities
 *
 * @see Collect
 * @see Strings
 */
public final class Fn {

    private Fn() { }	// block instantiation

    /**
     * Calls the indicated method.
     */
    public static Object apply(Object object,
			       String methodName,
			       List arguments) {
	return apply(object, methodName, arguments.toArray());
    }

    /**
     * Calls the indicated method.
     */
    public static Object apply(Object object,
			       String methodName,
			       Object[] arguments) {
	Class c = object.getClass();
	try {
	    Class[] sig = new Class[arguments.length];
	    for (int i = 0; i < sig.length; i++) {
		sig[i] = arguments[i].getClass();
	    }
	    Method meth = c.getMethod(methodName, sig);
	    try {
		return meth.invoke(object, arguments);
	    }
	    catch (IllegalAccessException e) {
		Debug.noteException(e);
		throw new Error("Cannot access method " + methodName +
				" for class " + c);
	    }
	    catch (InvocationTargetException e) {
		Debug.noteException(e);
		throw new RuntimeException
		    ("Exception from method " +
		     methodName + " from class " + c + ": " +
		     e.getTargetException());
	    }
	}
	catch (NoSuchMethodException e) {
	    Debug.noteException(e);
	    throw new Error("Cannot find method " + methodName +
			    " for class " + c);
	}
    }

    /**
     * Calls the indicated method.
     */
    public static Object apply(Object object,
			       Method meth,
			       Object[] arguments) {
	try {
	    return meth.invoke(object, arguments);
	}
	catch (IllegalAccessException e) {
	    Debug.noteException(e);
	    throw new Error("Cannot access method " + meth);
	}
	catch (InvocationTargetException e) {
	    Debug.noteException(e);
	    throw new RuntimeException
		("Exception from method " + meth + ": " +
		 e.getTargetException());
	}
    }

    /**
     * Calls the indicated static method.
     */
    public static Object applyStatic(Class c,
				     String methodName,
				     List arguments) {
	return applyStatic(c, methodName, arguments.toArray());
    }

    /**
     * Calls the indicated static method.
     */
    public static Object applyStatic(Class c,
				     String methodName,
				     Object[] arguments) {
	try {
	    Class[] sig = new Class[arguments.length];
	    for (int i = 0; i < sig.length; i++) {
		sig[i] = arguments[i].getClass();
	    }
	    Method meth = c.getMethod(methodName, sig);
	    try {
		return meth.invoke(null, arguments);
	    }
	    catch (IllegalAccessException e) {
		Debug.noteException(e);
		throw new Error("Cannot access method " + methodName +
				" for class " + c);
	    }
	    catch (InvocationTargetException e) {
		Debug.noteException(e);
		throw new RuntimeException
		    ("Exception from method " +
		     methodName + " from class " + c + ": " +
		     e.getTargetException());
	    }
	}
	catch (NoSuchMethodException e) {
	    Debug.noteException(e);
	    throw new Error("Cannot find method " + methodName +
			    " for class " + c);
	}
    }

    /**
     * Returns the specified method if it exists, else null.
     */
    public static Method getMethod(Object object,
				   String methodName,
				   Object[] arguments) {
	Class c = object.getClass();
	try {
	    Class[] sig = new Class[arguments.length];
	    for (int i = 0; i < sig.length; i++) {
		sig[i] = arguments[i].getClass();
	    }
	    Method meth = c.getMethod(methodName, sig);
	    return meth;
	}
	catch (NoSuchMethodException e) {
	    return null;
	}
    }

    /**
     * Returns the specified method if it exists, else null.
     */
    public static Method getMethod(Object object,
				   String methodName,
				   Class[] sig) {
	Class c = object.getClass();
	try {
	    Method meth = c.getMethod(methodName, sig);
	    return meth;
	}
	catch (NoSuchMethodException e) {
	    return null;
	}
    }

    /**
     * Returns a function that calls the 0-argument method specified
     * by a class and method name.
     */
    public static Function1 accessor(final Class c,
				     final String methodName) {
	final Class[] empty = new Class[] {}; // doubles as sig and arglist
	try {
	    final Method meth = c.getMethod(methodName, empty);
	    return new Function1() {
	        public Object funcall(Object a) {
		    try { 
			return meth.invoke(a, empty);
		    }
		    catch (IllegalAccessException e) {
			Debug.noteException(e);
			throw new Error("Cannot access method " + methodName +
					" for class " + c);
		    }
		    catch (InvocationTargetException e) {
			Debug.noteException(e);
			throw new RuntimeException
			    ("Exception from method " +
			     methodName + " from class " + c + ": " +
			     e.getTargetException());
		    }
		}
	    };
	}
	catch (NoSuchMethodException e) {
	    Debug.noteException(e);
	    throw new Error("Cannot find method " + methodName +
			    " for class " + c);
	}
    }

    /**
     * Returns a predicate that returns false when the given
     * predicate returns true, and true when it returns false.
     */
    public static Predicate1 negate(final Predicate1 p) {
	return new Predicate1() {
	    public boolean trueOf(Object a) {
		return !p.trueOf(a);
	    }
	};
    }

}
