/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Thu Apr  4 16:21:48 2002 by Jeff Dalton
 * Copyright: (c) 1998 - 2002, AIAI, University of Edinburgh
 */

package ix.util;

import java.util.*;

import ix.util.lisp.*;

/**
 * A class containing useful static string methods.
 */
public final class Strings {

    private Strings() {}	// block instantiation

    /**
     * Returns the index of the first occurrence of any of the specified
     * characters, or -1 if none are found.
     *
     * @param chars  the characters to look for
     * @param s      the string to look in
     */
    public static int indexOfAny(String chars, String s) {
	return indexOfAny(chars, 0, s);
    }

    /**
     * Returns the index of the first occurrence of any of the specified
     * characters, or -1 if none are found.
     *
     * @param chars  the characters to look for
     * @param start  index to start looking
     * @param s      the string to look in
     */
    public static int indexOfAny(String chars, int start, String s) {
	for (int i = start, len = s.length(); i < len; i++) {
	    int found = chars.indexOf(s.charAt(i));
	    if (found != -1)
		return i;
	}
	return -1;
    }

    /**
     * breakAtFirst takes a string containing fields separated by
     * a (string) delimiter and returns a two-element string array containing
     * the substring before the first occurrence of the char and the
     * substring after.  Neither substring contains the delimiter.  If
     * the delimiter does not appear in the string at all, the values
     * are the string and "".
     *
     * @param separator  the delimiter
     * @param s          the string that may contain it
     */
    public static String[] breakAtFirst(String separator, String s) {
	int i = s.indexOf(separator);
	if (i == -1)
	    return new String[]{s, ""};
	else
	    return new String[]{
		s.substring(0, i),
	        s.substring(i + separator.length())
	    };
    }

    /**
     * Returns a list of the substrings delimited by the given
     * separator.  The separator itself does not appear in any
     * element of the list, and if the separator does not occur,
     * the reault is a one-element list containing the string as-is.
     *
     * @param separator  the delimiter
     * @param s          the string that may contain it
     */
    public static List breakAt(String separator, String s) {
	// /\/: This uses breakAtFirst and so is less efficient than
	// it could be.
	// /\/: If the string ends with the spearator, there's
	// no final empty string in the result.
	LListCollector substrings = new LListCollector();
	while (!s.equals("")) {
	    String[] parts = breakAtFirst(separator, s);
	    substrings.add(parts[0]);
	    s = parts[1];
	}
	return substrings.contents();
    }

    /**
     * Returns the substring that ends directly before the first
     * occurrence of the separator.  If the separator does not
     * occur, the entire string is returned as-is.
     *
     * @param separator  the delimiter
     * @param s          the string that may contain it
     */
    public static String beforeFirst(String separator, String s) {
	int i = s.indexOf(separator);
	return (i == -1) ? s : s.substring(0, i);
    }

    /**
     * Returns the substring starting directly after the last
     * occurrence of the separator.  If the separator does not
     * occur, the entire string is returned as-is.
     *
     * @param separator  the delimiter
     * @param s          the string that may contain it
     */
    public static String afterLast(String separator, String s) {
	int i = s.lastIndexOf(separator);
	return (i == -1) ? s : s.substring(i + separator.length());
    }

    /**
     * Returns the first line of a string.
     */
    public static String firstLine(String s) {
	// For some reason, it doesn't work on Windows to use
	// System.getProperty("line.separator") here.  /\/
	return beforeFirst("\n", s);
    }

    /**
     * Returns a List of the lines in a string.
     */
    public static List breakIntoLines(String text) {
	// For some reason, it doesn't work on Windows to
	// use lineSpearator here.  /\/
	return breakAt("\n", text);
    }

    /**
     * Replaces some spaces with line separators to make a long
     * string more readable in contexts where it would have been
     * displayed as a single line.
     */
    public static String foldLongLine(String text) {
	List words = Strings.breakAt(" ", text);
	String newline = System.getProperty("line.separator");
	StringBuffer result = new StringBuffer();
	int maxLen = 75;
	int len = 0;
	for (Iterator wi = words.iterator(); wi.hasNext();) {
	    String word = (String)wi.next();
	    if (len + word.length() > maxLen) {
		result.append(newline);
		len = 0;
	    }
	    result.append(word).append(" ");
	    len += word.length();
	}
	return result.toString();
    }

    /**
     * Returns a string formed by replacing every occurrence
     * of <i>from</i> with <i>to</i> in <i>source</i>.  If any
     * replacements are made, the result is a new string without
     * any modifications to the original.  If, on the other hand,
     * <i>from</i> does not occur, the original string is
     * returned rather than a copy.
     *
     * @param from    the text to replace
     * @param to      the text to replace it with
     * @param source  the text in which to do the replacing
     *
     * @return  a new string if any replacements were needed;
     *          otherwise the original source string.
     */
    public static String replace(String from, String to, String source) {
	int i = source.indexOf(from);
	if (i < 0)
	    return source;
	StringBuffer result = new StringBuffer(source.length());
	int skip = from.length();
	int start = 0;
	while (i >= 0) {
	    result.append(source.substring(start, i))
		  .append(to);
	    start = i + skip;
	    i = source.indexOf(from, start);
	}
	result.append(source.substring(start));
	return result.toString();
    }

    /**
     * Puts double quotes around a string.  This is useful to indicate
     * the boundaries of the string when it's included in other text
     * such as in an error or warning message.
     */
    public static String quote(String text) {
	return "\"" + text + "\"";
    }

    /**
     * Returns a copy of the string in which the first character
     * is in upper case and the rest of the string is left as it was.
     */
    public static String capitalize(String s) {
	return s.substring(0,1).toUpperCase() +  s.substring(1);
    }

    /**
     * Returns a copy of the string in which the first character
     * is in lower case and the rest of the string is left as it was.
     */
    public static String uncapitalize(String s) {
	return s.substring(0,1).toLowerCase() +  s.substring(1);
    }

    /**
     * Returns a String made by appending a specified string <tt>count</tt>
     * times.
     */
    public static String repeat(int count, String s) {
	StringBuffer buf = new StringBuffer();
	for (int i = 0; i < count; i++)
	    buf.append(s);
	return buf.toString();
    }

    /**
     * Converts a name in which words are separate by dashes
     * into one that uses Java-style capitalisation instead.
     */
    public static String dashNameToJavaName(String name) {
	List words = breakAt("-", name);
	StringBuffer result = new StringBuffer(name.length());
	for (Iterator i = words.iterator(); i.hasNext();) {
	    String word = (String)i.next();
	    result.append(capitalize(word));
	}
	return result.toString();
    }

    /**
     * Converts a name that shows word boundaries using Java-style
     * capitalization to a name in which words are all lower case
     * and separated by dashes.
     *
     * @see JavaNameWordIterator
     */
    public static String javaNameToDashName(String name) {
	StringBuffer result = new StringBuffer(name.length());
	for (Iterator i = new JavaNameWordIterator(name); i.hasNext();) {
	    String word = (String)i.next();
	    if (word.length() > 1
		   && Character.isUpperCase(word.charAt(0))
		   && Character.isUpperCase(word.charAt(1)))
		// If 2 upper-case in a row, leave word as-is
		result.append(word);
	    else
		// else convert to lower case.
		result.append(word.toLowerCase());
	    if (i.hasNext())
		result.append("-");
	}
	return result.toString();
    }

}
