/* File: Util.java
 * Contains: A class for useful static methods
 * Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Created: January 1998
 * Updated: Tue May 28 00:28:01 2002 by Jeff Dalton
 * Copyright: (c) 1998 - 2002, AIAI, University of Edinburgh
 */

package ix.util;

import java.lang.reflect.*;
import java.util.*;
import java.io.*;
import java.net.*;

import java.awt.Component;
import javax.swing.*;

import ix.util.lisp.*;

/**
 * Class for useful static methods that don't belong anywhere else.
 *
 * @see Collect
 * @see Fn
 * @see Strings
 */

public class Util {

    /** 
     * A method to print the name of the system, the release version,
     * and the release date.
     */
    public static void printGreeting(String name) {
	System.out.println(name
			   + ", I-X version " + ix.Release.version
			   + ", " + ix.Release.date);
	System.out.println("");
    }

    /** 
     * Asks the GUI event-handling thread to displays a text message
     * in a dialogue box.
     *
     * <p>Always calls <code>SwingUtilities.invokeLater(Runnable)</code>,
     * which puts the request at the end of the queue even if called
     * from the event dispatching thread.
     *
     * @param parentComponent determines the Frame in which dialogs
     *        are displayed.
     * @param message the contents of the message.
     */
    public static void displayMessage(final java.awt.Component parentComponent,
				      final Object message) {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                JOptionPane.showMessageDialog
		    (parentComponent,
		     message,
		     "Message",
		     JOptionPane.INFORMATION_MESSAGE);
            }
        });
    }

    /** 
     * Asks the GUI event-handling thread to displays a text message
     * in a dialogue box and does not return until this has occurred.
     *
     * <p>If this method is called outside the event dispatching
     * thread, it calls <code>SwingUtilities.invokeAndWait(Runnable)</code>.
     *
     * @param parentComponent determines the Frame in which dialogs
     *        are displayed.
     * @param text the contents of the message.
     */
    public static void displayAndWait(final java.awt.Component parentComponent,
				      final Object message) {
	Runnable run = new Runnable() {
            public void run() {
                JOptionPane.showMessageDialog
		    (parentComponent,
		     message,
		     "Message",
                     JOptionPane.INFORMATION_MESSAGE);
            }
        };
	if (SwingUtilities.isEventDispatchThread())
	    run.run();
	else {
	    try {
		SwingUtilities.invokeAndWait(run);
	    }
	    catch (Throwable t) {
	    }
	}
    }

    /**
     * Brings up a message that the given string item is not yet supported.
     * e.g. "Deleting a schema is not yet supported"
     */
    public static void notImplemented(Component parent, String item){
      JOptionPane.showMessageDialog(parent, 
				    item + " is not yet supported.");
    }

    /**
     * Displays a confirmation dialog.
     */
    public static boolean dialogConfirms(Component parent, String text) {
	switch(JOptionPane.showConfirmDialog(parent,
	          text, "Confirm", JOptionPane.YES_NO_OPTION)) {
	case JOptionPane.YES_OPTION:
	    return true;
	case JOptionPane.NO_OPTION:
	    return false;
	}
	throw new Error("Confirm dialog failed to return YES or NO");
    }

    /**
     * Constructs an ImageIcon from the named (image) resource found
     * on the system's class path.  If the name contains "/", it is
     * used as-is; else "resources/images/" is added to the front.
     */
    public static ImageIcon resourceImageIcon(String name) {
	String fullName = name.indexOf('/') > -1
	    ? name
	    : "resources/images/" + name;
	// URL imageLocation = ClassLoader.getSystemResource(fullName);
	URL imageLocation = Util.class.getClassLoader().getResource(fullName);
	if (imageLocation != null) {
	    Debug.noteln("Making ImageIcon from", imageLocation);
	    return new ImageIcon(imageLocation);
	}
	else {
	    throw new RuntimeException("Can't find image resource named "
				       + Strings.quote(name));
	}
    }

    /**
     * Constructs an ImageIcon from the specified file in the "images"
     * subdirectory of the I-X "resources" directory.
     *
     * @deprecated As of April 1, 2002;
     * try {@link #resourceImageIcon(String)}.
     */
    public static ImageIcon resourceImageIconFromFile(String filename) {
	String base = System.getProperty("ix.base", ".");
	File resources = new File(base, "resources");
	File images = new File(resources, "images");
	File imageLocation = new File(images, filename);
	Debug.noteln("Making ImageIcon from", imageLocation);
	try {
	    return new ImageIcon(imageLocation.toURL());
	}
	catch (java.net.MalformedURLException e) {
	    Debug.noteException(e);
	    throw new RuntimeException("Can't handle image filename "
				       + filename);
	}
    }


    /**
     * Returns the name of the user who started this JVM.
     */
    public static String getUserName() {
	return System.getProperty("user.name");
    }


    /**
     * Returns the name of the machine that is running this JVM.
     */
    public static String getHostName() throws UnknownHostException {
	return Parameters.haveParameter("host")
	    ? Parameters.getParameter("host")
	    // N.B. the following call might fail.
	    : InetAddress.getLocalHost().getHostName();
    }


    /**
     * Puts double quotes around a string.  This is useful to indicate
     * the boundaries of the string when it's included in other text
     * such as in an error or warning message.
     */
    public static String quote(String text) {
	return "\"" + text + "\"";
    }


    /**
     * Name generator a la gensym.
     */
    public static class NameGenerator {
	protected Map counters = new HashMap();
	public NameGenerator() { }
	public String nextName(String base) {
	    Long count = (Long)counters.get(base);
	    if (count == null) {
		count = new Long(0);
		counters.put(base, count);
	    }
	    long c = count.longValue();
	    counters.put(base, new Long(c + 1));
	    return base + "-" + c;
	}
    }

    /**
     * The NameGenerator used by the <code>generateName(String)</code>
     * method.
     */
    private static NameGenerator defaultNameGenerator = new NameGenerator();

    /**
     * Generates a name unique within this VM that begins with the
     * specified base string.  (Note that "unique" means unique among
     * the names generated in this way; the same string might, of course,
     * be constructed in some other way.)
     */
    public static String generateName(String base) {
	return defaultNameGenerator.nextName(base);
    }


    /**
     * The ObjectCopier used by the copy methods in this class.
     */
    private static ObjectCopier defaultObjectCopier = new ObjectCopier();

    /**
     * Returns a top-level copy of the object.
     *
     * @see ObjectCopier
     */
    public static Object copy(Object obj)
	   throws InstantiationException,
	          IllegalAccessException,
		  InvocationTargetException {
	return defaultObjectCopier.copy(obj);
    }

    /**
     * Returns a top-level copy as an instance of the specified result
     * class.
     *
     * @see ObjectCopier
     */
    public static Object copy(Object obj, Class resultClass)
	   throws InstantiationException,
	          IllegalAccessException,
		  InvocationTargetException {
	return defaultObjectCopier.copy(obj, resultClass);
    }

    /**
     * Makes an instance of a class using the 0-argument constructor.
     *
     * @throws Error if the attempt fails.
     */
    public static Object makeInstance(Class c) {
	try {
	    return c.newInstance();
	}
	catch (Exception e) {
	    Debug.noteException(e);
	    throw new Error
		("Cannot make 0-arg instance of " + c + " because " + e);
	}
    }


    /**
     * Print the elements of a String[] array to System.out as lines.
     */
    public static void printLines(String[] lines) {
	for (int i = 0; i < lines.length; i++) {
	    System.out.println(lines[i]);
	}
    }


    /**
     * Simple, text-based user interaction.  askLine prints a prompt
     * to System.out and returns a String containing the next line
     * from System.in.<p>
     *
     * If askLine blocks when reading, we'd like other threads to be
     * able to run; but that doesn't seem to happen reliably.  Presumably,
     * this is a bug.  In any case, askLine works around the problem by
     * having a loop that checks whether input is available and sleeps
     * for a second if it isn't.
     */
    public static String askLine(String prompt) {
	System.out.print(prompt + " ");
	System.out.flush();
	// /\/: Should we synchronize on System.in?
	try {
	    while (System.in.available() == 0) {
		try { Thread.sleep(1000); }
		catch (InterruptedException e) {}
	    }
	}
	catch (IOException e) {
	    Debug.noteException(e);
	    return "";
	}
	return Util.readLine(System.in);
    }


    /** 
     * Reads a line from an InputStream and returns it as a String.
     * In Java, we seem to have to write this ourself unless we wrap
     * a special stream (or Reader) around whatever we want to read.
     * Here we provide a static method, because that's easier to
     * mix with other operations.  The only InputStream method
     * called is read().
     */
    public static String readLine(InputStream is) {
	ByteArrayOutputStream line = new ByteArrayOutputStream(80);

    readLoop:
	while (true) {
	    int c;		// a character as a "byte"
	    try { c = is.read(); }
	    catch (IOException e) {
		Debug.warn("IOException reading line from " + is);
		break readLoop;
	    }
	    if (c == '\n') 	// newline
		break readLoop;
	    else if (c == -1)	// end of stream
		break readLoop;
	    else if (c == '\r') // CR
		;		// (ignore)
	    else
		line.write(c);
	}
	return line.toString();
    }


}

// Issues:
// * The quote method is the same as the one in Strings and should
//   probably be eliminated; but it is used in many places.
