/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Tue Apr  2 06:45:05 2002 by Jeff Dalton
 * Copyright: (c) 2001, AIAI, University of Edinburgh
 */

package ix.util.xml;

import java.util.*;

import java.io.IOException;
import java.io.StringReader;
import java.io.File;

// Imports for using JDOM
// import java.io.IOException;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Attribute;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;

import ix.util.lisp.*;
import ix.util.*;

/**
 * A class useful static XML utilities.
 */

public class XML {

    private XML() { }		// don't allow instantiation

    /**
     * The translator used by static methods in this class.
     */
    static XMLTranslator defaultTranslator = new ConversionXMLTranslator();

    static { 
	defaultTranslator.setOmitImplAttributes(true);
    }

    /*
     *     XML to Objects
     */

    /**
     * Converts a string of XML to an Object using the default translator.
     */
    public static Object objectFromXML(String xml) {
	return defaultTranslator.objectFromXML(xml);
    }

    // /\/: Maybe names should be objectFromXMLString, ...FromXMLFile, etc.

    /**
     * Converts a File of XML to an Object using the default translator.
     */
    public static Object objectFromFile(File file) {
	try {
	    SAXBuilder builder = new SAXBuilder(SAXDriverClass);
	    Document doc = builder.build(file);
	    return objectFromDocument(doc);
	}
	catch (Exception e) {
	    Debug.noteln("While parsing", e);
	    e.fillInStackTrace();
	    throw new XMLException("Problem with file " + file + ": " + e);
	}
    }

    /**
     * Converts a JDOM Document to an object using the default translator.
     */
    public static Object objectFromDocument(Document doc) {
	return defaultTranslator.objectFromDocument(doc);
    }

    /**
     * Converts a JDOM Element to an object using the default translator.
     */
    public static Object objectFromElement(Element elt) {
	return defaultTranslator.objectFromElement(elt);
    }

    /**
     * The class of the SAX parser used by the <code>parseXML</code>
     * method.
     */
    static String SAXDriverClass = "org.apache.xerces.parsers.SAXParser";

    /**
     * Converts a string of XML to a JDOM Document.
     */
    public static Document parseXML(String text) {
	StringReader reader = new StringReader(text);
	try {
	    SAXBuilder builder = new SAXBuilder(SAXDriverClass);
	    Document doc = builder.build(reader);
	    return doc;		// for now /\/
	}
	catch (Exception e) {
	    Debug.noteln("While parsing", e);
	    // Debug.noteException(e);
	    // Throw because we have nothing to return
	    // and stray nulls are a pain.
	    throw new XMLException("Can't parse because " + e);
	}
    }


    /*
     *     Objects to XML
     */

    /**
     * Converts an object to a string of XML using the default translator.
     */
    public static String objectToXMLString(Object obj) {
	return defaultTranslator.objectToXMLString(obj);
    }

    /**
     * Converts an object to a JDOM Document using the default translator.
     */
    public static Document objectToDocument(Object obj) {
	return defaultTranslator.objectToDocument(obj);
    }

    /**
     * An XML outputter created by makePrettyXMLOutputter().
     */
    public static XMLOutputter prettyXMLOutputter = makePrettyXMLOutputter();

    /**
     * Constructs a JDOM XMLOutputter that outputs a JDOM Document
     * in a nicely indented fashion.  For example:
     * <pre>
     *    XMLOutputter outputter = makePrettyXMLOutputter()
     *    outputter.output(doc, System.out);
     *    System.out.flush();
     * </pre>
     */
    public static XMLOutputter makePrettyXMLOutputter() {
	// Create an outputter with 2 space indent and newlines=true
	XMLOutputter outputter = new XMLOutputter("  ", true);
	// outputter.setTrimText(true); // deprecated in beta-7 & wrong anyway
	outputter.setTextNormalize(false);
	return outputter;
    }


    /**
     * Encodes ampersands and angle brackets so that they can appear
     * in XML without being mistaken for markup.
     */
    public static String encodeText(String text) {
	return encodeForXML(text, false);
    }

    /**
     * Decodes text containing encoded ampersands and angle brackets.
     * This is the inverse of <code>encodeText(String)</code>.
     */
    public static String decodeText(String text) {
	return decodeForXML(text);
    }

    /**
     * Encodes double quotes as well as ampersands and angle brackets
     * to make text safe for use as XML attribute values.
     */
    public static String encodeAttribute(String text) {
	return encodeForXML(text, true);
    }

    /**
     * Decodes text containing encoded double quotes, ampersands,
     * and angle brackets.
     * This is the inverse of <code>encodeAttribute(String)</code>.
     */
    public static String decodeAttribute(String text) {
	return decodeForXML(text);
    }

    private static String encodeForXML(String text, boolean isForAttribute) {
	final String special = isForAttribute ? "&<>\"" : "&<>";
	int where = Strings.indexOfAny(special, text);
	if (where < 0)
	    return text;
	int increase = 0;	// how many more chars needed in result
	while (where >= 0) {
	    increase += charEncoding(text.charAt(where)).length() - 1;
	    where = Strings.indexOfAny(special, where + 1, text);
	}
	// Now that we know how many characters are needed, we construct
	// the result in a StringBuffer.  /\/: Use char[] instead?
	int textLen = text.length();
	StringBuffer buf = new StringBuffer(textLen + increase);
	for (int i = 0; i < textLen; i++) {
	    char c = text.charAt(i);
	    if (special.indexOf(c) >= 0)
		buf.append(charEncoding(c));
	    else
		buf.append(c);
	}
	Debug.assert(buf.length() == textLen + increase);
	return buf.toString();
    }

    private static String charEncoding(char c) {
	// Could use an array, I suppose.  /\/
	switch ("&<>\"".indexOf(c)) {
	case 0:		// &
	    return "&amp;";
	case 1:		// <
	    return "&lt;";
	case 2:		// >
	    return "&gt;";
	case 3:		// double quote
	    return "&quot;";
	}
	throw new RuntimeException("No encoding for " + String.valueOf(c));
    }


    private static String decodeForXML(String text) {
	if (text.indexOf('&') < 0)
	    return text;
	int i = 0, textLen = text.length();
	StringBuffer buf = new StringBuffer(textLen); // use char[]? /\/
	while (i < textLen) {
	    char c = text.charAt(i);
	    if (c != '&') { buf.append(c); i++; }
	    else {
		char clue = text.charAt(i + 1);
		char decode;
		String code;
		if      (clue == 'a') { decode = '&'; code = "&amp;"  ; }
		else if (clue == 'l') { decode = '<'; code = "&lt;"   ; }
		else if (clue == 'g') { decode = '>'; code = "&gt;"   ; }
		else if (clue == 'q') { decode = '"'; code = "&quot;" ; }
		else
		    throw new RuntimeException("Cannot decode " + text);
		Debug.assert(text.charAt(i + code.length() - 1) == ';');
		buf.append(decode);
		i += code.length();
	    }
	}
	return buf.toString();
    }

    /**
     * Test loop.
     */
    public static void main(String[] argv) {
	for (;;) {
	    String text = Util.askLine("Text:");
	    if (text.equals("bye"))
		return;
	    String encoded = encodeAttribute(text);
	    Debug.noteln("Encode:", encoded);
	    Debug.noteln("Decode:", decodeAttribute(encoded));
	}
    }

}
