/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Sat Jun  8 16:38:13 2002 by Jeff Dalton
 * Copyright: (c) 2002, AIAI, University of Edinburgh
 */

package ix.util.xml;

import java.lang.reflect.*;
import java.util.*;

// import java.io.StringReader;
import java.io.StringWriter;
import java.io.IOException;

// Imports for using JDOM
// import java.io.IOException;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Attribute;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;

import ix.util.*;
import ix.util.lisp.*;

/**
 * A class for translating between objects and XML.  Note that a
 * single XMLTranslator maps in both directions to make it easier
 * to create mappings that are consistent. <p>
 */

public class XMLTranslator {

    protected ClassFinder classFinder = new ClassFinder();

    protected XMLOutputter outputter = XML.makePrettyXMLOutputter();

    protected Map walkerCache = new HashMap();

    protected String implAttribute = externalFieldName("implClass");

    protected boolean omitImplAttributes = false;

    protected Element lastConvertedElement = null; // for error reporting

    protected Class defaultListClass = LinkedList.class;
    protected Class defaultSetClass = HashSet.class;
    protected Class defaultMapClass = HashMap.class;

    /** Add initial imports. */
    {
	addImport("ix.icore.*");
	addImport("ix.icore.domain.*");
	addImport("ix.ichat.ChatMessage");
	addImport("ix.ip2.TestItem");
	addImport("ix.util.Name");
	addImport("ix.util.lisp.*");
    }

    public XMLTranslator() {
    }

    public ClassFinder getClassFinder() {
	return classFinder;
    }

    public XMLOutputter getOutputter() {
	return outputter;
    }

    /**
     * Controlls whether or not <code>implClass</code> attributes
     * are added to List, Set, and Map elements when converting to XML.
     * Note that if it's set to true, the XML will not contain enough
     * information to determine the original implementation classes
     * and the translator will have to do its best to make reasonable
     * choices based on field classes and the defaults.
     */
    protected void setOmitImplAttributes(boolean v) {
	omitImplAttributes = v;
    }

    protected void addImport(String name) {
	classFinder.addImport(name);
    }


    /* ************************** *
     *      Name conversions      *
     * ************************** */

    public Class classForExternalName(String externalName) {
	return classFinder.classForName(internalName(externalName));
    }

    public String externalNameForClass(Class c) {
	return externalName(classFinder.nameForClass(c));
    }

    public String internalFieldName(String externalName) {
	return internalName(externalName);
    }

    public String externalFieldName(String internalName) {
	return externalName(internalName);
    }

    public String internalName(String externalName) {
	return externalName;
    }

    public String externalName(String internalName) {
	return internalName;
    }

    /* ************************ *
     *      Objects to XML      *
     * ************************ */

    /**
     * Converts an object to a string of XML.
     */
    public String objectToXMLString(Object obj) {
	Document doc = objectToDocument(obj);
	StringWriter writer = new StringWriter();
	try {
	    outputter.output(doc, writer);
	}
	catch (IOException e) {
	    Debug.noteException(e);
	    // Throw because we have nothing meaningful to return
	    // and stray nulls or ""s are a pain.
	    throw new XMLException
		("Can't convert " + obj + " to string because " + e);
	}
	writer.flush();
	return writer.toString();
    }

    /**
     * Converts an object to a JDOM Document.
     */
    public Document objectToDocument(Object obj) {
	return new Document(objectToElement(obj));
    }

    /**
     * Converts an object to a JDOM Element.  This is the central
     * recursion point for the conversion process; it determines
     * which of the <code><i>type</i>ToElement(Object)</code> 
     * methods to call.
     */
    public Element objectToElement(Object obj) {
	if (obj == null)
	    return new Element("null");
	else if (obj instanceof SemiPrimitive)
	    return primitiveObjectToElement(obj);
	else if (obj instanceof Number)
	    return primitiveObjectToElement(obj);
	else if (obj instanceof String)
	    return stringToElement(obj);
	else if (obj instanceof Boolean)
	    return primitiveObjectToElement(obj);
	else if (obj instanceof List)
	    return listToElement(obj);
	else if (obj instanceof Set)
	    return setToElement(obj);
	else if (obj instanceof Map)
	    return mapToElement(obj);
	else
	    return structToElement(obj);
    }

    public Element primitiveObjectToElement(Object obj) {
	// This is for instances of the Object equivalents of the primitive
	// types -- of Long, for example, rather than Long.TYPE.
	Class c = obj.getClass();
	String cname = externalNameForClass(c);
	Element elt = new Element(cname);
	elt.setText(obj.toString());
	return elt;
    }

    public Element stringToElement(Object obj) {
	// JDOM encodes when the string contains XML.
	return primitiveObjectToElement(obj);
    }

    public Element listToElement(Object obj) {
	return collectionToElement(List.class, (List)obj);
    }

    public Element setToElement(Object obj) {
	return collectionToElement(Map.class, (Set)obj);
    }

    public Element collectionToElement(Class iface, Collection obj) {
	// If the elt type is always "List" or "Set" (with the class
	// as an attribute), rather than itself being the class name,
	// it will be clear from the XML that we are looking
	// at a list (without having to look up the class) -
	// and thus that we're breaking the strict alternation
	// between field and class names.
	Class c = obj.getClass();
	String cname = externalNameForClass(c);
	String iname = externalNameForClass(iface);
	Element elt = new Element(iname);
	if (!omitImplAttributes)
	    elt.setAttribute(implAttribute, cname);
	for (Iterator i = obj.iterator(); i.hasNext();) {
	    Object item = i.next();
	    elt.addContent(objectToElement(item));
	}
	return elt;
    }

    public Element mapToElement(Object obj) {
	Map map = (Map)obj;
	Class c = map.getClass();
	String cname = externalNameForClass(c);
	String iname = externalNameForClass(Map.class);
	Element mapElt = new Element(iname);
	if (!omitImplAttributes)
	    mapElt.setAttribute(implAttribute, cname);
	for (Iterator i = map.entrySet().iterator(); i.hasNext();) {
	    Map.Entry entry = (Map.Entry)i.next();
	    Element entryElt = new Element(externalName("MapEntry"));
	    Element keyElt = new Element("key");
	    Element valElt = new Element("value");
	    mapElt.addContent(entryElt);
	    entryElt.addContent(keyElt);
	    entryElt.addContent(valElt);
	    keyElt.addContent(objectToElement(entry.getKey()));
	    valElt.addContent(objectToElement(entry.getValue()));
	    // Maybe just alternate keys and values without
	    // entry elements?  /\/
	}
	return mapElt;
    }

    public Element structToElement(Object obj) {
	Class c = obj.getClass();
	FieldWalker w = ensureFieldWalker(c);
	String cname = externalNameForClass(c);
	Element elt = new Element(cname);
	w.walkFields(obj, elt);
	return elt;
    }

    protected FieldWalker ensureFieldWalker(Class c) {
	FieldWalker w = (FieldWalker)walkerCache.get(c);
	if (w == null) {
	    w = new WalkerForXML(c);
	    walkerCache.put(c, w);
	}
	return w;
    }

    /**
     * A FieldWalker for XML generation.
     */
    protected final class WalkerForXML extends FieldWalker {

	public WalkerForXML(Class objClass) {
	    super(objClass);
	}

	public void visitField(Object obj, FieldInfo info, Object result) {
	    Element objElt = (Element)result;
	    String fieldName = externalFieldName(info.name);
	    try {
		Object value = info.getValue(obj);
		if (isAttributeClass(info.type)) {
		    // Add attribute
		    if (value != null)
			objElt.setAttribute(fieldName, value.toString());
		}
		else if (value != null) {
		    // Add child element
		    Element fieldElt = new Element(fieldName);
		    objElt.addContent(fieldElt);
		    fieldElt.addContent(objectToElement(value));
		}
	    }
	    catch (Exception e) {
		Debug.noteException(e);
		throw new XMLException
		    ("Cannot get value of field " + info.name
		     + " of class " + obj.getClass().getName()
		     + " because " + e);
	    }
	}

    }

    public static boolean isAttributeClass(Class c) {
	return c.isPrimitive()
	    || SemiPrimitive.class.isAssignableFrom(c)
	    || Number.class.isAssignableFrom(c)
	    || Boolean.class.isAssignableFrom(c);
    }

    /* ************************** *
     *      Objects from XML      *
     * ************************** */

    /**
     * Converts a string of XML to an object.
     */
    public Object objectFromXML(String text) {
	Document doc = XML.parseXML(text);
	return objectFromDocument(doc);
    }

    /**
     * Converts a JDOM Document to an object.  This method, or one
     * that calls it, is what should normally be called from outside
     * the translator - rather than calling objectFromElement(elt)
     * directly.
     */
    public Object objectFromDocument(Document doc) {
	Object result = objectFromElement(doc.getRootElement());
	lastConvertedElement = null;	// no longer needed
	return result;
    }

    /**
     * Returns the JDOM Element most recently given to the
     * objectFromElement method.  This may be useful when handling
     * or reporting an exception.
     */
    public Element getLastConvertedElement() {
	return lastConvertedElement;
    }

    /**
     * Converts a JDOM Element to an object.  This is the central
     * recursion point of the translation process; it determines
     * which of the <code><i>type</i>FromElement</code> methods
     * to call.
     */
    public Object objectFromElement(Element elt) {
	lastConvertedElement = elt;
	String className = elt.getName();
	if (className.equals("null"))
	    return null;	// /\/
	Class c = classForExternalName(className);
	if (c == null)
	    throw new XMLException("Can't find class " + className);
	else
	    return objectFromElement(c, elt);
    }

    protected Object objectFromElement(Class c, Element elt) {
	lastConvertedElement = elt;
	if (SemiPrimitive.class.isAssignableFrom(c))
	    return primitiveObjectFromElement(c, elt);
	else if (Number.class.isAssignableFrom(c))
	    return primitiveObjectFromElement(c, elt);
	else if (String.class.isAssignableFrom(c))
	    return stringFromElement(c, elt);
	else if (Boolean.class.isAssignableFrom(c))
	    return primitiveObjectFromElement(c, elt);
	else if (List.class.isAssignableFrom(c))
	    return listFromElement(c, elt);
	else if (Set.class.isAssignableFrom(c))
	    return setFromElement(c, elt);
	else if (Map.class.isAssignableFrom(c))
	    return mapFromElement(c, elt);
	else
	    return structFromElement(c, elt);
    }

    public Object primitiveObjectFromElement(Class c, Element elt) {
	return primitiveObjectFromString(c, elt.getText());
    }

    public Object primitiveObjectFromString(Class c, String s) {
	try {
	    Method meth = c.getMethod("valueOf", new Class[]{String.class});
	    return meth.invoke(null, new Object[]{s});
	}
	catch (InvocationTargetException e) {
	    Debug.noteException(e);
	    throw new XMLException
		("Can't convert to " + c + " because "
		 + e.getTargetException());
	}
	catch (Exception e) {
	    Debug.noteException(e);
	    throw new XMLException
		("Can't convert to " + c + " because " + e);
	}
    }

    public String stringFromElement(Class c, Element elt) {
	return elt.getText();
    }

    public List listFromElement(Class c, Element elt) {
	return (List)collectionFromElement(c, defaultListClass, elt);
    }

    public Set setFromElement(Class c, Element elt) {
	return (Set)collectionFromElement(c, defaultSetClass, elt);
    }

    public Collection collectionFromElement(Class c,
					    Class implDefault,
					    Element elt) {
	// If the element class, c, is an interface, get the name of
	// the implementing class from the implClass attribute.
	// Note that this also allows us to deal with XML in which
	// the element type is already the implementation class.
	Class implClass = c.isInterface()
	    ? getImplClass(elt, implDefault)
	    : c;

	// If the desired class is an LList, it's a bit of a special
	// case, because the actual class depends on whether the
	// list is empty or not, so we can't just call the usual
	// 1-arg constructor.  /\/
	if (LList.class.isAssignableFrom(implClass))
	    return makeLListFromElement(elt);

	// /\/: We used to create the result as an empty collection
	// and then add elements to it, but that wouldn't work
	// if the result class was for a fixed-size collection.
	List contents = new LinkedList();
	for (Iterator i = elt.getChildren().iterator(); i.hasNext();) {
	    Element childElt = (Element)i.next();
	    Object childObj = objectFromElement(childElt);
	    // col.add(childObj);
	    contents.add(childObj);
	}

	// If we've already built a collection of the desired class,
	// we are done.
	if (contents.getClass() == implClass)
	    return contents;

	// Otherwise, construct an instance of the correct class
	// and add the contents by calling the 1-arg constructor
	// we assume is standard with Collections.
	try {
	    Constructor cons =
		implClass.getConstructor(new Class[] {Collection.class});
	    return (Collection)cons.newInstance(new Object[] {contents});
	}
	catch (Exception e) {
	    Debug.noteException(e);
	    throw new XMLException
		("Could not create " + implClass + " collection because " + e);
	}
    }

    public LList makeLListFromElement(Element elt) {
	// While we're here, we make sublists default to LLists.
	LListCollector col = new LListCollector();
	for (Iterator i = elt.getChildren().iterator(); i.hasNext();) {
	    Element childElt = (Element)i.next();
	    String childClassName = childElt.getName();
	    if (childClassName.equals("null")) {
		// Irritating special case /\/
		col.add(null);
		continue;
	    }
	    Class childClass = classForExternalName(childClassName);
	    Object childObj = 
		childClass == List.class
		    && childElt.getAttributeValue(implAttribute) == null
		? makeLListFromElement(childElt)
		: objectFromElement(childElt);
	    col.add(childObj);
	}
	return col.contents();
    }

    public Class getImplClass(Element elt, Class implDefault) {
	String iname = elt.getAttributeValue(implAttribute);
	Class implClass = iname == null
	    ? implDefault
	    : classForExternalName(iname);
	if (implClass == null)
	    throw new XMLException
		("Cannot find implementation class " + iname);
	return implClass;
    }

    public Map mapFromElement(Class c, Element elt) {
	Class implClass = c.isInterface()
	    ? getImplClass(elt, defaultMapClass)
	    : c;
	Map map = (Map)makeInstance(implClass);
	String entryEltName = externalName("MapEntry");
	for (Iterator i = elt.getChildren().iterator(); i.hasNext();) {
	    Element entryElt = (Element)i.next();
	    requireEltType(entryEltName, entryElt);
	    requireEltSize(2, entryElt);
	    List entryChildren = entryElt.getChildren();
	    Element keyElt = (Element)entryChildren.get(0);
	    Element valElt = (Element)entryChildren.get(1);
	    requireEltType("key", keyElt);
	    requireEltType("value", valElt);
	    requireEltSize(1, keyElt);
	    requireEltSize(1, valElt);
	    Element keyObjElt = (Element)keyElt.getChildren().get(0);
	    Element valObjElt = (Element)valElt.getChildren().get(0);
	    Object key = objectFromElement(keyObjElt);
	    Object val = objectFromElement(valObjElt);
	    map.put(key, val);
	}
	return map;
    }

    protected void requireEltType(String name, Element elt) {
	if (!elt.getName().equals(name))
	    throw new XMLException
		("Expecting " + name + " element but found " + elt.getName());
    }

    protected void requireEltSize(int size, Element elt) {
	int actual = elt.getChildren().size();
	if (actual != size)
	    throw new XMLException
		("Expecting " + size + " children but found " + actual +
		 " in " + elt);
    }

    public Object structFromElement(Class c, Element elt) {
	// /\/: Should check for fields specified > once.
	Object obj = makeInstance(c);
	FieldWalker walker = ensureFieldWalker(c);
	// Process attributes
	for (Iterator i = elt.getAttributes().iterator(); i.hasNext();) {
	    Attribute fieldAttr = (Attribute)i.next();
	    String fieldName = internalFieldName(fieldAttr.getName());
	    String fieldValue = fieldAttr.getValue();
	    FieldWalker.FieldInfo fieldInfo = walker.getFieldInfo(fieldName);
	    if (fieldInfo == null)
		throw new XMLException
		    ("No field for attribute " + fieldAttr.getName());
	    else try {
		Class fieldClass = fieldInfo.type.isPrimitive()
		    ? wrapperClass(fieldInfo.type)
		    : fieldInfo.type;
		Object fieldObj = fieldClass == String.class
		    ? fieldValue
		    : primitiveObjectFromString(fieldClass, fieldValue);
		fieldInfo.setValue(obj, fieldObj);
	    }
	    catch (Exception e) {
		Debug.noteException(e);
		throw new XMLException
		    ("Cannot set field " + fieldName + " because " + e);
	    }
	}
	// Process subelements
	for (Iterator i = elt.getChildren().iterator(); i.hasNext();) {
	    Element fieldElt = (Element)i.next();
	    requireEltSize(1, fieldElt);
	    String fieldName = internalFieldName(fieldElt.getName());
	    FieldWalker.FieldInfo fieldInfo = walker.getFieldInfo(fieldName);
	    Element valElt = (Element)fieldElt.getChildren().get(0);
	    String valClassName = valElt.getName();
	    Class valClass;
	    Object fieldVal;
	    if (!valClassName.equals("null") // irritating special case /\/
		  && (valClass = classForExternalName(valClassName))
		     != null
		  && valClass.isInterface()
		  && valElt.getAttributeValue(implAttribute) == null
		  && !fieldInfo.type.isInterface()) {
		// Looks like we should try for an instance of
		// the field class.
		fieldVal = objectFromElement(fieldInfo.type, valElt);
	    }
	    else
		fieldVal = objectFromElement(valElt);
	    try {
		walker.setField(obj, fieldName, fieldVal);
	    }
	    catch (Exception e) {
		Debug.noteException(e);
		throw new XMLException
		    ("Cannot set field " + fieldName + " because " + e);
	    }
	}
	return obj;
    }

    /**
     * Returns the wrapper class for a primitive class, such as
     * <code>Long.class</code> for <code>long.class</code>.
     */
    public static final Class wrapperClass(Class pc) {
	// /\/: Is there really no better way to do this?
	if (pc == byte.class) return Byte.class;
	else if (pc == char.class) return Character.class;
	else if (pc == short.class) return Short.class;
	else if (pc == int.class) return Integer.class;
	else if (pc == long.class) return Long.class;
	else if (pc == float.class) return Float.class;
	else if (pc == double.class) return Double.class;
	else if (pc == boolean.class) return Boolean.class;
	else if (pc == void.class) return Void.class;
	else
	    throw new IllegalArgumentException
		(pc + " is not a primitive class");
    }

    /**
     * Makes an instance of a class using the 0-argument constructor.
     *
     * @throws XMLException if the attempt fails.
     */
    public Object makeInstance(Class c) {
	try {
	    return c.newInstance();
	}
	catch (Exception e) {
	    Debug.noteException(e);
	    throw new XMLException
		("Cannot make 0-arg instance of " + c + " because " + e);
	}
    }


    /* ***************** *
     *      Testing      *
     * ***************** */

    // N.B. It's important that this inlude the implClass attributes
    // in the XML so that we can use it to see what sort of object
    // we've actually got when testing a translator that omits
    // the impl class information.
    protected static XMLTranslator standardTranslator = new XMLTranslator();

    protected static String standardXML(Object obj) {
	return standardTranslator.objectToXMLString(obj);
    }

    protected static void test(XMLTranslator xmlt, Object obj) {
	test(xmlt, obj, null);
    }

    protected static void test(XMLTranslator xmlt,
			       Object obj,
			       Function2 look) {
	try {
	    Debug.noteln("Object:", obj);
	    Debug.noteln("Class:", obj.getClass());

	    String xml = xmlt.objectToXMLString(obj);
	    Debug.noteln("XML:", xml);

	    Object back = xmlt.objectFromXML(xml);
	    Debug.noteln("Back to:", back);
	    Debug.noteln("Class:", back.getClass());
	    Debug.noteln("obj.equals(back)? " + obj.equals(back));
	    Debug.noteln("back.equals(obj)? " + back.equals(obj));
	    Debug.noteln("Standard XML:", standardXML(back));
	    if (look != null)
		look.funcall(obj, back);
	}
	catch (Exception e) {
	    Debug.noteException(e);
	}
	Debug.noteln("- - - - - - - - - - - - - - - - - - - -");
    }

    /**
     * Simple main program used for testing.
     */
    public static void main(String[] argv) {
	do_main(new XMLTranslator());
    }

    public static void do_main(XMLTranslator xmlt) {

	Debug.noteln("----- With implClass attributes");

	test(xmlt, new Long(123));

	test(xmlt, Lisp.NIL);

	Debug.noteln("----- Without implClass attributes");
	xmlt.setOmitImplAttributes(true);

	test(xmlt, new Long(123));

	test(xmlt, Lisp.NIL);

	test(xmlt, Lisp.readFromString("(10 11 12)"));

	test(xmlt, new LinkedList((LList)Lisp.readFromString("(10 11 12)")));

	ix.icore.Issue issue = new ix.icore.Issue("Note this");
	issue.setPriority(ix.icore.Priority.HIGH);
	issue.setRef(Name.valueOf("ref-1"));
	issue.setSenderId(Name.valueOf("no one"));
	test(xmlt, issue);

	List ll = new LinkedList();
	ll.add("a"); ll.add("b"); ll.add("c");
	test(xmlt, ll);

	Map mm = new HashMap();
	mm.put("a", "a-val");
	mm.put("b", "b-val");
	test(xmlt, mm);

	test(xmlt, "String with <boo f=\"x\">embedded XML &c</boo>");

	// Try double quote in attribute
	issue.setSenderId(Name.valueOf("probably \"no one\""));
	test(xmlt, issue, new Function2() {
	    public Object funcall(Object issue, Object back) {
		Debug.noteln("Decoded senderId =",
			     ((ix.icore.Issue)back).getSenderId());
		return null;
	    };
	});

	test(xmlt, ix.icore.Status.COMPLETE);

	test(xmlt, new ix.ichat.ChatMessage("Hello", "sender-1"));

	test(xmlt, Lisp.readFromString("(a nil b)"));

	test(xmlt, new ix.icore.domain.NodeSpec("node-1", Lisp.NIL));

    }

}

// Issues:
// * For circularity and sharing, give objects that are shared an
//   attribute id="someid" and refer to it using an attribute
//   ref="someref" and then declare these attribute names as
//   type ID and IDREF respectively.  An element is allowed only
//   one attribute of type ID.
