/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Sat Jun  8 15:27:50 2002 by Jeff Dalton
 * Copyright: (c) 2002, AIAI, University of Edinburgh
 */

package ix.util.xml;

import javax.swing.*;

import java.awt.Container;
import java.awt.Component;
import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.event.*;

import java.io.*;

import java.util.*;

// Imports for using JDOM
// import java.io.IOException;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Attribute;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;

import ix.icore.IXAgent;

import ix.iface.domain.DomainParser;
import ix.iface.domain.LTF_Parser;

import ix.iface.util.CatchingActionListener;

import ix.iface.util.*;
import ix.util.*;
import ix.util.lisp.*;

/** 
 * A frame that contains an XML editing panel.
 * 
 * @see XMLTreeEditor
 * @see XMLTreeEditPanel
 */
public class XMLTreeEditFrame extends JFrame implements ActionListener {

    protected IXAgent agent;

    protected Container contentPane = getContentPane();

    protected XMLTreeEditPanel editPanel;
    
    XMLTreeEditFrame(IXAgent agent, String title) {
	// Default is BorderLayout.
	super(title);
	this.agent = agent;

	setIconImage(IconImage.getIconImage(this));
	setJMenuBar(makeMenuBar());

	editPanel = new XMLTreeEditPanel();
	contentPane.add(editPanel, BorderLayout.CENTER);

	setSize(500, 600);	// width, height
    }

    protected boolean isStandalone() {
	return agent instanceof XMLTreeEditor;
    }

    /**
     * Creates the menu bar
     */
    protected JMenuBar makeMenuBar() {
	JMenuBar bar = new JMenuBar();

	// "File" menu
	JMenu fileMenu = new JMenu("File");
	bar.add(fileMenu);
	if (isStandalone()) {
	    fileMenu.add(makeMenuItem("Load From ..."));
	    fileMenu.add(makeMenuItem("Save As ..."));
	    fileMenu.add(makeMenuItem("Reset"));
	    fileMenu.add(makeMenuItem("Exit"));
	}
	else {
	    // Part of another agent
	    fileMenu.add(makeMenuItem("Reset"));
	    fileMenu.add(makeMenuItem("Close"));
	}

	// "Edit" menu
	JMenu editMenu = new JMenu("Edit");
	bar.add(editMenu);
	editMenu.add(makeMenuItem("Check As Document"));
	editMenu.add(makeMenuItem("Check As Object"));

	return bar;
    }

    // /\/: Must find a better way to do this rather than have
    // this method everywhere.
    protected JMenuItem makeMenuItem(String text) {
	JMenuItem item = new JMenuItem(text);
	item.addActionListener(CatchingActionListener.listener(this));
	return item;
    }

    /**
     * Action interpreter for items in the frame's "File" menu.
     */
    public void actionPerformed(ActionEvent e) {
	String command = e.getActionCommand();
	Debug.noteln("XMLTreeEditFrame action:", command);

	if (command.equals("Load From ...")) {
	    loadFrom();
	}
	else if (command.equals("Save As ...")) {
	    saveAs();
	}
	else if (command.equals("Reset")) {
	    reset();
	}
	else if (command.equals("Close")) {
	    setVisible(false);
	}
	else if (command.equals("Exit")) {
	    exitIfUserConfirms();
	}
	else if (command.equals("Check As Document")) {
	    editPanel.getDocument();
	}
	else if (command.equals("Check As Object")) {
	    editPanel.getObject();
	}
	else
	    Debug.noteln("Nothing to do for", command);
    }

    protected void reset() {
	editPanel.reset();
    }

    public void exitIfUserConfirms() {
	if (Util.dialogConfirms(null,
				"Are you sure you want to exit?"))
	    System.exit(0);
    }

    protected void loadFrom() {
	File libDir = DomainParser.getLibraryDirectory();

	// Loop until either the data has successfully been read
	// from a file or else the user has decided to "Cancel"
	// rather than try a(nother) file.
	while (true) {

	    // Ask user to select a file.
	    File file = chooseFileToRead(this, libDir);
	    if (file == null)
		return;			// user cancels

	    Document doc = readFrom(file);
	    if (doc != null) {
		editPanel.editDocument(doc);
		return;			// success!
	    }

	}

    }

    /**
     * Converts the contents of the specified file to a JDOM Document.
     * In addition to .xml files, it is possible to read .lsp files
     * that contain domain definitions in LTF syntax.
     * 
     * @see ix.iface.domain.LTF_Parser
     */
    protected Document readFrom(File file) {
	Debug.noteln("Editor reading from", file);
	try {
	    String name = file.getName();
	    if (name.endsWith(".lsp"))
		return readLTF(file);
	    else if (!name.endsWith(".xml"))
		throw new IllegalArgumentException
		    ("Don't know how to read " + file);
	    SAXBuilder builder = new SAXBuilder(XML.SAXDriverClass);
	    Document doc = builder.build(file);
	    return doc;
	}
	catch (Exception e) {
	    Debug.noteException(e);
	    JOptionPane.showMessageDialog(this,
	        new Object[] {
		     "Problem reading " + Strings.quote(file.getPath()),
		     e.getMessage()},
		"Exception",
		JOptionPane.ERROR_MESSAGE);
	    return null;
	}
    }

    protected Document readLTF(File file) 
	      throws java.io.FileNotFoundException {
	return XML.objectToDocument(new LTF_Parser(file).readDomain());
    }
    
    private static File chooseFileToRead(Component frame, File directory) {
	JFileChooser chooser = new JFileChooser(directory);
	chooser.setFileFilter(new XMLFileFilter());
	int option = chooser.showOpenDialog(frame);
	if (option == JFileChooser.APPROVE_OPTION)
	    return chooser.getSelectedFile();
	else
	    return null;
    }

    protected void saveAs() {
	Document doc = editPanel.getDocument();
	File libDir = DomainParser.getLibraryDirectory();

	// Loop until either the data has successfully been written
	// to a file or else the user has decided to "Cancel" rather
	// that select a(nother) file.
	while (true) {

	    // Ask user to select a file.
	    File file = chooseFileToWrite(this, libDir);
	    if (file == null)
		return;			// user cancels

	    // See if file already exists.
	    if (file.exists()) {
		switch(JOptionPane.showConfirmDialog(this,
			"Overwrite " + Strings.quote(file.getPath()),
			"Confirm",
			JOptionPane.YES_NO_OPTION)) {
		case JOptionPane.YES_OPTION:
		    // Continue
		    break;
		case JOptionPane.NO_OPTION:
		    // Return now and don't save.
		    return;
		}
	    }

	    // Return if written successfully.
	    if (writeTo(file, doc))
		return;
	}

    }

    protected boolean writeTo(File file, Document doc) {
	Debug.noteln("Editor writing to", file);
	// Try to write out the data.
	try {
	    // Get a stream to write to
	    OutputStream out = 
		new BufferedOutputStream(
		    new FileOutputStream(file));
	    // Create an outputter with 3 space indent and newlines=true
	    XMLOutputter outputter = new XMLOutputter("  ", true);
	    outputter.setLineSeparator(System.getProperty("line.separator"));
	    // deprecated in beta-7 & wrong anyway:
	    // outputter.setTrimText(true);
	    outputter.setTextNormalize(false);
	    outputter.output(doc, out);
	    out.flush();
	    return true;	// success
	}
	catch (IOException e) {
	    JOptionPane.showMessageDialog(this,
		 new Object[] {
		     "Problem saving " + Strings.quote(file.getPath()),
		     e.getMessage()},
		 "Trouble saving to file",
		 JOptionPane.ERROR_MESSAGE);
	    return false;	// failure
	}
    }

    protected File chooseFileToWrite(Component frame, File directory) {
	JFileChooser chooser = new JFileChooser(directory);
	chooser.setFileFilter(new XMLFileFilter());
	int option = chooser.showSaveDialog(frame);
	if (option == JFileChooser.APPROVE_OPTION)
	    return chooser.getSelectedFile();
	else
	    return null;
    }

}
