/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Tue Jun 11 17:11:58 2002 by Jeff Dalton
 * Copyright: (c) 2002, AIAI, University of Edinburgh
 */

package ix.util.xml;

import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.event.*;

import java.awt.Container;
import java.awt.Component;
import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.event.*;

import java.util.*;

// import java.io.StringReader;
import java.io.StringWriter;
import java.io.IOException;

// Imports for using JDOM
// import java.io.IOException;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Attribute;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;

import ix.iface.util.*;

import ix.util.*;
import ix.util.lisp.*;

/**
 * A panel that contains an XML tree-editor.
 *
 * @see XMLTreeEditor
 * @see XMLTreeEditFrame
 */

public class XMLTreeEditPanel extends JPanel
       implements ActionListener,
		  TreeSelectionListener {

    /** The document this panel is editing.
     */
    protected Document document;

    protected TNodeFactory nodeFactory = new TNodeFactory();
    protected TNodeParser nodeParser = new TNodeParser();

    protected TemplateSyntax syntax = new TemplateSyntax();

    protected EditorTree docTree;
    protected DefaultTreeModel docModel;

    protected EditorTree templateTree;

    protected EditorTree mostRecentlySelectedTree;

    protected EditingTextArea editText;
    protected JButton submitButton;

    public XMLTreeEditPanel() {
	// Default is FlowLayout.
	super(new BorderLayout());

	// Root node
	TNode docRoot = new TNode("root");
	docRoot.setShouldBeExpanded(true);

	// Tree model
	docModel = new DefaultTreeModel(docRoot);
	// docModel.setAsksAllowsChildren(true);

	// Doc Tree
	docTree = new EditorTree("document tree", docModel);
        docTree.setEditable(true);
        docTree.getSelectionModel()
	    .setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
        docTree.setShowsRootHandles(true);
	docTree.addTreeSelectionListener(this);
	docTree.addMouseListener(new TreeMouseListener());

	// Template tree
	templateTree = makeTemplateTree();
	addTemplatesFor(new Class[] {
	    ix.icore.Activity.class,
	    ix.icore.Issue.class,
	    ix.icore.Report.class,
	    ix.icore.domain.Domain.class,
	    ix.icore.domain.PatternAssignment.class,
	    ix.ip2.TestItem.class
	});
	addTemplate(syntax.makeMapEntryNode());

	// Text editing
	editText = new EditingTextArea(2, 30);
	editText.setEditable(false);
	submitButton = makeButton("Submit Text");
	submitButton.setEnabled(false);

	// Scroll panes and borders
        JScrollPane docScroll = new JScrollPane(docTree);
	docScroll.setBorder
	    (BorderFactory.createTitledBorder("Document"));

	JScrollPane templateScroll = new JScrollPane(templateTree);
	templateScroll.setBorder
	    (BorderFactory.createTitledBorder("Templates"));

	JScrollPane textScroll = new JScrollPane(editText);
	textScroll.setBorder
	    (BorderFactory.createTitledBorder("Text"));

	// Split panes
	JSplitPane split = new JSplitPane
	    (JSplitPane.HORIZONTAL_SPLIT, docScroll, templateScroll);
	split.setDividerSize(3);
	split.setResizeWeight(0.80);

	JSplitPane split2 = new JSplitPane
	    (JSplitPane.VERTICAL_SPLIT, split, textScroll);
	split2.setDividerSize(3);
	split2.setResizeWeight(0.90);

	// Panel contents
        add(split2, BorderLayout.CENTER);
	add(makeButtonPanel(), BorderLayout.SOUTH);

    }

    protected void reset() {
	setDocRoot(new TNode("root"));
	editText.clear();
	// /\/: Need to reset template tree
    }

    protected void setDocRoot(TNode root) {
	docModel.setRoot(root);
	root.setShouldBeExpanded(true);
    }

    public void editDocument(Document doc) {
	document = doc;
	setDocRoot(nodeFactory.nodeFrom(doc));
    }

    public Document getDocument() {
	document = nodeParser.documentFrom((TNode)docModel.getRoot());
	return document;
    }

    public void editObject(Object object) {
	editDocument(syntax.xmlt().objectToDocument(object));
    }

    public Object getObject() {
	Document doc = getDocument(); // will select culprit node if error
	try {
	    return syntax.xmlt().objectFromDocument(doc);
	}
	catch (Exception e) {
	    Debug.noteException(e);
	    // Try to show and select the responsible node in the docTree.
	    Element probableBadElt = syntax.xmlt().getLastConvertedElement();
	    if (probableBadElt != null) {
		Debug.noteln("Bad elt", probableBadElt);
		TNode culprit = nodeParser.getNodeForElement(probableBadElt);
		if (culprit != null) {
		    TreePath path = culprit.getTreePath();
		    Debug.noteln("Bad node's path", path);
		    docTree.scrollPathToVisible(path);
		    docTree.setSelectionPath(path);
		}
	    }
	    throw new InvalidNode(e.toString());
	}
    }

    public void addTemplatesFor(Class[] classes) {
	addTemplatesFor(Seq.toLList(Seq.elements(classes)));
    }

    public void addTemplatesFor(List classes) {
	DefaultTreeModel model = (DefaultTreeModel)templateTree.getModel();
	TNode root = (TNode)model.getRoot();

	List templates = syntax.makeTemplateNodes(classes);
	for (Iterator i = templates.iterator(); i.hasNext();) {
	    TNode t = (TNode)i.next();
	    root.add(t);
	}
	model.nodeStructureChanged(root);
    }

    public void addTemplate(TNode template) {
	DefaultTreeModel model = (DefaultTreeModel)templateTree.getModel();
	TNode root = (TNode)model.getRoot();
	root.add(template);
	model.nodeStructureChanged(root);
    }

    protected EditorTree makeTemplateTree() {
	TNode root = new TNode("Templates");
	root.setShouldBeExpanded(true);
	EditorTree tree = new EditorTree("template tree", root);
        tree.setEditable(false);
        tree.getSelectionModel()
	    .setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
        tree.setShowsRootHandles(true);
	tree.addTreeSelectionListener(this);
	tree.addMouseListener(new TreeMouseListener());
	return tree;
    }

    protected JPanel makeButtonPanel() {
	JPanel panel = new JPanel(); 	// defaults to FlowLayout
	panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS));
	panel.add(Box.createHorizontalGlue());
	panel.add(makeButton("Add Child"));
	panel.add(makeButton("Copy Subtree"));
	panel.add(makeButton("Cut Subtree"));
	panel.add(submitButton);
	panel.add(Box.createHorizontalGlue());
	return panel;
    }

    protected JButton makeButton(String command) {
	JButton b = new JButton(command);
	b.addActionListener(CatchingActionListener.listener(this));
	return b;
    }

    /**
     * Action interpreter for panel buttons
     */
    public void actionPerformed(ActionEvent e) {
	String command = e.getActionCommand();
	Debug.noteln("XMLTreeEditPanel action:", command);
	if (command.equals("Add Child")) {
	    addChild();
	}
	else if (command.equals("Copy Subtree")) {
	    copySubtree();
	}
	else if (command.equals("Cut Subtree")) {
	    cutSubtree();
	}
	else if (command.equals("Submit Text")) {
	    editText.submit();
	}
	else
	    Debug.noteln("Nothing to do for", command);
    }

    /**
     * Adds a new node after the existing children of the selected
     * node.
     */
    public void addChild() {
	TNode selected = (TNode)docTree.getLastSelectedPathComponent();
	if (selected == null)
	    complain("No selected node in document");
	else {
	    TNode newNode = new TNode("blank");
	    docModel.insertNodeInto
		(newNode, selected, selected.getChildCount());
	    docTree.scrollPathToVisible(newNode.getTreePath());
	}
    }

    /** 
     * Copy the subtree below and including the currently selected node.
     * This is essentially the same as a "Cut" without deletion, but it
     * also takes a copy of the subtree rather than taking those very
     * nodes.
     */ 
    public void copySubtree() {
	TNode selected = (TNode)docTree.getLastSelectedPathComponent();
	if (selected == null)
	    complain("No selected node in document");
	else
	    new CutFrame(selected.copySubtree(), "Copy");
    }

    /** 
     * "Cut" the subtree below and including the currently selected node.
     */ 
    public void cutSubtree() {
	TNode selected = (TNode)docTree.getLastSelectedPathComponent();
	if (selected == null)
	    complain("No selected node in document");
	else {
	    TNode parent = (TNode)selected.getParent();
	    if (parent != null) {
		docModel.removeNodeFromParent(selected);
		new CutFrame(selected, "Cut");
	    }
	    else
		complain("The root cannot be cut; try File: Reset");
	}
    }

    protected void complain(Object message) {
	JOptionPane.showMessageDialog
	    (this, message, "Error", JOptionPane.ERROR_MESSAGE);
    }

    class InvalidCommand extends RuntimeException {
	InvalidCommand(String message) {
	    super(message);
	}
    }

    /*
     * TreeSelectionListener methods.
     * 
     * Called for both the document and the template trees.
     */

    public void valueChanged(TreeSelectionEvent e) {
	EditorTree source = (EditorTree)e.getSource();
        TNode node = (TNode)source.getLastSelectedPathComponent();

        if (node == null)
	    return;

	Debug.noteln(source.getName() + " selection", node.getTreePath());

	if (mostRecentlySelectedTree != null
	        && mostRecentlySelectedTree != source) {
	    Debug.noteln("Changing selected tree");
	    mostRecentlySelectedTree.getSelectionModel().clearSelection();
	}
	mostRecentlySelectedTree = source;

	TNode parent = (TNode)node.getParent();
	if (parent != null && nodeParser.isAtomicValue(parent))
	    editText.display(parent);
	else
	    editText.clear();

    }


    /*
     * Node classes
     */

    /**
     * A TreeNode for use in EditorTrees.
     */
    class TNode extends DefaultMutableTreeNode {

	boolean shouldBeExpanded = false;

	TNode(Object userObject) {
	    super(userObject);
	}

	boolean shouldBeExpanded() {
	    return shouldBeExpanded;
	}

	void setShouldBeExpanded(boolean sbe) {
	    shouldBeExpanded = sbe;
	}

	void setSubtreeShouldBeExpanded(boolean sbe) {
	    for (Enumeration e = preorderEnumeration();
		 e.hasMoreElements();) {
		TNode n = (TNode)e.nextElement();
		n.setShouldBeExpanded(sbe);
	    }
	}

	public TreePath getTreePath() {
	    return new TreePath(getPath());
	}

	public void setUserObject(Object object) {
	    Debug.noteln("Setting user object to", object);
	    super.setUserObject(object);
	}

	public TNode copySubtree() {
	    TNode result = new TNode(getUserObject());
	    result.shouldBeExpanded = shouldBeExpanded;
	    for (Enumeration e = children(); e.hasMoreElements();) {
		TNode child = (TNode)e.nextElement();
		result.add(child.copySubtree());
	    }
	    return result;
	}

    }

    /**
     * Makes nodes from JDOM Documents and Elements.
     */
    class TNodeFactory {

	TNodeFactory() { }

	TNode nodeFrom(Document doc) {
	    return nodeFrom(doc.getRootElement());
	}

	TNode nodeFrom(Element elt) {
	    TNode node = new TNode(elt.getName());
	    if (elt.getAttributes().isEmpty()) {
		// No attributes
		if (elt.getChildren().isEmpty()) {
		    // No children either - so take only the text
		    node.add(new TNode(elt.getText()));
		    return node;
		}
	    }
	    else {
		// Add attributes
		node.add(makeAttributesNode(elt));
	    }
	    // Add any children
	    for (Iterator i = elt.getChildren().iterator(); i.hasNext();) {
		Element fieldElt = (Element)i.next();
		node.add(nodeFrom(fieldElt));
	    }
	    return node;
	}

	TNode makeAttributesNode(Element elt) {
	    TNode attrNode = new TNode("attributes");
	    for (Iterator i = elt.getAttributes().iterator(); i.hasNext();) {
		Attribute fieldAttr = (Attribute)i.next();
		String fieldName = fieldAttr.getName();
		String fieldValue = fieldAttr.getValue();
		attrNode.add(new TNode(fieldName + "=" + fieldValue));
	    }
	    return attrNode;
	}

    }

    /**
     * Responsible for understanding the object-like structure of tree
     * nodes and for converting nodes into JDOM Documents or Elements.
     */
    class TNodeParser {

	Map elementToNodeMap = new HashMap();

	TNodeParser() { }

	Document documentFrom(TNode node) {
	    elementToNodeMap.clear();
	    Element rootElt = elementFrom(node);
	    return new Document(rootElt);
	}

	Element elementFrom(TNode node) {
	    Element result;
	    if (isAttributesList(node))
		throw syntaxError(node, "Misplaced attributes");
	    else if (isAtomicValue(node))
		result = valueElementFrom(node);
	    else
		result = structElementFrom(node);
	    elementToNodeMap.put(result, node);
	    return result;
	}

	TNode getNodeForElement(Element elt) {
	    return (TNode)elementToNodeMap.get(elt);
	}

	protected Element valueElementFrom(TNode node) {
	    String name = (String)node.getUserObject();
	    TNode valueNode = (TNode)node.getChildAt(0);
	    String value = (String)valueNode.getUserObject();
	    Element elt = new Element(name);
	    elt.setText(value);
	    return elt;
	}

	protected Element structElementFrom(TNode node) {
	    String name = (String)node.getUserObject();
	    Enumeration children = node.children();

	    Element elt = new Element(name);
	    if (hasAttributes(node)) {
		addAttributes(elt, node);
		children.nextElement();		// remove atributes
	    }
	    while (children.hasMoreElements()) {
		TNode child = (TNode)children.nextElement();
		elt.addContent(elementFrom(child));
	    }
	    return elt;

	}

	protected void addAttributes(Element elt, TNode node) {
	    TNode attributesNode = (TNode)node.getChildAt(0);
	    for (Enumeration e = attributesNode.children();
		 e.hasMoreElements();) {
		TNode attrNode = (TNode)e.nextElement();
		addAttribute(elt, attrNode);
	    }
	}

	protected void addAttribute(Element elt, TNode attrNode) {
	    String[] parts = parseAttribute(attrNode);
	    elt.setAttribute(parts[0], parts[1]);
	}

	String[] parseAttribute(TNode attrNode) {
	    if (!isAttribute(attrNode))
		throw syntaxError(attrNode, "Invalid attribute");
	    String text = (String)attrNode.getUserObject();
	    String[] parts = Strings.breakAtFirst("=", text);
	    if (attrNode.getChildCount() != 0
		    || parts[0].equals("")
		 /* || parts[1].equals("") */ )
		throw syntaxError(attrNode, "Invalid attribute");
	    return parts;
	}

	boolean hasAttributes(TNode node) {
	    return node.getChildCount() >= 1
		&& isAttributesList((TNode)node.getChildAt(0));
	}

	boolean isAttributesList(TNode node) {
	    return node.getUserObject().equals("attributes");
	}

	boolean isAttribute(TNode node) {
	    TNode parent = (TNode)node.getParent();
	    return parent != null
		&& isAttributesList(parent)
		&& node.getChildCount() == 0;
	}

	String getAttributeName(TNode node) {
	    String[] parts = parseAttribute(node);
	    return parts[0];
	}

	String getAttributeValue(TNode node) {
	    String[] parts = parseAttribute(node);
	    return parts[1];
	}

	TNode getAttributeContainingObjectNode(TNode attrNode) {
	    // The node that corresponds to the object that has
	    // the attribute as a field.
	    Debug.assert(isAttribute(attrNode));
	    TNode typeNode = (TNode)attrNode.getParent().getParent();
	    if (typeNode == null)
		throw syntaxError(attrNode, "Attribute outside object");
	    return typeNode;
	}

	boolean isAtomicValue(TNode node) {
	    return node.getChildCount() == 1
		&& !isAttributesList(node)
		&& !isAttributesList((TNode)node.getChildAt(0))
		&& node.getChildAt(0).getChildCount() == 0;
	}

	void setAtomicValue(TNode node, Object userObj) {
	    Debug.assert(isAtomicValue(node));
	    TNode valChild = (TNode)node.getChildAt(0);
	    valChild.setUserObject(userObj);
	}

	InvalidNode syntaxError(TNode culprit, String message) {
	    TreePath path = culprit.getTreePath();
	    docTree.scrollPathToVisible(path);
	    docTree.setSelectionPath(path);
	    return new InvalidNode(message);
	}

    }

    class InvalidNode extends RuntimeException {
	InvalidNode(String message) {
	    super(message);
	}
    }

    /**
     * Knows about the syntax of Java classes.
     */
    class TemplateSyntax extends Syntax {

	TemplateSyntax() {
	    super(new DashSyntaxXMLTranslator());
	}

	List makeTemplateNodes(List topClasses) {
	    // First find all relevant classes.
	    List relevantClasses = new LinkedList();
	    for (Iterator i = topClasses.iterator(); i.hasNext();) {
		Class c = (Class)i.next();
		walkRelevantClasses(c, relevantClasses);
	    }
	    // Create a template TNode for each relevant class
	    // if this will be useful and we know how.
	    List result = new LinkedList();
	    for (Iterator i = relevantClasses.iterator(); i.hasNext();) {
		Class c = (Class) i.next();
		TNode template = templateFrom(c);
		if (template != null) 		// useful and know how?
		    result.add(template);
	    }
	    return result;
	}

	TNode templateFrom(Class c) {
	    ClassDescr cd = new ClassDescr(c);
	    if (cd.isStruct())
		return structTemplateFrom(c);
	    else
		return null;
	}

	TNode structTemplateFrom(Class c) {
	    String className = xmlNameForClass(c);
	    List fields = getFields(c);
	    List attrFields = attributeFields(fields);
	    List eltFields = elementFields(fields);
	    Debug.assert(fields.size() == attrFields.size()
			                  + eltFields.size());
	    TNode result = new TNode(className);
	    if (attrFields.isEmpty()) {
		if (eltFields.isEmpty()) return null;
	    }
	    else
		result.add(makeAttributesNode(attrFields));
	    for (Iterator ei = eltFields.iterator(); ei.hasNext();) {
		FieldDescr fd = (FieldDescr)ei.next();
		result.add(makeFieldNode(fd));
	    }
	    return result;
	}

	TNode makeAttributesNode(List attrFields) {
	    TNode attrNode = new TNode("attributes");
	    for (Iterator ai = attrFields.iterator(); ai.hasNext();) {
		FieldDescr fd = (FieldDescr)ai.next();
		String attrName = fd.getElementName();
		attrNode.add(new TNode(attrName + "="));
	    }
	    return attrNode;
	}

	TNode makeFieldNode(FieldDescr fd) {
	    TNode fieldNode = new TNode(fd.getElementName());

	    if (fd.type.isList()) {
		String listName = xmlNameForClass(List.class);
		TNode listNode = new TNode(listName);
		listNode.add(new TNode(listName + "-element..."));
		fieldNode.add(listNode);
	    }
	    else if (fd.type.isSet()) {
		String setName = xmlNameForClass(Set.class);
		TNode setNode = new TNode(setName);
		setNode.add(new TNode(setName + "-element..."));
		fieldNode.add(setNode);
	    }
	    else if (fd.type.isMap()) {
		String mapName = xmlNameForClass(Map.class);
		TNode mapNode = new TNode(mapName);
		mapNode.add(makeMapEntryNode());
		fieldNode.add(mapNode);
	    }
	    else {
		String fieldClassName = xmlNameForClass(fd.type.theClass);
		TNode fieldValueNode = new TNode(fieldClassName);
		fieldValueNode
		    .add(new TNode(fieldClassName + "-value"));
		fieldNode.add(fieldValueNode);
	    }
	    return fieldNode;
	}

	TNode makeMapEntryNode() {
	    String entryName = xmlt.externalName("MapEntry");
	    TNode entryNode = new TNode(entryName);
	    entryNode.add(new TNode("key"));
	    entryNode.add(new TNode("value"));
	    return entryNode;
	}

	XMLTranslator xmlt() {
	    return xmlt;
	}

	String xmlNameForClass(Class c) {
	    return xmlt.externalNameForClass(c);
	}

	Class classForXmlName(String name) {
	    return xmlt.classForExternalName(name);
	}

	Class getFieldClass(Class c, String fieldName) {
	    // fieldName is the true / internal name
	    List fields = getFields(c);
	    for (Iterator fi = fields.iterator(); fi.hasNext();) {
		FieldDescr fd = (FieldDescr)fi.next();
		if (fd.name.equals(fieldName))
		    return fd.type.theClass;
	    }
	    return null;
	}

	List getEnumerationValues(Class c) {
	    if (new ClassDescr(c).isEnumeration())
		return (List)Fn.applyStatic(c, "values", new Object[]{});
	    else
		return null;	// not an enumeration
	}

    }

    /*
     * Mouse listener
     *
     * /\/: There are separate listeners for the document and template
     * trees.  The main reason for using a separate class is that
     * it can extend MouseAdapter; however, for symmetry, it might
     * be better for the TreeSelectionListener to be handled in the
     * same way even though there isn't an adapter class to extend.
     */

    class TreeMouseListener extends MouseAdapter {

	public void mousePressed(MouseEvent e) {
	    EditorTree selTree = (EditorTree)e.getSource();
	    int selRow = selTree.getRowForLocation(e.getX(), e.getY());
	    TreePath selPath = selTree.getPathForLocation(e.getX(), e.getY());

	    if (selRow == -1)
		return;
	    if (!SwingUtilities.isRightMouseButton(e))
		return;

	    Debug.noteln("Right click in " + selTree.getName(), selPath);
	    TNode selNode = (TNode)selPath.getLastPathComponent();
	    NodePopupMenu popup = new NodePopupMenu(selTree, selNode);
	    popup.show(e.getComponent(), e.getX(), e.getY());
	}

    }


    /**
     * The popup menu that appears when the user right-clicks on a tree node.
     */
    class NodePopupMenu extends JPopupMenu
                        implements ActionListener {

	EditorTree tree;
	DefaultTreeModel model;
	TNode node;

	NodePopupMenu(EditorTree tree, TNode node) {
	    super();
	    this.tree = tree;
	    this.model = (DefaultTreeModel)tree.getModel();
	    this.node = node;

	    if (nodeParser.isAttribute(node))
		addValueMenuItemIfKnown();

	    add(makeMenuItem("Fully Expand"));
	    add(makeMenuItem("Fully Collapse"));
	    add(makeMenuItem("Add Child"));
	    add(makeMenuItem("Copy Subtree"));
	    add(makeMenuItem("Cut Subtree"));

	    if (nodeParser.isAttributesList(node))
		add(makeMenuItem("Delete Valueless Attributes"));
	    else if (node.getChildCount() > 0)
		add(makeMenuItem("Delete Closed Subnodes"));

	}

	JMenuItem makeMenuItem(String text) {
	    JMenuItem item = new JMenuItem(text);
	    item.addActionListener(CatchingActionListener.listener(this));
	    return item;
	}

	void addValueMenuItemIfKnown() {
	    // We have to find the class of the object whose attribute
	    // it is, and then look up the class of the field that
	    // corresponds to the attribute.
	    TNode objNode = nodeParser
		.getAttributeContainingObjectNode(node);
	    String objClassName = (String)objNode.getUserObject();
	    Class objClass = syntax.classForXmlName(objClassName);
	    String fieldName = nodeParser.getAttributeName(node);
	    Class fieldClass = syntax.getFieldClass
		(objClass,
		 syntax.xmlt().internalFieldName(fieldName));
	    if (fieldClass != null) {
		List values = syntax.getEnumerationValues(fieldClass);
		if (values != null) {
		    JMenu valueMenu = makeValueMenu(values);
		    add(valueMenu);
		}
	    }
	}

	JMenu makeValueMenu(List values) {
	    JMenu menu = new JMenu("Set Value");
	    for (Iterator i = values.iterator(); i.hasNext();) {
		Object value = i.next();
		JMenuItem item = makeMenuItem(value.toString());
		item.setActionCommand("setValue");
		menu.add(item);
	    }
	    return menu;
	}

	public void actionPerformed(ActionEvent e) {
	    String command = e.getActionCommand();
	    Debug.noteln("Item popup command", command);
	    if (command.equals("Fully Expand")) {
		fullyExpand();
	    }
	    else if (command.equals("Fully Collapse")) {
		fullyCollapse();
	    }
	    else if (command.equals("Add Child")) {
		addChild();
	    }
	    else if (command.equals("Copy Subtree")) {
		copySubtree();
	    }
	    else if (command.equals("Cut Subtree")) {
		cutSubtree();
	    }
	    else if (command.equals("setValue")) {
		JMenuItem item = (JMenuItem)e.getSource();
		String value = item.getText();
		setValue(value);
	    }
	    else if (command.equals("Delete Valueless Attributes")) {
		deleteValuelessAttributes();
	    }
	    else if (command.equals("Delete Closed Subnodes")) {
		deleteClosedSubnodes();
	    }
	    else {
		Debug.noteln("Nothing to do for", command);
	    }
	}

	void fullyExpand() {
	    node.setSubtreeShouldBeExpanded(true);
	    tree.fixExpansions();
	}

	void fullyCollapse() {
	    node.setSubtreeShouldBeExpanded(false);
	    tree.fixExpansions();
	}

	void addChild() {
	    TNode newChild = new TNode("blank");
	    model.insertNodeInto
		(newChild, node, node.getChildCount());
	    tree.scrollPathToVisible(newChild.getTreePath());
	}

	void copySubtree() {
	    new CutFrame(node.copySubtree(), "Copy");
	}

	void cutSubtree() {
	    TNode parent = (TNode)node.getParent();
	    if (parent != null) {
		if (tree == templateTree
		    && !Util.dialogConfirms(XMLTreeEditPanel.this,
			    "Are you sure you want to cut a template?"))
		    return;
		model.removeNodeFromParent(node);
		new CutFrame(node, "Cut");
	    }
	    else
		throw new InvalidCommand
		    ("A root cannot be cut");
	}

	void setValue(String value) {
	    String text = (String)node.getUserObject();
	    String typeName = Strings.beforeFirst("=", text);
	    node.setUserObject(typeName + "=" + value);
	    model.nodeChanged(node);
	}

	void deleteValuelessAttributes() {
	    // Must extract a separate list of the children so that
	    // we're not going through the node.children() Enumeration
	    // while we're deleting.
	    LList children = Seq.toLList(node.children());
	    for (Iterator i = children.iterator(); i.hasNext();) {
		TNode child = (TNode)i.next();
		if (nodeParser.getAttributeValue(child).trim().equals(""))
		    model.removeNodeFromParent(child);
	    }
	}

	void deleteClosedSubnodes() {
	    // Must extract a separate list of the children so that
	    // we're not going through the node.children() Enumeration
	    // while we're deleting.
	    LList children = Seq.toLList(node.children());
	    for (Iterator i = children.iterator(); i.hasNext();) {
		TNode child = (TNode)i.next();
		if (!child.shouldBeExpanded())
		    model.removeNodeFromParent(child);
	    }
	}

    }

    /**
     * JTree subclass used for all trees in the editor.
     */
    class EditorTree extends JTree {

	// Inherits get/setName from Component

	RecordingExpansionListener expansionListener
	    = new RecordingExpansionListener(this);

	EditorTree(String name, TreeNode root) {
	    this(name, new DefaultTreeModel(root));
	}

	EditorTree(String name, TreeModel model) {
	    super(model);
	    setName(name);
	    addTreeExpansionListener(expansionListener);
	    putClientProperty("JTree.lineStyle", "Angled");

	    // Here's how we would change the "folder" icons
// 	    DefaultTreeCellRenderer tcr = new DefaultTreeCellRenderer();
// 	    tcr.setOpenIcon(Util.resourceImageIcon("ix-symbol-fold.gif"));
// 	    tcr.setClosedIcon(Util.resourceImageIcon("ix-symbol-unfold.gif"));
// 	    tcr.setLeafIcon(Util.resourceImageIcon("ix-symbol-unfold.gif"));
// 	    setCellRenderer(tcr);

	}

	void fixExpansions() {
	    fixExpansions((TNode)getModel().getRoot());
	}

	void fixExpansions(TNode node) {
    	    try {
		expansionListener.active = false;
		fixExpansionsRec(node);
	    }
	    finally {
		expansionListener.active = true;
	    }
	}

	private void fixExpansionsRec(TNode node) {
	    // First, recursively process children
	    for (Enumeration e = node.children(); e.hasMoreElements();) {
		fixExpansionsRec((TNode)e.nextElement());
	    }
	    // Now this node
	    TreePath path = node.getTreePath();
	    if (node.shouldBeExpanded())
		expandPath(path);
	    else
		collapsePath(path);
	}

    }

    class RecordingExpansionListener implements TreeExpansionListener {

	JTree tree;
	boolean active = true;

	RecordingExpansionListener(JTree tree) {
	    this.tree = tree;
	}

	public void treeExpanded(TreeExpansionEvent e) {
	    if (active) {
		Debug.noteln("Tree expanding", e.getPath());
		theNode(e).setShouldBeExpanded(true);
	    }
	}

	public void treeCollapsed(TreeExpansionEvent e) {
	    if (active) {
		Debug.noteln("Tree collapsing", e.getPath());
		theNode(e).setShouldBeExpanded(false);
	    }
	}

	TNode theNode(TreeExpansionEvent e) {
	    Debug.assert(e.getSource() == tree);
	    return (TNode)e.getPath().getLastPathComponent();
	}

    }

    /**
     * A text area for editing string values
     */
    class EditingTextArea extends JTextArea {

	TNode editingNode;
	TNode valueNode;
	Element editingElt;

	EditingTextArea(int height, int width) {
	    super(height, width);
	}

	void display(TNode node) {
	    editingNode = node;
	    editingElt = nodeParser.valueElementFrom(node);
	    setText(editingElt.getText());
	    setEditable(true);
	    submitButton.setEnabled(true);
	}

	void clear() {
	    if (editingNode != null) {
		editingNode = null;
		editingElt = null;
		setText("");
		setEditable(false);
		submitButton.setEnabled(false);
	    }
	}

	void submit() {
	    editingElt.setText(getText());
	    nodeParser.setAtomicValue(editingNode, editingElt.getText());
	    ((DefaultTreeModel)mostRecentlySelectedTree.getModel())
		.nodeStructureChanged(editingNode);
	    clear();
	}

    }


    /**
     * A frame that shows a subtree cut from the main editor panel.
     */
    class CutFrame extends JFrame implements ActionListener {

	EditorTree subtree;
	TNode subtreeNode;

	EditorTree destinationTree;
	DefaultTreeModel destinationModel;

	Container contentPane = getContentPane();

	CutFrame(TNode subtreeNode, String title) {
	    super(title);
	    this.subtreeNode = subtreeNode;

	    // Tree
	    subtree = new EditorTree(title + " tree", subtreeNode);
	    subtree.setEditable(false);
	    subtree.setSelectionModel(null);
	    subtree.setShowsRootHandles(true);
	    subtree.fixExpansions();

	    // Frame contents
	    JScrollPane scrollPane = new JScrollPane(subtree);
	    scrollPane.setBorder
		(BorderFactory.createTitledBorder("Subtree"));
	    contentPane.add(scrollPane, BorderLayout.CENTER);
	    JPanel buttons = makeButtonPanel();
	    buttons.setBorder
		(BorderFactory.createTitledBorder("Insert ..."));
	    contentPane.add(buttons, BorderLayout.SOUTH);

	    // pack();
	    setSize(400, 300);	// width, height
	    setVisible(true);
	}

	JPanel makeButtonPanel() {
	    JPanel panel = new JPanel(); 	// defaults to FlowLayout
	    panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS));
	    panel.add(Box.createHorizontalGlue());
	    panel.add(makeButton("As"));
	    panel.add(makeButton("Before"));
	    panel.add(makeButton("After"));
	    panel.add(makeButton("Child"));
	    panel.add(makeButton("Discard"));
	    panel.add(Box.createHorizontalGlue());
	    return panel;
	}

	JButton makeButton(String command) {
	    JButton b = new JButton(command);
	    b.addActionListener(CatchingActionListener.listener(this));
	    return b;
	}

	/**
	 * Action interpreter for cut-frame buttons
	 */
	public void actionPerformed(ActionEvent e) {
	    String command = e.getActionCommand();
	    Debug.noteln("XMLTreeEditPanel action:", command);

	    if (command.equals("Discard")) {
		subtreeNode = null;
		finished();
		return;
	    }

	    destinationTree = mostRecentlySelectedTree;
	    if (destinationTree == null) {
		complain("No selected destination");
		return;
	    }
	    destinationModel = (DefaultTreeModel)destinationTree.getModel();

	    TNode selected =
		(TNode)destinationTree.getLastSelectedPathComponent();
	    if (selected == null) {
		complain("No selected node");
		return;
	    }

	    if (command.equals("As")) {
		insertAs(selected);
	    }
	    if (command.equals("Before")) {
		insertBefore(selected);
	    }
	    else if (command.equals("After")) {
		insertAfter(selected);
	    }
	    else if (command.equals("Child")) {
		insertUnder(selected);
	    }
	    else
		Debug.noteln("Nothing to do for", command);
	}

	void insertAs(TNode selected) {
	    TNode parent = (TNode)selected.getParent();
	    if (parent == null) {
		destinationModel.setRoot(subtreeNode);
	    }
	    else {
		int i = destinationModel.getIndexOfChild(parent, selected);
		destinationModel.removeNodeFromParent(selected);
		destinationModel.insertNodeInto(subtreeNode, parent, i);
	    }
	    finished();
	}

	void insertBefore(TNode selected) {
	    TNode parent = (TNode)selected.getParent();
	    if (parent == null)
		complain("Cannot insert before root");
	    else {
		destinationModel.insertNodeInto
		    (subtreeNode,
		     parent,
		     destinationModel.getIndexOfChild(parent, selected));
		finished();
	    }
	}

	void insertAfter(TNode selected) {
	    TNode parent = (TNode)selected.getParent();
	    if (parent == null)
		complain("Cannot insert at same level as root");
	    else {
		destinationModel.insertNodeInto
		    (subtreeNode,
		     parent,
		     destinationModel.getIndexOfChild(parent, selected) + 1);
		finished();
	    }
	}

	void insertUnder(TNode selected) {
	    destinationModel.insertNodeInto(subtreeNode, selected, 0);
	    finished();
	}

	void finished() {
	    if (subtreeNode != null) {
		// Subtree was inserted somewhere
		destinationTree.fixExpansions();
		destinationTree.scrollPathToVisible(subtreeNode.getTreePath());
	    }
	    // /\/: Need to really get rid of this frame somehow.
	    // setVisible(false);
	    dispose();
	}

    }

}
